
// Copyright (c) 2009 Massachusetts Institute of Technology

// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:

// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.

// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
/*
JV3.Analysis.engines.datejs = {
    name: "datejs",
    initialize: function(analysis) {
	this.analysis = analysis;
	this.log = this.analysis.JV3.log;
	this.log("initializing datejs");
	xuload("chrome://listit/content/frameworks/datejs/date-en-US.js"); // this might possibly break something!
    },
    analyze:function(string, continuation) {
	// if you provide a continuation, we will use a "slowmap" and return asynchronously 
	var this_ = this;
	var results = {};
	this.analysis.filter_over_substrings(
	    // test function 
	    this.test,  // --> to be "this"-safe: function(start,end,s) { return this_.test.apply(this_, arguments); },
	    // input string 
	    string,
	    // success 
	    function(s,result) { this_.analysis.merge_results_choose_maximum_spanning_parses(results,result);  },
	    // continuation? 
	    continuation ? { continuation: continuation } : undefined );
	return results;
    },
    test:function(start,end,s)  {
	var p = Date.parse(s);
	if (p) {
	    return { success: p !== null && p !== undefined,
		     start: start, end: end,
		     value: {type:"date", exact:true, timestamp:p.valueOf()}, //(p !== null && p !== undefined ? JV3.plumutil.toJSON(p) : undefined),
		     type: "date" };
	} 
	return false;
    }
};
*/


(function(JV3,analysis) {

    try {
	JV3.xuload("chrome://listit/content/analysis/plumdate.js")(JV3,analysis);
    } catch(e) {
	JV3.log(e + " " + e.lineNumber + " " + e.fileName);
    }
    
    analysis.engines.plumdate = {
	name: "plumdate",
	initialize: function() {
	    try {
	    JV3.log("plumdate is " + analysis.PlumDate);
	    this.log = JV3.log;
	    this.util = JV3.plumutil;
	    analysis.PlumDate.log = this.log;
	    analysis.PlumDate.init();
	    this.log("::initializing plumdate::" + analysis.PlumDate);
	    } catch(e) {
		JV3.log(e + " " + e.lineNumber + " " + e.fileName);
	    }
	},    
	analyze:function(string, continuation) {
	    /* if you provide a continuation, we will use a "slowmap" and return asynchronously */
	    this.log("JV3.Analysis.engines.plumdate.analyze() >> " + string.substring(0,25).replace("\n"," "));
	    var this_ = this;
	    var results = [];	
	    
	    var successes = {};
	    var util =this.util;

	    try { 
		// boobar.
		var ptr = 0;
		while (ptr < string.length) {
		    var nextn = string.indexOf("\n",ptr);
		    var line = nextn >= 0 ? string.substring(ptr,nextn) : string.substring(ptr) ;
		    this.log("engines.analyze(): analyzing line: " + line);
		    var lineparses = analysis.PlumDate.parse(line);
		    this.log("engines.analyze(): parses on line: " + lineparses.length);
		    lineparses.map(function(parse) {
			try {
			    var tomatch = parse.text;
			    var findNext = function(s,frompos) {
				if (frompos === undefined) { frompos = 0; }
				// returns the index of the next tomatch in s
				if (tomatch instanceof RegExp) {
				    return s.substring(frompos).search(tomatch)+frompos;
				}
				return s.indexOf(tomatch,frompos);
			    };
			    var nextMatch = function(s, frompos) {
				if (frompos === undefined) { frompos = 0; }
				if (tomatch instanceof RegExp) {
				    var matched = s.substring(frompos).match(tomatch);
				    if (matched) { return matched[0]; }
				} else {
				    if (s.substring(frompos).indexOf(tomatch) >= 0) {
					return tomatch;
				    }
				}
			    };
			    // sometimes we get an array back, which means we get a date-time combination6
			    if (this_.util.isArrayLike(tomatch)) {
				tomatch = new RegExp(tomatch.join(".*"),"g");
				this_.log("!!!!!!! tomatch is a REGEXP : " + tomatch);
			    }
			    var inext = findNext(line);
			    var imatch = nextMatch(line);
			    //this_.log("analyze(): \"" + tomatch + "\" inext " + inext + " imatch \"" + imatch + "\"" );			
			    while (inext >= 0 && imatch) {
				var ilength = imatch.length;
				var datum = {
				    start: ptr + inext,
				    end : ptr + inext + ilength,
				    success: true,
				    value:parse.parse,
				    type:"date"
				};		    
				results.push(datum);
				inext = findNext(line, inext + ilength);
				imatch = nextMatch(line, inext + ilength);
				//this_.log("analyze(): " + line.substring(indext+ilength) + " inext " + inext + "  imatch " + imatch + " " + ilength);
			    }
			} catch(e3) {
			    this_.log(e3 + " " + e3.lineNumber + " " + e3.fileName);
			}
		    });
		    ptr += line.length + 1;
		}
	    } catch(e) {
		this.log(e + " " + e.lineNumber + " " + e.fileName);
	    }
	    this.log("Analyze(); returning # of parses " + results.length);
	    return results;
	},
	test:function(start,end,s)  {
	    var p = Date.parse(s);
	    if (p) {
		return { success: p !== null && p !== undefined,
			 start: start, end: end,
			 value: {type:"date", exact:true, timestamp:p.valueOf()}, //(p !== null && p !== undefined ? JV3.plumutil.toJSON(p) : undefined),
			 type: "date" };
	    } 
	    return false;
	}
    };
})

	
	
/*	lines.map(function(line) {
	    try {
		JV3.Analysis.PlumDate.parse(line).map(function(parse) {
		    var text = parse.text;
		    if (this_.util.isArrayLike(text)) { text = new RegExp(text.join("\\s*"),"g");  }			
		    successes[text] = parse.parse;
		});
	    } catch(e) { this_.log(e + " " + e.lineNumber + " " + e.fileName); }
	});
	util.objKeys(successes).map(function(hit) {
	    var i = string.indexOf(hit);
	    var hit_ctr = 0;
	    while (i >= 0) {
		var datum = {
		    start: i,
		    end : i + hit.length,
		    success: true,
		    timestamp:successes[hit],
		    type:"date"
		};
		results.push(datum);
		i = string.indexOf(string,datum.end);
		hit_ctr++;
	    }

	    if (hit_ctr === 0) {
		this_.log("JV3.Analysis.engines.plumdate(): ERROR Could not find string \"" + hit + "\"" + typeof(hit));
	    }
	});
	this.log("JV3.Analysis.engines.plumdate.analyze() <<  hits : " + results.length);
*/
