
// Copyright (c) 2009 Massachusetts Institute of Technology

// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:

// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.

// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

(function(JV3,Analysis) {
    
    var plumutil = JV3.plumutil;

    var PlumDate = {
	log: function(x) { try { console.log(x); } catch(e) { } },
	months:{
	    type:"patterns",
	    patterns:[
		{pattern:'january', semantics:{type:"month", month:1}},
		{pattern:'jan', semantics:{type:"month", month:1}},
		{pattern:'february', semantics:{type:"month", month:2}},
		{pattern:'feb', semantics:{type:"month", month:2}},
		{pattern:'march', semantics:{type:"month", month:3}},
		{pattern:'mar', semantics:{type:"month", month:3}},
		{pattern:'april', semantics:{type:"month", month:4}},
		{pattern:'apr', semantics:{type:"month", month:4}},
		{pattern:'may', semantics:{type:"month", month:5}},
		{pattern:'june', semantics:{type:"month", month:6}},
		{pattern:'jun', semantics:{type:"month", month:6}},
		{pattern:'july', semantics:{type:"month", month:7}},
		{pattern:'jul', semantics:{type:"month", month:7}},
		{pattern:'august', semantics:{type:"month", month:8}},
		{pattern:'aug', semantics:{type:"month", month:8}},
		{pattern:'september', semantics:{type:"month", month:9}},
		{pattern:'sep', semantics:{type:"month", month:9}},
		{pattern:'october', semantics:{type:"month", month:10}},
		{pattern:'oct', semantics:{type:"month", month:10}},
		{pattern:'november', semantics:{type:"month", month:11}},
		{pattern:'nov', semantics:{type:"month", month:11}},
		{pattern:'december', semantics:{type:"month", month:12}},
		{pattern:'dec', semantics:{type:"month", month:12}}]
	},
	dow:{
	    type:"patterns",
	    patterns:[{pattern:'monday', semantics:{type:"dow", dow:1}},
		      {pattern:'mon', semantics:{type:"dow", dow:1}},
		      {pattern:'tuesday', semantics:{type:"dow", dow:2}},
		      {pattern:'tues', semantics: {type:"dow", dow:2}},
		      {pattern:'tue', semantics:{type:"dow", dow:2}},
		      {pattern:'wednesday', semantics:{type:"dow", dow:3}},
		      {pattern:'weds', semantics:{type:"dow", dow:3}},
		      {pattern:'wed', semantics:{type:"dow", dow:3}},
		      {pattern:'thursday', semantics:{type:"dow", dow:4}},
		      {pattern:'thurs', semantics:{type:"dow", dow:4}},
		      {pattern:'thur', semantics:{type:"dow", dow:4}},
		      {pattern:'friday', semantics:{type:"dow", dow:5}},
		      {pattern:'fri', semantics:{type:"dow", dow:5}},
		      {pattern:'saturday', semantics:{type:"dow", dow:6}},
		      {pattern:'sat', semantics:{type:"dow", dow:6}},
		      {pattern:'sunday', semantics:{type:"dow", dow:0}},
		      {pattern:'sun', semantics:{type:"dow", dow:0}}]},
	special_dow_modifiers: {
	    type: "patterns",
	    patterns:[ {pattern:"every",semantics:{repeat:"every"}},
		       {pattern:"next",semantics:{relative:"next"}}]
	},
	special_day: {
	    type:"patterns",
	    patterns:[ {pattern:"tomorrow", semantics:{value:"tomorrow"}},
		       {pattern:"today", semantics:{value:"today"}},
		       {pattern:"tonight", semantics:{value:"tonight"}}
		     ]
	}
    };
    PlumDate.date_patterns = [
	{name: "month-day", pattern:[PlumDate.months,/(^\d{1,2}$)/], semantics:function(x) {
	    return {month:x[0].month, day: parseInt(x[1].toString())  }; }},
	{name: "month-day-yr", pattern:[PlumDate.months,/(\d\d),?/,/(^\d\d\d\d$)/],
	 semantics:function(x) {
	     return {month:x[0].month, day: parseInt(x[1].toString()), year:parseInt(x[2].toString())}; }},
	{name: "month-day-year", pattern:[PlumDate.months,/(\d\d),?/,/(^\d\d$)/],
	 semantics:function(x) {
	     return {month:x[0].month, day: parseInt(x[1].toString()), year:parseInt(x[2].toString())}; }},
	{name: "mm/dd/yyyy", pattern: [/^(\d{1,2})\/(\d{1,2})\/(\d{4})$/],
	 semantics: function(x) {
	     x = x[0];
	     return {month:parseInt(x[1]), day:parseInt(x[2]), year:parseInt(x[3])};
	 }},    
	{name: "mm/dd/yy", pattern: [/^(\d{1,2})\/(\d{1,2})\/(\d{2})$/],
	 semantics: function(x) {
	     x = x[0];
	     var year = parseInt(x[3]);
	     if (year < 20) {  year += 2000; } else { year += 1900; }
	     return {month:parseInt(x[1]), day:parseInt(x[2]), year:year};
	 }},
	{name: "mm/dd", pattern: [/(\d{1,2})\/(\d{1,2})/],
	 semantics: function(x) {
	     x = x[0];
	     return {month:parseInt(x[1]), day:parseInt(x[2])};	 
	 }},
	{name: "relative dow", pattern: [PlumDate.special_dow_modifiers,PlumDate.dow], semantics:function(x) {
	    return { dow: x[1].dow, granularity:"day" };
	}},     
	{name: "special_day", pattern: [PlumDate.special_day], semantics:function(x) {
	    if (x[0].value == 'tomorrow') { return { type: "relative", day_offset:"1", granularity:"day" }; }
	    if (x[0].value == 'today') { return { type: "relative", day_offset:"0", granularity: "day" }; }
	    if (x[0].value == 'tonight') { return { type: "relative", day_offset:"0", granularity: "day", start_hour:5, end_hour:24 }; }
	}},
	{name: "dow", pattern: [PlumDate.dow], semantics:function(x) {
	    return x[0];
	}} // "thursday"
    ];

    PlumDate.time_patterns = [
	{name:"time-standard-base", pattern:[ /^(\d{1,2}):(\d{2})$/ , /(^a\.m\.|am|p\.m\.|pm)$/ ], semantics:
	 function(x) {
	     var hour=parseInt(x[0][1]);
	     var min=0;
	     if (x[0][1]) { min = parseInt(x[0][1]); }
	     if (x[1][1] && x[1][1]=='p.m.' || x[1][1]=='pm') { hour += 12; }
	     return { hour:hour, min:min };
	 }},    
	{name:"time-standard", pattern:[/^(\d{1,2})(:\d{2})(a\.m\.|am|p\.m\.|pm)?$/], semantics:
	 function(x) {
	     x = x[0];	 
	     var hour=parseInt(x[1]);
	     var min=0;
	     if (x[2] && !isNaN(parseInt(x[2].slice(1)))) {  min = parseInt(x[2].slice(1));  } // if (x[2] && x[2]=='p.m.' || x[2]=='pm') {    hour += 12; }
	     if (x[3] && x[3]=='p.m.' || x[3]=='pm') {    hour += 12; }
	     return { hour:hour, min:min };
	 }},
	{name:"at-time2", pattern:[/^(\@|at)(\d{1,2})(\:\d{1,2})?(a\.m\.|am|p\.m\.|pm)?/], semantics: // @3pm
	 function(x) {
	     PlumDate.log("at-time2");
	     PlumDate.log(x);
	     x = x[0];
	     var hour=parseInt(x[2]);
	     var min = 0;
	     if (x[3]) { min = parseInt(x[3].slice(1)); }
	     if (x[4] && x[4]=='p.m.' || x[4]=='pm') { hour += 12; }
	     return { hour:hour, min:min };
	 }},        
	{name:"time-vague", pattern:[/^(morning|noon|afternoon|evening)$/], semantics:
	 function(x) {
	     PlumDate.log(x);
	     if (x[0][0] === 'morning') {   return { type:"timerange", start_hour:6, end_hour:12 }; }
	     if (x[0][0] === 'noon') {   return { type:"timerange", start_hour:11, end_hour:13 }; }
	     if (x[0][0] === 'afternoon') {   return { type:"timerange", start_hour:12, end_hour:17 }; }
	     if (x[0][0] === 'evening') {   return { type:"timerange", start_hour:6, end_hour:12 }; }
	 }}
    ];

    PlumDate.combined_patterns = [
	{
	    name:"foo-bar",
	    pattern:[PlumDate.time_patterns[0], "on", PlumDate.date_patterns[0]],
	    semantics: function(x) {
		return { semantics: "foo bar " };
	    }
	}
    ];

    PlumDate.init = function() {
	this.log("plumdate.init---->");
	var this_ = this;
	var strict_start = false;
	this.util = plumutil;
	var trim = plumutil.trim;	
	var makeRecognizer = function(specs) {
	    // returns a serial recognizer that takes a series of accepting modules
	    // like < hello > < my > < name > < is > 
	    return {	    
		parse: function(s) {
		    var successes = [];
		    var h_p = function(string_left,string_so_far,parser_left,parse_so_far) {
			if (parser_left.length === 0) {
			    PlumDate.log("success parsing " + s);
			    successes.push({text:string_so_far,segments:parse_so_far});
			    return;
			} // success!h
			if (string_left.length === 0) { return undefined; } // fail!
			
			var car = string_left[0];
			var par = parser_left[0];

			if (this_.util.trim(string_left[0]).length == 0) {
			    // advance on
			    h_p(string_left.slice(1),
				string_so_far.length > 0 ? string_so_far+" "+string_left[0] : string_left[0],
				parser_left,
				parse_so_far );
			}
			
			var result = parser_left[0].parse(car);
			if (result) {
			    //PlumDate.log("WIN next.. " + par + " " + car);
			    h_p(string_left.slice(1),
				string_so_far.length > 0 ? string_so_far+" "+string_left[0] : string_left[0],
				parser_left.slice(1),
				par.semantics ? parse_so_far.concat([ (typeof(par.semantics) == 'function' ? par.semantics(result) : par.semantics) ]) : parse_so_far );
			    return;
			}
		    };

		    // now since we're less aggressive than the last version and are not tolerant
		    // of interwoven tokens we are going to have to scan down the tokens one at a time
		    // manually
		    
		    var tokenized = s.split(' ');

		    if (strict_start) {
			// front aligned
			h_p(tokenized, "", specs.concat(), []);
		    } else {
			// keep looking starting at every token till we find one
			while (successes == 0 && tokenized.length > 0) {
			    h_p(tokenized, "", specs.concat(), []);
  			    tokenized.shift();
			}
		    }
		    return successes;
		}	    
	    };
	};
	
	// compile a recognizer! -- takes 	
	var recs = [];
	var helper = function(remaining, recognizer) {
	    if (remaining.length === 0) {
		// done, add it to the pile
		return recs.push(makeRecognizer(recognizer));
	    }
	    var _m = function(pr) {
		return {
		    rec: pr,
		    parse: function(s) {
			if (typeof pr.pattern == 'string') {
			    return s == pr.pattern;
			}
			// regexp in simple pattern
			if (typeof pr.pattern == 'object' && pr.pattern.exec) {
			    return s.match(pr.pattern);
			}
			return false;
		    },
		    semantics: pr.semantics };
	    };
	    var _s = function(pr) {
		return { rec: pr,  parse: function(s) { return s == pr; } };
	    };
	    var car = remaining[0];
	    if (typeof car == 'object' && car.type == "patterns") {
		// list of "patterns"
		car.patterns.map(function(x) {
		    helper(remaining.slice(1), recognizer.concat(_m(x)));
		});		
	    } else {
		helper(remaining.slice(1),
		       recognizer.concat( _m( { pattern:car, semantics:function(x) { return x; }} ) ));
	    }
	};
	
	// compile a recognizers out of the patterns
	this.date_patterns = this.date_patterns.map(function(ds) {
	    recs = [];
	    helper(ds.pattern,[]);
	    return {name:ds.name, parsers:recs, semantics:ds.semantics};
	});
	this.time_patterns = this.time_patterns.map(function(ds) {
	    recs = [];
	    helper(ds.pattern,[]);
	    return {name:ds.name, parsers:recs, semantics:ds.semantics};
	});
	
    };

    PlumDate.parse=function(str) {
	var parses = [];
	str = this.util.trim(str.toLowerCase());

	// apply date patterns
	this.date_patterns.map(function(pattern) {
	    pattern.parsers.map(function(po) {
		var rs = po.parse(str);
		parses = parses.concat(rs.map(function(r) {
		    return {name: pattern.name, text: r.text, parse:pattern.semantics(r.segments)};
		}));
	    });	
	});

	var time_parses = [];
	this.time_patterns.map(function(pattern) {
	    pattern.parsers.map(function(po) {
		var rs = po.parse(str);
		time_parses = time_parses.concat(rs.map(function(r) {
		    return {name: pattern.name, text: r.text, parse:pattern.semantics(r.segments)};
		}));
	    });	
	});

	//PlumDate.log(parses);
	//PlumDate.log(time_parses);
	
	if (time_parses.length == 0) { return parses;    }
	if (parses.length == 0) {	return time_parses;    }

	JV3.log("MERGING DATE AND TIME");
	
	var this_ = this;

	// now we want to merge them.
	var new_parses = [];
	parses.map(function(dp) {
	    time_parses.map(function(tp) {
		var texthits = str.indexOf(dp.text) < str.indexOf(tp.text) ? [dp.text,tp.text]:[tp.text,dp.text];
		this_.log("analyze(): texthits " + texthits);
		new_parses.push( { name: [dp.name,tp.name],
				   text: texthits,
				   parse: plumutil.clone(dp.parse, plumutil.clone(tp.parse))
				 });
	    });
	});
	return new_parses;
    };


    PlumDate.instantiate_relative = function(pd, anchor){
	if (pd.parse.type == 'relative' && pd.parse.day_offset !== undefined && pd.parse.granularity == "day") {
	    var offsetdate = this.util.plusDays(anchor,pd.parse.day_offset);
	    this.log(offsetdate);
	    var clone = this.util.clone(pd);
	    this.util.clone( { day : offsetdate.getDate(),  month : offsetdate.getMonth(), year:offsetdate.getYear() }, clone.parse );
	    return clone;
	}
	return pd;    
    };

    PlumDate.toMinsOfDay = function (hr,min) {  return hr*60 + min;  };

    PlumDate.instantiate_partials = function(pd, anchor) { 
	// the idea is to fill in missing bits from dates that have pieces missing

	var parse = this.util.clone(pd.parse);
	var toMinsOfDay = this.toMinsOfDay;

	// our algorithm: find the soonest after anchor that fulfills the constraints
	// established by parse.
	anchor = new Date(anchor.valueOf());
	anchor.setMilliseconds(0);
	anchor.setSeconds(0);

	if ( parse.hour !== undefined && parse.min === undefined) {
	    parse.min = 0;
	}

	// dow specified, no date
	// instantiate {dow} -> { dow, date,  next-day-of-week
	if (parse.dow !== undefined && parse.month === undefined && parse.day === undefined) {
	    // watch out! modulo 6
	    if (parse.dow - anchor.getDay() >= 0) {
		parse.day = this.util.plusDays(anchor, parse.dow - anchor.getDay()).getDate();
		parse.month = this.util.plusDays(anchor, parse.dow - anchor.getDay()).getMonth();
		parse.year = this.util.plusDays(anchor, parse.dow - anchor.getDay()).getYear();
	    } else {
		parse.day = this.util.plusDays(anchor, 7 + parse.dow - anchor.getDay()).getDate();
		parse.month = this.util.plusDays(anchor, 7 + parse.dow - anchor.getDay()).getMonth();
		parse.year = this.util.plusDays(anchor, 7 + parse.dow - anchor.getDay()).getYear();
	    }
	}

	// day and month specified, no year
	if (parse.day !== undefined && parse.month !== undefined && parse.year === undefined) {
	    // canonicalize
	    var toDaysInYear = function(month,date) { return month*31 + date; }
	    var parsediy = toDaysInYear(parse.month,parse.day);
	    var anchoriy = toDaysInYear(anchor.getMonth(),anchor.getDate());
	    if (parsediy < anchoriy) {
		parse.year = anchor.getYear() + 1;
	    } else {
		parse.year = anchor.getYear();
	    }	
	}
	
	// time specified, no date
	if ( (parse.day === undefined || parse.month === undefined) && parse.hour !== undefined && parse.min !== undefined) {
	    //console.log("no date");
	    var parsemins = toMinsOfDay(parse.hour,parse.min);
	    //console.log("parsemins " + parsemins);
	    var anchormins = toMinsOfDay(anchor.getHours(),anchor.getMinutes());
	    //console.log("anchormins " + anchormins);
	    var plusdays = 0;
	    if (parsemins < anchormins) {  plusdays = 1; } // the time already passed today, so tomorrow!
	    var parsedate = this.util.plusDays(anchor, plusdays);
	    parse.day = parsedate.getDate();
	    parse.month = parsedate.getMonth();
	    parse.year = parsedate.getYear();
	    //console.log(parse);
	}
	// date specified, no time
	var toret = this.util.clone({ parse: parse },this.util.clone(pd));
	return toret;
    };

    PlumDate.compareTo = function(plumdate,d,relative_to) {
	if (isNaN(relative_to.valueOf())) { /* warning: invalid */ return 0;  }
	plumdate = this.instantiate_relative(plumdate,relative_to);
	plumdate = this.instantiate_partials(plumdate,relative_to);

	var parse = plumdate.parse;

	// parse should now have day/month/year
	this.util.assert(parse.day !== undefined && parse.month !== undefined, "Plumdate does not have day/month: " + plumdate.text);

	if ( parse.hour === undefined  ) {

	    // only days
	    var toDaysInYear = function(year,month,date) { return year*365+ month*31 + date; }
	    var diffdays =
		toDaysInYear(parse.year,parse.month,parse.day) - toDaysInYear(d.getYear(),d.getMonth(),d.getDate());
	    //console.log("diffdays: " + diffdays);
	    if (diffdays < 0) { return 0; }
	    if (diffdays == 0) { return 0.6; }
	    return 0.6 - Math.min(0.6,0.6*diffdays/3.0);
	}

	// hour and day specified, so we convert to msec and compare there
	//console.log("hour and day specified");
	//console.log(parse);

	var plumDate = new Date(relative_to.valueOf());
	plumDate.setMonth( parse.month );
	plumDate.setDate( parse.day );
	plumDate.setHours( parse.hour );
	plumDate.setMinutes( parse.min );
	plumDate.setSeconds( 0 );
	plumDate.setSeconds( 0 );

	var diffmsec = plumDate.valueOf() - d.valueOf();

	if (diffmsec < 0) { return 0;  }
	var N = 3;
	var daystail = N*24*60*60*1000*1.0;
	if (diffmsec > daystail) { return 0; }
	return 1 - Math.pow((diffmsec/daystail),0.5);    
    };


    /*
    
    if ( parse.month == d.getMonth() &&
	 parse.day == d.getDate() &&
	 parse.year == d.getYear() &&
	 parse.hour == d.getHours() &&
	 parse.min == d.getMinutes()  ) {
	this.log("exact minute/hour");
	// exact match!
	return 	1.0;
    } 
    if ( parse.month == d.getMonth() &&
	 parse.day == d.getDate() &&
	 parse.hour == d.getHours() &&
	 parse.min == d.getMinutes()  ) {
	// no year. we'll forgive them.
	this.log("exact minute/hour sans year");
	return 	1.0;
    }
    if ( parse.month == d.getMonth() && parse.day == d.getDate() && parse.day !== undefined && parse.hour !== undefined )  {
	// today, so scale linearly	
	this.log("same day with hour/min specified");
	var pdminofday = parse.hour*60 + parse.min;
	var dminofday = d.getHours()*60+d.getMinutes();
	if (pdminofday > dminofday) {
	    // scale to 0.5-1.0
	    return 1.0 - 0.4 * Math.pow(pdminofday-dminofday,2)/Math.pow(60.0*24.0,2);
	}
	// past our expiry, so don't show it.
	return 0.0;
    }
    if ( parse.month == d.getMonth() && parse.day == d.getDate() ) {
	this.log("same day ");
	return 0.6;
    }
    if ( parse.hour && parse.min ) {
	var pdhour = parse.hour * 60 + parse.min;
	var dhour = date.getHours() * 60 + d.getMinutes();
	if (pdhour < dhour) { return 0; }
	return 0.6 - 0.3 * (pdhour-dhour)/(60.0*24);
    }
    if ( parse.month && parse.day ) {
	var plumdayofyear = parse.month * 31 + parse.day;
	var ddayofyear = d.getMonth() * 31 + d.getDate();	
	if (plumdayofyear < ddayofyear) {
	    return 0;
	} else {
	    return 0.3-0.3*(plumdayofyear - ddayofyear)/365.0;
	}
    }
*/

    Analysis.PlumDate = PlumDate;
})