// Copyright (c) 2009,2008 Massachusetts Institute of Technology

// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:

// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.

// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

__defn_module(
    "context-maiden-eventstore",
    function(JV3,ContextMaiden) {
	ContextMaiden.mozGetType = ({
					"int": function(statement,x) { return statement.getInt64(x); },
					"double": function(statement,x) { return statement.getDouble(x); },
					"string": function(statement,x) { return statement.getUTF8String(x); }
				    });
	
	var getParameterIndex = function(statement,name) {
	    try { return statement.getParameterIndex(name);   } catch(e) { }
	    try { return statement.getParameterIndex(":"+name); } catch(e) {
		JV3.log("getParameterIndex FAIL.");
	    }
	};
	
	ContextMaiden.CMStore = {
	    superclass:JV3.PlumStore,
	    schemas:JV3.schemas,
	    initialize:function(params) {
		JV3.PROTOCROCK.superclass(this,"initialize", params);
		this.store_listeners = {};	
		this.log = function() { JV3.log.apply(JV3,arguments); };
		this.log("CMStore init " + this.plumutil.toJSON(params));
		// this.vacuum();
	    },
	    forStatementsMatching: function(fields, fn) {
		// fields should be an object like { context_type : value, context_type : value, .... }
		var select_keys = this.schemacols;
		var select_clause = this._makeSelectCols(select_keys);
		if (select_clause.length === 0)  { select_clause = " * "; }
		var _temp = this._makeWhere(fields, "=", " AND " );
		var where_selector = _temp[0], bindings = _temp[1];
		var query = "SELECT " + select_clause + " FROM " + this.table + (where_selector.length > 0 ? " WHERE " + where_selector : "") + ";";
		var db = this._db;
		// JV3.log("forstatementsmatching " + query);
		var statement = db.createStatement(query);
		bindings.map(function(f) { f(statement); });
		try {
		    while(statement.executeStep()) {    fn(select_keys, statement);		}
		} catch (e) {
		    this.log("forStatementsMatching error " + e);
		} finally {
		    statement.reset();
		}
	    },
	    vacuum:function() { try {  this.execute("VACUUM;"); } catch (x) { JV3.log(x); } },
	    _execute_async_statement:function(statement,callback,fn) {
		var this_ = this;
		if (fn === undefined) { fn = function(x) { return x; }; };
		try {
		    var results = [];
		    if (statement.executeAsync) {
			// ASYNCHRONOUS new skool for 3.5.x and newer ->
			var unpack_row = function(r) {
			    var o = {};			    
			    this_.schemacols.map(function(k) { o[k] = r.getResultByName(k); });
			    return fn(o);
			};	    			
			statement.executeAsync(
			    {
				handleResult:function(resultSet) {
				    try {
					for (var row = resultSet.getNextRow(); row; row = resultSet.getNextRow()) {
					    results.push(unpack_row(row));
					}
				    } catch (x) {  JV3.log(x); }
				},
				handleError:function(err) {
				    JV3.log(err);
				    if (callback !== undefined) {callback({error:err});}

				},
				handleCompletion:function(aReason) {
				    if (aReason != Components.interfaces.mozIStorageStatementCallback.REASON_FINISHED) {
					return callback({error:"query canceled or aborted"});				
				    }
				    if (callback !== undefined) { callback(results); };
				    try { statement.reset(); } catch (x) {JV3.log(x); }				    
				    return undefined;
				}
			    });			
		    } else {
			// SYNCHRONOUS old skool for 3.0.x
			var unpack_row = function(statement) {
			    var result = {};
			    this_.schemacols.map(function(key) {    result[key] = this_.column_grabbers[key](statement, this_.schemacols.indexOf(key)); });
			    return fn(result);
			};	    
			// synchronously
			while(statement.executeStep()) {  results.push( unpack_row(statement) ); }
			if (callback !== undefined) { callback(results); }
		    }
		} catch (e) {
		    this._dberror("execute_async", e, statement);
		    try { callback({error:e}); } catch (x) { JV3.log("FATAL. EXCEPTION ON EXCEPTION >>> ", x);  }
		} 
	    },	
	    forStatementsMatchingAsync: function(fields, fn, callback) {
		// fields should be an object like { context_type : value, context_type : value, .... }
		var select_keys = this.schemacols;
		var select_clause = this._makeSelectCols(select_keys);
		if (select_clause.length === 0)  { select_clause = " * "; }
		var _temp = this._makeWhere(fields, "=", " AND " );
		var where_selector = _temp[0], bindings = _temp[1];
		var query = "SELECT " + select_clause + " FROM " + this.table + (where_selector.length > 0 ? " WHERE " + where_selector : "") + ";";
		var db = this._db;
		// JV3.log("async query: " + query);
		var statement = db.createStatement(query);	    
		bindings.map(function(f) { f(statement); });
		this._execute_async_statement(statement,callback,fn);
	    },	
	    countEntitiesMatching:function(fields) {
		var count = 0;
		this.forStatementsMatching(fields, function() { count = count+1; });
		return count;
	    },
	    countEntitiesMatchingAsync:function(fields,callback) {
		var count = 0;
		this.forStatementsMatchingAsync(fields,
						function() { count = count+1; },
						function() { callback(count); }); // continuation
	    },	
	    insert:function(obj) {
 		var _temp = this._makeInsertValues(this.schemacols, obj);
		var bindings = _temp[0],fns = _temp[1];
		var stmt_sql = "INSERT INTO " + this.table + " VALUES(" + bindings + ");";
		var statement = this._db.createStatement(stmt_sql);
		fns.map( function(f) { f(statement); } );	
		try {   statement.executeStep(); } catch(e)  {
		    this._dberror("INSERT",e,statement,bindings);
		} finally {
		    statement.reset();
		}	    
	    },
	    insertAsync:function(obj,cont) {
		var _temp = this._makeInsertValues(this.schemacols, obj);
		var bindings = _temp[0],fns = _temp[1];
		var stmt_sql = "INSERT INTO " + this.table + " VALUES(" + bindings + ");";
		var statement = this._db.createStatement(stmt_sql);
		fns.map( function(f) { f(statement); } );
		this._execute_async_statement(statement,cont);
	    },
	    _dberror:function(type,err,statement,bindings) {
		JV3.log(type + "::ERROR____ ");
		JV3.log(err);
		JV3.log("statement: ");
		JV3.log(statement);
		JV3.log(JV3.plumutil.toJSON(bindings));
	    },
	    remove:function(select_fields) {
		var _select_temp = this._makeWhere(select_fields, "=", " AND " );
		var where_selector = _select_temp[0], where_fns = _select_temp[1];
		var stmt_sql = "DELETE FROM  " + this.table +" WHERE " + where_selector + ";";
		var statement = this._db.createStatement(stmt_sql);
		// JV3.log("remove:"+stmt_sql);
		where_fns.map( function(f) { f(statement); } );
		try {
		    //this.log("Executing " + statement);
		    return statement.executeStep();
		} catch(e)  {
		    this._dberror("REMOVE",e,statement,bindings);
		} finally {
		    statement.reset();
		}
	    },	
	    removeAsync:function(select_fields,cont) {
		var _select_temp = this._makeWhere(select_fields, "=", " AND " );
		var where_selector = _select_temp[0], where_fns = _select_temp[1];
		var stmt_sql = "DELETE FROM  " + this.table +" WHERE " + where_selector + ";";
		var statement = this._db.createStatement(stmt_sql);
		// JV3.log("removeasync: " + stmt_sql);
		where_fns.map( function(f) { f(statement); } );
		this._execute_async_statement(statement,cont);
	    },	
	    update:function(select_fields,dehydrated){
		var _select_temp = this._makeWhere(select_fields, "=", " AND " );
		var where_selector = _select_temp[0], where_fns = _select_temp[1];
		var _set_temp = this._makeWhere(dehydrated,"=",", ");
		var set_selector = _set_temp[0], set_fns = _set_temp[1];
		
		var stmt_sql = "UPDATE  " + this.table + " SET " + set_selector  + " WHERE " + where_selector + ";";
		// JV3.log(">>>>>>UPDATE ASYNC");
		//JV3.log("update_async:" + stmt_sql);
		// JV3.log(dehydrated);
		// JV3.log("|||||||||");
		var statement = this._db.createStatement(stmt_sql);
		set_fns.concat(where_fns).map( function(f) { f(statement); } );
		try {
		    //this.log("Executing " + statement);
		    statement.executeStep();
		} catch(e)  {
		    this._dberror("UPDATE",e,statement,bindings);
		} finally {
		    statement.reset();
		}
		return "update";	
	    },
	    updateAsync:function(select_fields,dehydrated,cont){
		var _select_temp = this._makeWhere(select_fields, "=", " AND " );
		var where_selector = _select_temp[0], where_fns = _select_temp[1];
		var _set_temp = this._makeWhere(dehydrated,"=",", ");
		var set_selector = _set_temp[0], set_fns = _set_temp[1];
		
		var stmt_sql = "UPDATE  " + this.table + " SET " + set_selector  + " WHERE " + where_selector + ";";
		//JV3.log(stmt_sql);
		//JV3.log(JV3.plumutil.toJSON(dehydrated));	    
		var statement = this._db.createStatement(stmt_sql);
		set_fns.concat(where_fns).map( function(f) { f(statement); } );
		this._execute_async_statement(statement,cont);
	    },	
	    // -------------------------------------------------------------------------------------
	    saveCommon:function(select_fields,dehydrated) {
		if (this.countEntitiesMatching(select_fields) == 0) {
		    return this.insert(dehydrated);
		} 
		// update!
		return this.update(select_fields,dehydrated);
	    },
	    saveCommonAsync:function(select_fields,dehydrated,cont) {
		var this_ = this;
		this.countEntitiesMatchingAsync(select_fields,function(c) {
						    if (c == 0) {
							return this_.insertAsync(dehydrated,cont);	    
						    } else {
							return this_.updateAsync(select_fields,dehydrated, cont);   
						    }
						});
	    },	
	    addStoreUpdateListener:function(type,l) {
		if (type === undefined) { type = "__ALL__"; }
		if (this.store_listeners[type]) { this.store_listeners[type].push(l); return; }
		this.store_listeners[type] = [l];
		return;
	    },
	    _createTable:function() {
		var this_ = this;
		this.execute("CREATE TABLE IF NOT EXISTS " + this.table + this.schema + ";"); /* create table */
		if (this.index) {
		    this.index.map(function(colname) {
				       try {
					   this_.execute("CREATE INDEX IF NOT EXISTS " + "index_"+colname + " ON " +  this_.table + " ( " + colname + " ); "); // 
				       } catch (x) {  JV3.log(x);      }				   
				   });
		}
	    },
	    _makeSelectCols:function(cols) { return cols.join(",");    },    
	    _fireUpdate:function(item) {
		var this_ = this;
		var rel_listeners = [];
		if (this.store_listeners[item.type]) { rel_listeners = this.store_listeners[item.type]; }
		if (this.store_listeners["__ALL__"]) { rel_listeners = rel_listeners.concat(this.store_listeners["__ALL__"]); }
		// JV3.log("_CMStore fireUpdate ", JV3.plumutil.objKeys(item), item.id, item.type, "listeners: ", rel_listeners.length);
		rel_listeners.map(function(x) {
				      if (typeof(x) === 'object') {
					  try {
					      x.itemUpdated(JV3.plumutil.clone(item));
					  } catch(e) {
					      this_.log(e + " " + e.lineNumber + " " + e.fileName);
					  }		    
				      } else if (typeof(x) === 'function') {
					  x(item);
				      }
				  });
	    },
	    _makeSet : function(obj, op, conjunctor) {
		var s = '';
		for (var field in obj) {
		    s += field + op + (typeof(obj[field]) == 'string' ? "\""+obj[field]+"\"" : obj[field]) + conjunctor;
		}
		return s.slice(0,s.length-conjunctor.length);
	    },
	    _make_statement_parameter_binder:function(key,val) {
		return function(statement) {
		    var index = getParameterIndex(statement,key);
		    if (typeof val == "string") {  return statement.bindUTF8StringParameter(index, val);   }
		    if (typeof val == "number") {  return statement.bindInt64Parameter(index, val);   }
		    if (typeof val == "object" && JV3.plumutil.isDateLike(val)) {  return statement.bindInt64Parameter(index, val.valueOf());   }					      
		    if (typeof val == "object" && JV3.plumutil.isArrayLike(val)) {  return statement.bindUTF8StringParameter(index, plumutil.toJSON(val)); }
		    // what happened to boolean
		    if (val === undefined || val === null) { return statement.bindNullParameter(index);  }
		    JV3.plumutil.assert(false, "contextMaidenInsertVals: Error figuring out type of val to insert for key " + key + " val " + val + " " + typeof(val));
		};
	    },
	    _makeInsertValues:function(cols, values) {
		// ORDER is crucial
		var this_ = this;
		var fns = [];
		var plumutil = this.plumutil;
		var clause = cols.map(function(key) {
					  var val = values[key];
					  fns.push( this_._make_statement_parameter_binder(key,val) );
					  return ":"+key;
				      }).join(",");
		return [clause,fns];
	    },    
	    _makeWhere : function(obj, ops, conjunctor) {
		// return "k=\"obj[k]\" AND ... " for all keys in obj
		var this_ = this;
		with(JV3.plumutil) {
		    var getOp = function(key) {
			if (key && typeof(ops) == 'object') { return ops[key]; }
			return ops;
		    };
		    var fns = [];
		    var clause = objKeys(obj).map(function(key) {
						      var val = obj[key];
						      fns.push( this_._make_statement_parameter_binder(key,val) );
						      return key+""+getOp(key)+":"+key;  // covers cases like: foo=:foo AND bar=:bar --or-- foo:foo, bar=:bar 
						  }).join(conjunctor);
		    return [clause,fns];
		}
	    },    
	    // more modern version --------
	    _makeStatement: function(cols,table,clauses) {
		var make_where_c_fn = this._makeWhereArray(clauses);
		var stmt_sql = "SELECT " + this._makeSelectCols(cols) + " FROM " + table + " WHERE " + make_where_c_fn[0];
		var statement = this._db.createStatement(stmt_sql);
		make_where_c_fn[1].map( function(f) { f(statement); } );
		return statement;
	    },
	    _makeWhereArray : function(clauses, conjunctor) {
		// uses an array structure for clauses to support repeats 
		// return "k=\"obj[k]\" AND ... " for all keys in obj
		var this_ = this;
		var conjunctor = conjunctor === undefined ? "AND" : conjunctor;
		with(JV3.plumutil) {
		    var fns = [];
		    var clause = clauses.map(function(clause) {
						 var key = guid();
						 var val = clause.val;
						 fns.push( this_._make_statement_parameter_binder(key,val) );						 
						 return clause.col+""+clause.op + ":"+key;  // covers cases like: foo=:foo AND bar=:bar --or-- foo:foo, bar=:bar 
					     }).join(" " + conjunctor + " " );
		    return [clause,fns];
		}	
	    },    
	    _forResults:  function(fields, fn) {
		// fields should be an object like { context_type : value, context_type : value, .... }
		var this_ = this;
		var select_keys = this.schemacols;
		var select_clause = this._makeSelectCols(select_keys);
		if (select_clause.length === 0)  { select_clause = " * "; }
		var _temp = this._makeWhere(fields, "=", " AND " );
		var where_selector = _temp[0], bindings = _temp[1];
		var query = "SELECT " + select_clause + " FROM " + this.table + (where_selector.length > 0 ? " WHERE " + where_selector : "") + ";";
		var db = this._db;
		// JV3.log("forResults: " + query);
		var statement = db.createStatement(query);
		bindings.map(function(f) { f(statement); });
		try {
		    while(statement.executeStep()) {
			var result = {};
			select_keys.map(function(key) {
					    result[key] = this_.column_grabbers[key](statement, select_keys.indexOf(key));
					});
			try { fn(result); } catch(e) { JV3.log(e + " " + e.lineNumber + " " + e.fileName); }
		    }
		} catch (e) {
		    this.log("forStatementsMatching error " + e);
		} finally {
		    statement.reset();
		}
	    },
	    _forResultsArray:  function(clauses, conjunctor, fn) {
		// fields should be an object like [ {field:"foo", op:"=", val="12903"} ... ]
		var this_ = this;
		var select_keys = this.schemacols;	    
		var select_clause = this._makeSelectCols(select_keys);
		if (select_clause.length === 0)  { select_clause = " * "; }
		var _temp = this._makeWhereArray(clauses,conjunctor);
		var where_selector = _temp[0], bindings = _temp[1];
		var query = "SELECT " + select_clause + " FROM " + this.table + (where_selector.length > 0 ? " WHERE " + where_selector : "") + ";";
		var db = this._db;
		// JV3.log("_forResultsarray: " + query);
		var statement = db.createStatement(query);
		bindings.map(function(f) { f(statement); });
		try {
		    while(statement.executeStep()) {
			var result = {};
			select_keys.map(function(key) {  result[key] = this_.column_grabbers[key](statement, select_keys.indexOf(key)); });
			try { fn(result); } catch(e) { JV3.log(e); }
		    }
		} catch (e) {
		    JV3.log(e);
		} finally {
		    statement.reset();
		}
	    },
	    _forResultsArrayAsync:  function(clauses, conjunctor, fn, cont) {
		// fields should be an object like [ {field:"foo", op:"=", val="12903"} ... ]
		var this_ = this;
		var select_keys = this.schemacols;	    
		var select_clause = this._makeSelectCols(select_keys);
		if (select_clause.length === 0)  { select_clause = " * "; }
		var _temp = this._makeWhereArray(clauses,conjunctor);
		var where_selector = _temp[0], bindings = _temp[1];
		var query = "SELECT " + select_clause + " FROM " + this.table + (where_selector.length > 0 ? " WHERE " + where_selector : "") + ";";
		var db = this._db;
		// JV3.log("forResultsArrayAsync : " + query);
		var statement = db.createStatement(query);
		bindings.map(function(f) { f(statement); });
		this._execute_async_statement(statement,cont,fn);
	    }	
	};

	/** these guys were lifted from **/

	
	ContextMaiden.EventStore = {
	    superclass: ContextMaiden.CMStore,    
	    schema: "(type TEXT, start NUMERIC, end NUMERIC, entityid TEXT, entitytype TEXT, data TEXT)",
	    schemacols: ["type", "start", "end", "entityid", "entitytype", "data"],
	    index:["type","start","end","entityid"],	
	    column_grabbers: {
		type : ContextMaiden.mozGetType["string"],
		start : ContextMaiden.mozGetType["double"],
		end : ContextMaiden.mozGetType["double"],
		entityid : ContextMaiden.mozGetType["string"],
		entitytype : ContextMaiden.mozGetType["string"],
		data : ContextMaiden.mozGetType["string"]
	    },
	    index : ["start", "end"],
	    instant_smooth_window_msec: 3*60000, // 1 minute
	    initialize: function(JV3,params, entity_store) {
		JV3.PROTOCROCK.superclass(this,"initialize", params ? params : {});
		this.table = (params === undefined || params.tablename === undefined) ? "event" : params.tablename;
		this._createTable();
		this.entity_store = entity_store; // used in reconstituting dudes
		this.util = JV3.plumutil;
	    },
	    _makeTemporalWhereClause:function(timespec_or_whereclause) {
		var s = "";
		var plumutil = this.plumutil;
		//this.log("timespec or whereclause " + timespec_or_whereclause);
		if (typeof(timespec_or_whereclause) == "string") {
		    // ALREADY a where clasue
		    return timespec_or_whereclause;
		}
		
		var timespec = timespec_or_whereclause;
		
		if (timespec_or_whereclause.type == "TimeStruct") {
		    //this.log("timestruct start:" + timespec_or_whereclause.asDuration().start + " " + timespec_or_whereclause.asDuration().end);
		    var dur = timespec_or_whereclause.asDuration();
		    return this._makeTemporalWhereClause([dur.start,dur.end]);
		}	
		if (timespec.length >= 1 && typeof(timespec[0]) == 'number') {
		    s = "start > " + timespec[0];
		    if (timespec.length == 2) {	s = s + " AND end <=" + timespec[1];   }
		    return s;		
		} else if (timespec.length >= 1 && plumutil.isDateLike(timespec[0])) {
		    s = "start > " + timespec[0].valueOf();
		    if (timespec.length == 2) { s = s + " AND end <=" + timespec[1].valueOf();    }
		    return s;		
		} else if (timespec.type === 'datetime') {
		    // instant
		    var timepair = timespec.getEffectiveTime().valueOf();
		    if (timepair.length && timepair[0] && timepair[1]) {
			// then we've been given a range back
			return "start > " + timepair[0].valueOf() + " AND end <=" + timepair[1].valueOf();
		    }
		    if (timepair.length && timepair[0]) {
			var window = this._smoothInstant(timepair[0]);
			return "start > " + window[0].valueOf() + " AND end <= "+ window[1].valueOf();
		    }
		}
		return "";
	    },
	    _makeTemporalIntersectionWhereClause:function(timespec) {
		return "start <= " + timespec[1].valueOf() + " and end >= " + timespec[0].valueOf();
	    },	    	    	    
	    _smoothInstant:function(jsdate) {
		return [ new Date(jsdate.valueOf() - this.instant_smooth_window_msec), new Date(jsdate.valueOf() + this.instant_smooth_window_msec) ];
	    },
	    // THESE FOUR DO NOT RECONSTITUTE ENTITY.
	    getEventsRaw:function(event_or_template) {
		var this_ = this;
		var dehydrated = event_or_template && JV3.PROTOCROCK.isinstanceof(event_or_template,this.schemas.Event) ? this._deconstitute(event_or_template) : event_or_template;
		var results = [];
		this.forStatementsMatching( dehydrated, function(select_keys,statement) {
						var result = {};
						select_keys.map(function(key) {
								    result[key] = this_.column_grabbers[key](statement, select_keys.indexOf(key) );
								});
						results.push(result);
					    });
		return results.map(function(x) { return this_._reconstitute(x); } );
	    },
	    getEventsAsync:function(event_or_template,cont) {
		var this_ = this;
		this._getEventsRawAsync(event_or_template, function(result) { return this_._reconstitute(result);  }, cont);
	    },	    
	    getEventsRawAsync:function(event_or_template,cont) {
		this._getEventsRawAsync(event_or_template, function(result) { return result; }, cont);
	    },
	    _getEventsRawAsync:function(event_or_template,fn,cont) {
		var this_ = this;
		var dehydrated = event_or_template && JV3.PROTOCROCK.isinstanceof(event_or_template,this.schemas.Event) ? this._deconstitute(event_or_template) : event_or_template;
		var results = [];
		this.forStatementsMatchingAsync( dehydrated, fn, cont);
	    },	    
	    getEventsRawTime:function(type,timespec) {
		var results = [];
		var typeclause = type === undefined ? [] : [{col:"type", op:"=", val:type}];
		this._forResultsArray(typeclause.concat([{col:"start", op:">=", val:timespec[0]},   {col:"end", op:"<=", val:timespec[1]}]),"AND", function(r) { results.push(r);});
		return results;
	    },
	    getEventsRawTimeAsync:function(type,timespec,cont) {
		var results = [];
		var typeclause = type === undefined ? [] : [{col:"type", op:"=", val:type}];
		var clauses = typeclause.concat([{col:"start", op:">=", val:timespec[0]},{col:"end", op:"<=", val:timespec[1]}]);
		var _temp = this._makeWhereArray(clauses,"AND");
		var where_selector = _temp[0], bindings = _temp[1];
		var query = "SELECT * FROM " + this.table + (where_selector.length > 0 ? " WHERE " + where_selector : "") + ";";
		//JV3.log(query);
		var db = this._db;
		var statement = db.createStatement(query);
		bindings.map(function(f) { f(statement); });
		this._execute_async_statement(statement,cont);		
	    },
	    
	    getEventTypes:function() {
		return this.getStringResults("select distinct type from " + this.table + ";");
	    },
	    getDistinctEntityIds:function(evttype) {
		return this.getStringResults("select distinct entityid from " + this.table + (evttype ? " where type=\""+evttype+"\"" : "") + ";");
	    },
	    getEntityTypesforEventType:function(evttype) {
		return this.getStringResults("select distinct entitytype from " + this.table + (evttype ? " where type=\""+evttype+"\"" : "") + ";");
	    },	    
	    getEvents:function(type,timespec) {
		var where_type = type ? "type == \""+type+"\" " : undefined;
		var where_time = timespec ? this._makeTemporalWhereClause(timespec) : undefined;
		var where_selector = where_type || where_time ? " WHERE " + [where_type, where_time].filter(function(x) { return x; }).join("AND") : "";
		var keys = this._makeSelectCols(this.schemacols);
 		var cols = this.schemacols;	
		var query = "SELECT " + keys + " FROM " + this.table + where_selector + " ORDER BY start;";
 		// JV3.log(query);
		var db = this._db;	
		var statement = db.createStatement(query);
		var results = [];
		try {
		    while(statement.executeStep()) {
			var result = {};
			for (var c_i = 0 ; c_i < cols.length; c_i++) {
			    var prop = cols[c_i];
			    result[prop] = this.column_grabbers[prop](statement,c_i);
			}
			results.push(this._reconstitute(result));
		    }
		} catch (e) {
		    //self.ERROR = e;
 		    this.log("getStringResults() error in statement " + statement + " : " + e); //alert(e);
		} finally {
		    statement.reset();
		}
		return results;
	    },

	    // this does soemthign bad right now, which is to make a blocking call to getEntities at every fn(). 
	    _get_events_async:function(where_time,type,timespec,cont) {
	    	var where_type = type ? "type=\""+type+"\"" : undefined;
	    	var where_selector = where_type || where_time ? " WHERE " + [where_type, where_time].filter(function(x) { return x; }).join(" AND ") : "";
	    	var keys = this._makeSelectCols(this.schemacols);
	    	var cols = this.schemacols;
	    	var query = "SELECT " + keys + " FROM " + this.table + where_selector + " ORDER BY start;";		
	    	// JV3.log(query);
	    	var this_ = this;
	    	var statement = this._db.createStatement(query);

		// this goes out to brenn >> 
		this._execute_async_statement(statement,
					      function(drieds) {
						  (function(todo,done) {
						       try {
							   var _this_fn = arguments.callee;
							   if (todo.length == 0) { return cont(done); }
							   this_._reconstituteAsync( todo[0], function(bar) {
											 try {
											     _this_fn(todo.slice(1),done.concat(bar));
											 } catch (x) {    JV3.log(x);	 }
										     });
						       } catch (x) {  JV3.log(x);       }
						   })(drieds,[]);
					      });
		
	    },
	    getEventsAsyncTime:function(type,timespec,cont) {
	    	var where_time = timespec ? this._makeTemporalWhereClause(timespec) : undefined;
	    	this._get_events_async(where_time,type,timespec,cont);
	    },
	    getEventsIntersectionAsync:function(type,timespec,cont) {
	    	var where_time = timespec ? this._makeTemporalIntersectionWhereClause(timespec) : undefined;
	    	this._get_events_async(where_time,type,timespec,cont);		
	    },
	    getMostRecentEventAsync:function(type,timespec,cont) {
		var where_type = type ? "type = \""+type+"\" " : undefined;
		var where_time = timespec ? this._makeTemporalWhereClause(timespec) : undefined;
		var where_selector = where_type || where_time ? " WHERE " + [where_type, where_time].filter(function(x) { return x; }).join("AND") : "";		
		var keys = this._makeSelectCols(this.schemacols);
		var cols = this.schemacols;
		var query = "SELECT " + keys + " FROM " + this.table + where_selector + " ORDER BY START DESC LIMIT 1;";		
		var this_ = this;
		// JV3.log(query);
		var statement = this._db.createStatement(query);
		this._execute_async_statement(statement, cont, function(x) { return this_._reconstitute(x);});
	    },	    
	    _reconstitute:function(serial) {
		var plumutil = this.plumutil;	
		var rehydrated = this.plumutil.clone(serial);
		
		if (rehydrated.entityid && rehydrated.entitytype) {
		    //JV3.log("!!!!!!!!!!!!!! rehydrating : " + rehydrated.entityid + " " + rehydrated.entitytype);
		    var rehyd_ent = undefined; 
		    try {
			// BAD! we're using a blocking call :(
			rehyd_ent = this.entity_store.getEntities({id:rehydrated.entityid, type:rehydrated.entitytype});
		    } catch(e) {
			JV3.log(e);
		    }
		    if (rehyd_ent !== undefined && rehyd_ent.length > 0) {
			rehydrated.entity = rehyd_ent[0];
		    } else {
			JV3.log("Warning could not restore entity belonging to event " +  rehydrated.type + " [[  "  + rehydrated.entityid + " " + rehydrated.entitytype + "]]");

			// we want to keep this around just in case someone else wants to restore the real dudes
			rehydrated._entityid = rehydrated.entityid;
			rehydrated._entitytype = rehydrated.entitytype;			
		    }
		    delete rehydrated.entityid;
		    delete rehydrated.entitytype;
		}
		rehydrated.start = parseInt(rehydrated.start); 
		rehydrated.end = parseInt(rehydrated.end);
		if (rehydrated.data && rehydrated.data !== 'undefined') {
		    rehydrated.data = plumutil.fromJSONAggressive(rehydrated.data);
		}
		return rehydrated;
	    },
	    _reconstituteAsync:function(serial,cont) {
		var plumutil = this.plumutil;	
		var rehydrated = this.plumutil.clone(serial);		
		if (rehydrated.entityid && rehydrated.entitytype) {
		    //JV3.log("!!!!!!!!!!!!!! rehydrating : " + rehydrated.entityid + " " + rehydrated.entitytype);
		    rehyd_ent = this.entity_store.getEntitiesAsync({id:rehydrated.entityid, type:rehydrated.entitytype},
								   undefined,
								   function(rehyd_ent) {
								       if (rehyd_ent !== undefined && rehyd_ent.length > 0) {
									   rehydrated.entity = rehyd_ent[0];
								       } else {
									   JV3.log("Warning could not restore entity belonging to event " +  rehydrated.type + " [[  "  + rehydrated.entityid + " " + rehydrated.entitytype + "]]");
								       }
								       delete rehydrated.entityid;
								       delete rehydrated.entitytype;
								       rehydrated.start = parseInt(rehydrated.start); 
								       rehydrated.end = parseInt(rehydrated.end);
								       if (rehydrated.data && rehydrated.data !== 'undefined') {
									   rehydrated.data = plumutil.fromJSONAggressive(rehydrated.data);
								       }
								       cont(rehydrated);								  
								   });
		}
	    },	    
	    _deconstitute:function(value) {
		var this_ = this;
		var plumutil = this.plumutil;
		var val = plumutil.clone(value);
		if (val.entityid && val.entitytype) {
		    // we are okay as-is!!	    
		} else if (val.entity) {
		    
		    // this is commented out because for atomate functionalty
		    // we can't be savin' entities attached to events all the time!
		    // (that completely messes up the rule trigger mechanism.)
		    
		    // if (this.entity_store) { 
		    // 	// not sure this is a good thing but we will try to save the attached entity
		    // 	// just in case!
		    // 	//
		    // 	// try {
		    // 	//     this.entity_store.saveAsync(val.entity);
		    // 	// } catch(e) {
		    // 	//     /* entity already exists then we oky */
		    // 	//     this_.log(e + " " + e.lineNumber + " " + e.fileName);
		    // 	// }
		    // }
		    
		    val.entitytype = val.entity.type;
		    val.entityid = val.entity.id;
		    delete val.entity;
		} else {	    
		    val.entityid = 'undefined';  val.entitytype = 'undefined';
		}
		
		if (val.start.type == 'schemas.DateTime') {  val.start = val.start.getDate().valueOf();  /* can only be absolute */}
		if (val.end.type == 'schemas.DateTime') {  val.end = val.end.getDate().valueOf();  /* can only be absolute */}
		if (plumutil.isDateLike(val.start)) { val.start = val.start.valueOf(); }
		if (plumutil.isDateLike(val.end)) { val.end = val.end.valueOf(); }
		if (val.data) { val.data = plumutil.toJSON(val.data); }
		return val;
	    },
	    save:function(event) {
		//this.log("saving event : " + event);
		var this_ = this;
		var dehydrated = this._deconstitute(event);
		// we want to omit end from the select_clauses so that we can easily overwrite previous observations
		// starting at the same point, e.g., extend previous observations

		var last = this.getEventsRaw({type : event.type, entitytype: dehydrated.entitytype, entityid: dehydrated.entityid, start : dehydrated.start});
		//this.log(">> savecommon : " + event);
		this.saveCommon( {type : event.type, entitytype: dehydrated.entitytype, entityid: dehydrated.entityid, start : dehydrated.start /*, end : dehydrated.end */ }, dehydrated );
		//this.log("<< savecommon : " + event);
		
		if (last.length === 0) {
		    // this.log("event! new! " + event.type + " " + (event.entity ? event.entity.id : event.entityid) + " " + event.start + " " );
		    this._fireUpdate(this.plumutil.clone(event, {changed:true}));
		} else {
		    // this.log("event! old! " + event.type + " " + (event.entity ? event.entity.id : event.entityid) + " " + last[0].start.valueOf() + " " + last[0].end.valueOf() + " " + event.start.valueOf() + " " + event.end.valueOf() + " "  + (event.end.valueOf() - last[0].end.valueOf())); // + " " + last[0].start + " " + event.end + " " + event.end.valueOf() - last[0].end.valueOf());
		    last = this_.util.clone(last[0]);
		    if (last.end.valueOf() <= event.end.valueOf()) {
			var timeextension = event.end.valueOf() - last.end.valueOf();
			//JV3.log("timeextension :" + timeextension + " new end : " + event.dn);
			this._fireUpdate(this_.util.clone({end:event.end,changed:false, timeextension:timeextension},last));
		    } else {
			this_.log("last end is not < event.end. fail? " + last.end.valueOf() + " " + event.end.valueOf() + " " + dehydrated.entityid);
		    }
		}
	    },
	    saveAsync:function(event,cont) {
		//this.log("saving event : " + event);
		var this_ = this;
		var dehydrated = this._deconstitute(event);
		// we want to omit end from the select_clauses so that we can easily overwrite previous observations
		// starting at the same point, e.g., extend previous observations
		this.getEventsAsync({type : event.type, entitytype: dehydrated.entitytype, entityid: dehydrated.entityid, start : dehydrated.start},
				       function(last) {						 						 
					   this_.saveCommonAsync( {type:event.type, entitytype: dehydrated.entitytype, entityid: dehydrated.entityid, start: dehydrated.start /*, end : dehydrated.end */ },
								  dehydrated, cont);
					   //JV3.log("raw hits " + dehydrated.entityid + " " + last.length);					   
					   if (last.length === 0) {
					       this_._fireUpdate(this_.plumutil.clone(event, {changed:true}));
					   } else {
					       last = JV3.plumutil.clone(last[0]);
					       if (last.end.valueOf() <= event.end.valueOf()) {
						   var timeextension = event.end.valueOf() - last.end.valueOf();
						   //JV3.log("timeextension :" + timeextension + " new end : " + event.dn);
						   this_._fireUpdate(this_.util.clone({end:event.end,changed:false, timeextension:timeextension},last));
					       } else {
						   this_.log("last end is not < event.end. fail? " + last.end.valueOf() + " " + event.end.valueOf() + " " + dehydrated.entityid);
					       }
					   }
				       });
		//this.log(">> savecommon : " + event);
		//this.log("<< savecommon : " + event);	    
		
	    },
	    
	    /* schema-specific stuff */
	    saveCalEvent: function(calEvent) {
		// stores a schema
		var this_ = this;

		// fails bross-browser env plumutil.assert(PROTOCROCK.isinstanceof(calEvent,schemas.Event), "Must be a type of Event");
		return calEvent.dates.map(function(d) {
					      this_.save({ type: calEvent.type, entity: calEvent, start: d[0], end:d[1] });
					  }).length;	
	    },
	    getMostRecentEvents: function() {
		// loads the most recent single event of each type into changeCache
		// so that we can compare new events to the past.
		var query = "SELECT max(end),* from " + this.table + " GROUP BY type;";
		var db = this._db;
		var this_ = this;
		var statement = db.createStatement(query);
		var results = {};
		var fn = function(select_keys,statement) {
		    var result = {};
		    select_keys.map(function(key) { result[key] = this_.column_grabbers[key](statement, select_keys.indexOf(key) ); });
		    return this_._reconstitute(result);
		};
		try {
		    while(statement.executeStep()) {
			var r = fn(["end"].concat(this.schemacols), statement);
			if (r) {  results[r.type] = r;}
		    }	    
		} catch (e) {
		    this.log("forStatementsMatching error " + e);
		} finally {
		    statement.reset();
		}
		return results;
	    }
	};

	ContextMaiden.EntityInflaterMixin = {
	    _reconstitute:function(dehydrated_entity) {
		var this_ = this;
		try {
		    JV3.plumutil.assert(dehydrated_entity.type !== undefined, "Entity type cannot be undefined.");
		    var rehydrated = { id: dehydrated_entity.id, type: dehydrated_entity.type } ;
		    if (dehydrated_entity.data) { JV3.plumutil.clone(JV3.plumutil.fromJSONAggressive(dehydrated_entity.data),rehydrated); }
		    return this._reconstitute_h(rehydrated);
		} catch(e) {
		    this_.log(dehydrated_entity.type + " " + dehydrated_entity.id + " " + e + " " + e.lineNumber + " " + e.fileName);
		}
		return undefined;
	    },	    
	    _reconstitute_h:function(v) {
		// new reconstitute_helper 
		var this_ = this;
		if (v === null || v === undefined) { return undefined;  }
		if (typeof(v) == 'string' && v.indexOf('(new Date(') == 0) {
		    var ss = v.substring('(new Date('.length,v.indexOf(')'));
		    JV3.plumutil.assert(!isNaN(parseInt(ss)), "Cannot parse " + ss);
		    return new Date(parseInt(ss));
		}
		// check if it's a type first.
		if (typeof(v) == 'object' && v.type == 'type' && JV3.schemas.getType(v.id) !== undefined) { return JV3.schemas.getType(v.id);  }
		// not a type, regular entity.
		if (typeof(v) == 'object' && v.type !== undefined && v.id !== undefined) {
		    // first restore all the keys
		    v = JV3.plumutil.objMap(function(vv) { return this_._reconstitute_h(vv); },v);
		    var t = JV3.schemas.getType(v.type);
		    if (t === undefined) {
			JV3.log("Warning using generic schemas.Entity for " + v.id); t = JV3.schemas.Entity;
		    }
		    return JV3.newify(t, v); 		    
		}
		if (typeof(v) == 'object' && JV3.plumutil.isArrayLike(v)) {
		    return v.map(function(xv) { return this_._reconstitute_h(xv); });
		}
		return v;		
	    },
	    /* _reconstitute_helper:function(v) {
		// for every field
		var this_ = this;
		if (v === null || v === undefined) { return undefined;  }
		if (typeof(v) == 'string' && v.indexOf('(new Date(') == 0) {
		    var ss = v.substring('(new Date('.length,v.indexOf(')'));
		    JV3.plumutil.assert(!isNaN(parseInt(ss)), "Cannot parse " + ss);
		    return new Date(parseInt(ss));
		}
		if (typeof(v) == 'object' && v.type && v.id ) {
		    // entity! retrieve it and hook it up

		    // check if it's a type first.
		    if (v.type == 'type' && JV3.schemas.getType(v.id) !== undefined) { return JV3.schemas.getType(v.id);  }
	     
		    // if it's not a type then it's an entity.
		    var en = this.getEntities( {id:v.id, type:v.type}, undefined, true ); // avoid the cache since we might be cycling indefinitely
		    if (en.length > 0) { return en[0];}
		    try { JV3.plumutil.assert(false, "Could not restore entity (type: " + v.type + "; id: " + v.id + ")"); }
		    catch (x) {	JV3.log(x); }
		    JV3.plumutil.assert(false, "Could not restore entity (type: " + v.type + "; id: " + v.id + ")");	
		    return;
		}
		if (typeof(v) == 'object' && JV3.plumutil.isArrayLike(v)) {
		    return v.map(function(xv) { return this_._reconstitute_helper(xv); });
		}
		return v;
	    }, 
	     */
	    _deconstitute_skel : function(ent) { return {type:ent.type,id:ent.id}; },
	    _deconstitute_helper:function(v) {

		var this_ = this;
		if (v === undefined || v === null) { return undefined; }
		if (typeof(v) == 'object' && JV3.plumutil.isDateLike(v)) {
		    return "(new Date("+v.valueOf()+"))";
		}
		if (typeof(v) == 'object' && v.type && v.id ) {
		    return this._deconstitute_skel(v);
		}
		if (typeof(v) == 'object' && JV3.plumutil.isArrayLike(v)) {
		    return JV3.plumutil.toJSON( v.map(function(xv) { return this_._deconstitute_helper(xv); }));
		}
		return v;
	    },
	    _deconstitute: function(entity) {
		try {
		    var this_ = this;
		    var basis = {};
		    // << debug

		    // unncessary check -- now that we have things that can have types that people don't know about
		    // var etype = JV3.schemas.getType(entity.type);
		    // if (etype === undefined) {
		    // 	throw new Error("cannot deconstitute  entity " + entity.id + " of type " + entity.type + " because that type was not found! ");
		    // }
		    
		    // new mojo
		    JV3.schemas.getPublicFields(entity).map(function(key) { basis[key] = this_._deconstitute_helper(entity[key]);  });
		    var dried = {};
		    ["id","type"].map(function(key) { dried[key] = entity[key]; });		
		    dried.data = JV3.plumutil.toJSON(JV3.plumutil.clone(basis,{},["id","type"]));
		    return dried;
		} catch(ee) {
		    JV3.log(ee+ " " + ee.fileName + " " + ee.lineNumber + " "+ ee.stack + " " + entity.id + " " + entity.type);
		}	  
	    }
	};

	ContextMaiden.EntityStore = {
	    superclass: ContextMaiden.CMStore,
	    schema: "(id TEXT, type TEXT, data TEXT)",
	    schemacols: ["id","type","data"], // "id" identifies logical ids
	    index:["id","type"],
	    column_grabbers: {
		id:    ContextMaiden.mozGetType.string,
		type : ContextMaiden.mozGetType.string,
		data : ContextMaiden.mozGetType.string
	    },
	    caching:false,
	    special_handlers:[],
	    initialize: function(params) {
		JV3.PROTOCROCK.superclass(this,"initialize", params ? params : {});
		JV3.plumutil.clone(ContextMaiden.EntityInflaterMixin,this);
		this.table = "entities"; // (params === undefined || params.tablename === undefined) ? "entities" : params.tablename;
		this._createTable();
	    },
	    get_or_create:function(template) {
		// Convenience method which looks up entity by the keys  ------------
		// if a matching entity [or entities] exist returns that entity
		JV3.plumutil.assert(template.type !== undefined, "get_or_create(): template must have a type");
		var type = JV3.schemas.getType(template.type);
		var this_ = this;
		var create_new_guy = function() {
		    JV3.log("++++++++++++++++ create new guy type is: " + type + " " + template.type);
		    var new_guy = JV3.newify(type, template);
		    this_.save(new_guy);
		    return new_guy;		
		}
		JV3.log("++++++++++++++++ getting new guy type is: " + type + " " + template.type);
		var ent = this.getEntities(template);
		if (ent.length == 0) { return create_new_guy();   }
		if (ent.length > 1) {
		    JV3.log("WARNING: got more than one entity when we're supposed to be talking unique here : " + ent.map(function(X) { return X.id; })).join(",");
		}
		return ent[0];
	    },
	    get_or_createAsync:function(template,cont) {
		// Convenience method which looks up entity by the keys  ------------
		// if a matching entity [or entities] exist returns that entity
		JV3.plumutil.assert(template.type !== undefined, "get_or_create(): template must have a type");
		var type = JV3.schemas.getType(template.type);
		var this_ = this;
		var create_new_guy = function() {
		    // JV3.log("createasync CREATING.. ++++++++++++++++ type is: " + type + " " + template.type + " " + template.id);
		    var new_guy = JV3.newify(type, template);
		    this_.saveAsync(new_guy, function() { cont(new_guy); });
		}
		this.getEntitiesAsync(template,undefined,function(ent) {
					  if (ent.length == 0) { return create_new_guy();			  }
					  if (ent.length > 1) { JV3.log("WARNING: got more than one entity when we're supposed to be talking unique here : " + ent.map(function(X) { return X.id; })).join(",");  }
					  return cont(ent[0]);				
				      });
	    },	
	    getEntityById: function(id) {
		var results = this.getEntities( {id:id});
		return results.length ? results[0] : undefined;
	    },
	    getEntityByIdRaw:function(id) {
		// get entity by id, but don't dehydrate
		var results = [];
		this._forResultsArray([{col:"id", op:"=", val:id}], "AND", function(r) { results.push(r); });
		return results;
	    },
	    addSpecialQueryHandler:function(qh) {
		// useful for adding special query handling certain
		// kinds of entitis
		this.special_handlers.push(qh);
	    },
	    _compare: function(x,y) {
		var cmp = arguments.callee;
		if (x === undefined) { return y === undefined; }
		if (x == y) { return true; } // avoid later checks.
		// deep compare in case we get entities or arrays as properties
		if (JV3.plumutil.isArrayLike(x) && JV3.plumutil.isArrayLike(y)) {
		    if (x.length !== y.length) { return false; }
		    for (var i = 0; i < x.length; i++) {
			if (!JV3.ContextMaiden.EntityStore._compare(x[i],y[i])) { // for some reason referring by this or this_ fails
			    return false;
			}
		    }
		    return true;
		}
		if (JV3.plumutil.isDateLike(x) && JV3.plumutil.isDateLike(y)) {
		    return x.valueOf() == y.valueOf();
		}
		if (typeof(x) == "object" && typeof(y) == "object") {
		    return JV3.plumutil.objKeys(x).concat(JV3.plumutil.objKeys(y)).map(function(xk) {
											   if (! cmp(x[xk],y[xk]) ) {  return false;  }
											   return true;
										       }).filter(function(x) { return x; }).length == JV3.plumutil.objKeys(x).concat(JV3.plumutil.objKeys(y)).length;
		}
		return false; // fell through.
	    },
	    getEntities:function(template,datalike,punt_cache) {
		// template is something like
		// { key : val, key2: [op1(val)
		// datalike : an array of strings that are likely to be in data

		//JV3.log("getEntities" + " " + JV3.plumutil.printObj(template));
		
		var this_ = this;	    
		// try special qhs first
		var qhresult = JV3.plumutil.mapappend(function(qh) { return qh(template); }, this.special_handlers).filter(function(x) { return x !== undefined; });	    
		if (qhresult.length > 0) {
		    return qhresult;
		}

		var possible_results = [];
		if (this.caching && !punt_cache) {
		    if (this._cache === undefined) {
			var _cache = [];
			this._forResultsArray([], "AND", function(r) {
						  var rer = this_._reconstitute(r);
						  if (rer === undefined) { JV3.log("ERROR reconstituting " + r.id); return; };
						  _cache.push(rer);
					      });
			this._cache = _cache;
		    }
		    possible_results = this._cache;
		} else {		
		    var dehydrated = [];
		    ['type','id'].map(function(key) {
					  if (template[key] !== undefined) {
					      dehydrated.push({ col:key, op:"=", val: template[key] });
					  }
				      });
		    // handle datalike clause
		    if (datalike !== undefined) {
			if (typeof(datalike) === 'string') {datalike = [datalike];}
			datalike.map(function(x) {
					 dehydrated.push({ col:"data", op:" LIKE ", val: "%"+x+"%" });
				     });
		    }		
		    this._forResultsArray(dehydrated, "AND", function(r) {
					      var reconst = this_._reconstitute(r);
					      possible_results.push(reconst);
					  });
		}
		return possible_results.filter(function(x) {
						   if (x === undefined) { return false;  } 
						   for (var i = 0; i < JV3.plumutil.objKeys(template).length; i++) {
						       var key = JV3.plumutil.objKeys(template)[i];
						       if (["id","type"].indexOf(key) >= 0) { continue; }
						       if (this_._compare(template[key],x[key]) == false) {
							   return false;
						       }
						   }
						   return true;
					       });
	    },
	    _getEntitiesAsync:function(template,reconstitution_fn,datalike,cont) {
		// template is something like
		// { key : val, key2: [op1(val)
		// datalike : an array of strings that are likely to be in data

		var this_ = this;
		template = template !== undefined ? template : {};
		
		// try special qhs first
		var qhresult = JV3.plumutil.mapappend(function(qh) { return qh(template); }, this.special_handlers).filter(function(x) { return x !== undefined; });	    
		if (qhresult.length > 0) {
		    return qhresult;
		}
		var dehydrated = [];
		['type','id'].map(function(key) {  if (template[key] !== undefined) { dehydrated.push({ col:key, op:"=", val: template[key] }); } });
		
		// handle datalike clause
		if (datalike !== undefined) {
		    if (typeof(datalike) === 'string') {datalike = [datalike];}
		    datalike.map(function(x) { dehydrated.push({ col:"data", op:" LIKE ", val: "%"+x+"%" });});
		}
		// JV3.log(dehydrated);
		this._forResultsArrayAsync(dehydrated, "AND", reconstitution_fn, 
					   function(possible_results) {
					       if (typeof(possible_results) == 'object' && possible_results.error) { return cont([]);     }
					       cont(
						   possible_results.filter(
						       function(x) {
							   try {
							       if (x === undefined) { return false;  } 
							       for (var i = 0; i < JV3.plumutil.objKeys(template).length; i++) {
								   var key = JV3.plumutil.objKeys(template)[i];
								   if (["id","type"].indexOf(key) >= 0) { continue; }
								   if (!this_._compare(template[key],x[key])) { return false; }
							       }
							       return true;							       
							   } catch (x) {  JV3.log(x);  }
							   return false;
						       }));
					   });
	    },
	    getEntitiesAsync:function(template,datalike,cont) {
		var this_ = this;
		this._getEntitiesAsync(template,function(r) { return this_._reconstitute(r); },datalike,cont);
	    },
	    getEntitiesAsyncRaw:function(template,datalike,cont) {
		this._getEntitiesAsync(template,function(r) { return r; },datalike,cont);
	    },
	    mergeEntities:function(entity1,entity2) {
		// sets entity1 equal to entity2
		var merged = this._merge_into_one__arrdupes(entity1,entity2);
		// JV3.log("merged : " + merged);
		// delete entity2
		this.deleteEntity(entity2);
		this.save(merged);
		return merged;
	    },	
	    deleteEntity:function(entity) {
		return this.remove( {id:entity.id, type:entity.type} );
	    },
	    addEntityPropertyUpdateListener:function(f) {
		if (this._entity_property_update_listeners.indexOf(f) < 0) {
		    this._entity_property_update_listeners.push(f);
		}
	    },
	    // entity property update dudes 
	    _entity_property_update_listeners : [],
	    _fire_diff:function(olde,newe) {
		var this_ = this;
		if (this_._entity_property_update_listeners.length == 0) { return; }
		var diffprops = JV3.plumutil.remove_dups(JV3.plumutil.objKeys(olde).concat(JV3.plumutil.objKeys(newe))).filter(
		    function(k) {
			return k.indexOf("_") !== 0 && typeof(olde[k]) !== 'function' && typeof(newe[k]) !== 'function' &&  olde[k] != newe[k];
		    }
		);	    
		this._entity_property_update_listeners.map(function(l) { try { l(JV3.plumutil.clone(newe),diffprops); } catch (x) {   JV3.log(x);  }  });			  
	    },
	    _fire_new:function(newe) {
		var this_ = this;
		if (this._entity_property_update_listeners.length == 0) { return; }
		this._entity_property_update_listeners.map(function(l) { try { l(JV3.plumutil.clone(newe),"new"); } catch (x) {   JV3.log(x);  }  });			  
	    },
	    save: function(entity) {
		if (entity === undefined) {  JV3.log("save rrror."); JV3.log(entity); throw "Entity is undefined"; }
		var old = this.getEntities({id:entity.id,type:entity.type},undefined,true); //this._cache.filter(function(x) { return x.id == entity.id && x.type == entity.type; });
		if (old.length > 0) {
		    if (this.cache) { JV3.plumutil.removeElement(this._cache,entity); }
		    this._fire_diff(old[0], entity); // old item
		} else {
		    // new item!
		    this._fire_new(entity);
		}
		if (this._cache) {this._cache.push(entity); }
		var dehydrated = this._deconstitute(entity);
		this.saveCommon( { type: dehydrated.type, id: dehydrated.id }, dehydrated );
		this._fireUpdate(entity);
		return entity;
	    },
	    saveAsync: function(entity,cont) {
		var this_ = this;
		if (entity === undefined) {   JV3.log("saveAsync Error."); JV3.log(entity); throw "Entity is undefined";}
		var old = this.getEntitiesAsync({id:entity.id,type:entity.type},undefined,
						function(old) {
						    if (old.length > 0) {
							if (this_.cache) { JV3.plumutil.removeElement(this_._cache,entity); }
							this_._fire_diff(old[0], entity); // old item
						    } else {
							// new item!
							this_._fire_new(entity);
						    }
						    if (this_._cache) { this_._cache.push(entity); }
						    var dehydrated = this_._deconstitute(entity);
						    this_.saveCommonAsync( { type: dehydrated.type, id: dehydrated.id }, dehydrated, cont );
						    this_._fireUpdate(entity);
						}); 
	    },	    
	    _merge_into_one__arrdupes: function(dest,src) {
		// merges fields of src into dest. if dest already has the key, either
		var id = dest.id;
		var type = dest.type;	    
		JV3.plumutil.objKeys(src).map(function(key) {
						  if (src[key] == undefined || src[key] == null) { return; }
						  if (dest[key] === undefined) { return dest[key] = src[key]; }
						  if (dest[key] == src[key]) { return; }
						  if (JV3.plumutil.isArrayLike(dest[key])) {
						      if ( dest[key].indexOf(src[key]) < 0 ) {return dest[key].push(src[key]);}
						  }
						  dest[key] = [dest[key],src[key]];
					      });
		dest.id = id;
		dest.type = type;
		return dest;
	    }
	};


	ContextMaiden.AliasStore = {
	    superclass:ContextMaiden.CMStore,
	    schema: "(alias TEXT, entityid TEXT, entitytype TEXT)",
	    schemacols: ["alias","entityid","entitytype"],
	    column_grabbers: { alias : ContextMaiden.mozGetType.string,
			       entityid : ContextMaiden.mozGetType.string,
			       entitytype : ContextMaiden.mozGetType.string},
	    
	    initialize:function(params,entity_store) {
		JV3.PROTOCROCK.superclass(this,"initialize", params ? params : {});
		this.table = "entity_aliases";
		this._createTable();
		this.entity_store = entity_store;
	    },	
	    lookup:function(alias) {
		var uris = [];
		alias = (alias+"").toLowerCase();
		var res = this._forResults( {alias:alias}, function(ent) {
						// should we attempt to get the entity?
						if (ent.entityid) { uris.push( ent.entityid ); }
					    });
		return uris.map(function(x) { return JV3.CMS.entity_store.getEntityById(x); });
	    },
	    getAliasesFor:function(entityid,entitytype) {
		var aliases = [];
		var res = this._forResults( {entityid:entityid,entitytype:entitytype}, function(ent) {
						// should we attempt to get the entity?
						if (ent.alias) { aliases.push( ent.alias ); }
					    });
		return aliases;
	    },	
	    match:function(alias) {
		// supports _like_ strings
		var query = "select alias,entityid,entitytype from " + this.table + " where alias LIKE \"" + alias + "%\";";
		//JV3.log(query);
		var results = this._getResults(query,function(x,i) { return x.getString(i); },[0,1,2]);
		//JV3.log(results.length);
		return results.map(function(idtype) {
				       var id = idtype[1];
				       var type = idtype[2];
				       //JV3.log("getting  id : " + idtype[0] + ": " + idtype[1]);
				       return JV3.CMS.entity_store.getEntityById(id);		
				   });
	    },
	    _lookup_uris:function(alias) {
		var uris = [];
		this._forResults( {alias:alias}, function(ent) {
				      // should we attempt to get the entity?
				      uris.push( ent.entityid );
				  });
		return uris;
	    },
	    addAlias:function(entity,alias) {
		alias = (alias+"").toLowerCase();
		var repeats = this._lookup_uris(alias).filter(function(x) { return x == entity.id; });
		if (repeats.length > 0) {
		    return;
		}
		this.insert( {alias:alias, entityid: entity.id, entitytype:entity.type } );
	    },
	    update_aliases:function() {
		var ents = this.entity_store.getEntities({});
		//JV3.log("entities: " + ents.length);
		var this_ = this;
		ents.map(function(entity) {
			     if (entity.aliases !== undefined) {
				 entity.aliases.map(function(key) {
							if (entity && entity[key] && key) {
							    //JV3.log("adding alias : " + entity[key] + " to " + entity.id);
							    this_.addAlias(entity,entity[key]);
							}
						    });
				 if (entity.getLabel !== undefined) {
				     this_.addAlias(entity.getLabel(),entity);
				 }
			     }
			 });	    
	    }
	};
    });


	    // getEntitiesAsyncFast:function(template,cont) {
	    // 	// template is something like
	    // 	// { key : val, key2: [op1(val)
	    // 	// datalike : an array of strings that are likely to be in data

	    // 	var this_ = this;
	    // 	template = template !== undefined ? template : {};
	    // 	// try special qhs first
	    // 	var qhresult = JV3.plumutil.mapappend(function(qh) { return qh(template); }, this.special_handlers).filter(function(x) { return x !== undefined; });	    
	    // 	if (qhresult.length > 0) {
	    // 	    return qhresult;
	    // 	}
	    // 	var dehydrated = [];
	    // 	['type','id'].map(function(key) {  if (template[key] !== undefined) { dehydrated.push({ col:key, op:"=", val: template[key] }); } });
	    // 	this._forResultsArrayAsync(dehydrated, "AND", function(r) JV3.plumutil.clone(r.data,{id:template.id,type:template.type},
	    // 				   function(possible_results) {
	    // 				       if (typeof(possible_results) == 'object' && possible_results.error) { return cont([]);     }
	    // 				       cont( possible_results.filter(
	    // 						 function(x) {
	    // 						     if (x === undefined) { return false;  } 
	    // 						     for (var i = 0; i < JV3.plumutil.objKeys(template).length; i++) {
	    // 							 var key = JV3.plumutil.objKeys(template)[i];
	    // 							 if (["id","type"].indexOf(key) >= 0) { continue; }
	    // 							 if (!this_._compare(template[key],x[key])) { return false; }
	    // 						     }
	    // 						     return true;
	    // 						 }));
	    // 				   });
	    // },	    

/*
	    //this.log("searching for template : " + this.plumutil.toJSON(dehydrated));
this.forStatementsMatching( dehydrated, function(select_keys,statement) {
		var flattened = {};
		select_keys.map(function(key) {
		    flattened[key] = statement.getUTF8String( select_keys.indexOf(key) );
		});
		JV3.plumutil.assert(flattened.id !== undefined, "No id on restored thing");
		// will be undefined if hasn't been reconstituted yet
		results[flattened.id] = this_._reconstitute(flattened,results[flattened.id]);
	    });
	    return JV3.plumutil.objVals(results);



*/
