// Copyright (c) 2008,2009 Massachusetts Institute of Technology
// Copyright (c) 2009 Nokia

// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:

// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.

// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

(function(JV3,ContextMaiden) {

    var plumutil = JV3.plumutil;
    var newify = JV3.newify;
    
    ContextMaiden.OilLocationXPCOM = {
	id: 'ContextMaiden.OilLocation',
	type:"location",
	server: "localhost",
	lastevent : undefined,
	initialize:function() {
	    var this_ = this;
	    this.log = JV3.log;
	    this.log("Initializing Oil Location services");
	    this.oil = this.__initOil();
	    plumutil.assert(JV3 !== undefined, "ContextMaiden.OilLocation JV3 is undefined");
	    plumutil.assert(this.oil !== undefined, "OIL failed to initialize");
	},
	__initOil : function() {
	    // get the oil running
	    try {
		var oilService = Components.classes["@mit.edu/oil/location;1"].getService();
		oilService = oilService.QueryInterface(Components.interfaces.nsIOilLocation);
		return oilService;
	    } catch(e) {
		this.log("__initOil: " + e + " " + e.lineNumber + " " + e.fileName);
	    }
	},
	shutdown:function() {
	    this.oil = undefined;
	},
	initOverlay:function(w, document, JV3) {
	    return newify(ContextMaiden.OilOverlay, w, document, JV3);
	},
	_process:function(location, event_store) {
	    try {	    				 
		this.log("OIL XPCOM LOCID : " + location);
		var new_loc =
		    location.type == 'schemas.Location' ? location :
		    JV3.CMS.entity_store.get_or_create({type:"schemas.Location", id:location, name: location, probability:1.0});

		JV3.plumutil.assert(new_loc !== undefined, "Location is undefined ");

		// current HACK: we dont know how long we've been here
		var new_start = new Date() - 1;
		var new_end = new Date();
		
		if (this.lastevent && this.lastevent.entity.id !== new_loc.id) {
		    // trim off prediction so that we don't overlap
		    this.lastevent.end = new_start;
		    event_store.save(this.lastevent);
		    // fall through so we can create a new event
		} else if (this.lastevent && this.lastevent.entity.id === new_loc.id) {
		    this.lastevent.end = new_end;
		    event_store.save(this.lastevent);
		    return this.lastevent;
		}
		// new event yo.
		var new_event = { type: "location", start: new_start, end:new_end, entity: new_loc };
		event_store.save( new_event );
		this.lastevent = new_event;  // update our last event to the new guy
		return new_event;	    
	    }  catch(e) {
		this.oil = undefined;
		this.log(e + " " + e.lineNumber + " " + e.fileName);
	    }
	},
	update: function(event_store) {
	    this.log("OIL XPCOM update");
	    if (!this.oil) { this.oil = this.__initOil();   }
	    var new_locid = this.oil.whereami();

	    if (new_locid == null || new_locid == "<nodata>") {
		this.log("oil-location: localizer returned Null. We'll call that nodata.");
		new_locid = "Somewhere";
	    }
	    var evt = this._process(new_locid, event_store);
	    JV3.informObservers("location-changed", evt.entity.id);
	},
	setLocation:function(s) {
	    // called when location is explicitly set in the ui
	    var event_store = JV3.CMS.event_store;
	    var entity_store = JV3.CMS.entity_store;
	    
	    if (s.length > 0) {
		this.log("Calling iam here : " + s);
		this.oil.iamhere(s);
	    }
	    
	    // should we create an event -- or look for it in the db?
	    // find nearest neighbor:
	    var evtmatches = JV3.CMS.entity_store.get_or_create({
		type: "schemas.Location",
		id:s,
		name:s
	    });
	    
	    this._process(evtmatches,event_store);
	    JV3.informObservers("location-changed", evtmatches.id);
	}
    };

    ContextMaiden.OilLocationTCP = {
	id: 'ContextMaiden.OilLocation',
	type:"location",
	server: "localhost",
	port: 8032,
	lastevent : undefined,
	time_smidge: 60*1000,
	initialize:function(params) {
	    this.log = params.log;
	    this.log("Initializing Oil Location services");
	},
	_get_url: function() { return "http://"+this.server+":"+this.port+"/whereami"; }, 
	_process:function(result,textStatus,event_store) {
	    if (result === undefined || !plumutil.isArrayLike(result)) {
		return this._fail(undefined,"malformed response from location server", undefined);
	    }
	    if (result.length === 0) { /*  we have no idea where we are. */  return; }
	    var new_locid = result[0].location;
	    var probability = result[0].probability;
	    var new_start = new Date();
	    
	    // current HACK: we dont know how long we've been here
	    var new_end = new Date(new_start.valueOf() + this.time_smidge);
	    if (this.lastevent) {
		if (this.lastevent.entity.id == new_locid) {
		    this.lastevent.end = new_end;
		    event_store.save(this.lastevent);
		    return;
		}
	    }
	    var new_event = {
		type: "location",
		start: new_start, end:new_end,
		entity: newify(JV3.schemas.Location, {id:new_locid, name: new_locid, probability: probability})
	    };
	    event_store.save( new_event );
	    this.lastevent = new_event;  // update our last event to the new guy
	    return new_event;
	},
	_fail:function(req,status,errorThrown) {
	    this.log(status + errorThrown);
	},
	update: function(event_store) {
	    var this_ = this;
	    jQuery.ajax({
		url: this._get_url(),
		type: "GET",
		dataType: "json",
		success:function(data,textStatus) {
		    this_._process(data,textStatus,event_store);
		},
		error: function(req, status, errorThrown) {
		    this_._fail(req,status,errorThrown);
		}
	    });
	}
    };

    ContextMaiden.OilOverlay = {
	time_to_stick_after_manual_override: 5*60000,
	initialize:function(overlaywindow,overlaydoc) {
	    var this_ = this;
	    this.log = JV3.log;
	    this.log("OilOverlay constructor for document " + overlaydoc);
	    this.overlay = overlaydoc;
	    var this_ = this;
	    try {
		var sb = overlaydoc.getElementById("status-bar");
		var el = overlaydoc.createElementNS("http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul", "statusbarpanel");
		el.setAttribute("id","oillocation");
		el.setAttribute("popup","locationPanel");
		sb.appendChild(el);
		el.addEventListener("keypress", function(event) { this_._keycb.apply(this_,[this,event]); }, false);
		this.el = el;
		this.setLocationDisplay("Localizing...");
		this._populate_places();
	    } catch(e) {
		this.log("OilOverlay Error : " + e);
	    }
	    this._add_observers(overlaywindow);
	},
	_add_observers:function(overlaywindow) {
	    var this_ = this;
	    var observer = {
		observe:function(subject,topic,data) {
		    if (topic == "location-changed") {
			this_._cb_location_changed(data);
		    } else {
			this_.log("Observer -- Got something weird : " + topic);
		    }
		}
	    };
	    JV3._getObserverService().addObserver(observer,"location-changed",false);
	    JV3.addCloseHookForWindow(overlaywindow, function() {  JV3._getObserverService().removeObserver(observer,"location-changed");  });
	},    
	_keycb: function(component,event) {
	    var key = event.data;
	    this.log("keyevent on the location : " + key);
	},
	setLocationDisplay:function(s) {
	    this.el.label = s;
	},
	getEntityStore:function() {
	    return JV3.CMS.entity_store;
	},
	_populate_places:function() {
	    var widget = this.overlay.getElementById("locationPanelTextbox");
	    if (widget === undefined || widget === null) {
		return;
	    }
	    var locations = this.getEntityStore().getEntities( {type: "schemas.Location"} );
	    this._flush_places();
	    locations.map(function(l){
		widget.appendItem(l.name,l.id,l.name);
	    });
	},
	_flush_places: function() {
	    //
	    var widget = this.overlay.getElementById("locationPanelTextbox");
	    // @TODO -- tofix. widget is sometimes null
	    if (widget && widget !== null) {
		try {
		    widget.removeAllItems();
		} catch(e) {
		    this.log(e + " " + e.fileName +":"+ e.lineNumber);
		}
	    } else {
		this.log("ERROR list.it context-maiden-oil error widget locationPanelTextbox is " + widget);
	    }
	},
	_cb_location_changed:function(locid) {
	    var entities =
		this.getEntityStore().getEntities( {id:locid, type:"schemas.Location"} );
	    if (entities.length > 0) {
		this.setLocationDisplay(entities[0].name);
	    } else {
		this.log("OILOverlay: Could not find entity with id : " + locid);
		this.setLocationDisplay(locid);
	    }
	    this._populate_places();
	}    
    };
    ContextMaiden.OilLocation = ContextMaiden.OilLocationXPCOM;
})
