// Copyright (c) 2008,2009 Massachusetts Institute of Technology

// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:

// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.

// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

(function(JV3,ContextMaiden) {
    ContextMaiden.Scheduler = {
	queue: [],
	initialize:function(sources, params) {
	    JV3.log("Scheduler init........");
	    var this_ = this;
	    this.log = JV3.log;
	    this.sources = [];
	    this.ContextMaiden = ContextMaiden;
	    var filename = JV3.prefs.getCharPref("context.filename");
	    try {
		this.entity_store = newify(ContextMaiden.EntityStore,  { filename:filename, debug: false, log: function(x) { this_.log(x); } });
		JV3.log("entity store initialized : " + this.entity_store);
	    } catch(ee) {
		JV3.log(ee + " " + ee.lineNumber + " " + ee.fileName + " " + ee.stack);
	    }
	    this.event_store = newify(ContextMaiden.EventStore, JV3,
				      { filename : filename, debug: false, log: function(x) { this_.log(x); } },
				      this.entity_store );
	    this.alias_store = newify(ContextMaiden.AliasStore,
				      { filename : filename, debug: false, log: function(x) { this_.log(x); } },
				      this.entity_store);
	    
	    if (params.commit_interval) {
		this.log("Setting commit interval to " + params.commit_interval);
		this.committer_alarm = JV3.setInterval(function() { this_.updateAll(); }, params.commit_interval);
	    }	    
	    if (JV3.schemas.PairwiseContextTable !== undefined) {
		this.pstats = JV3.schemas.PairwiseContextTable.getInstance(this); 
	    }
	    if (JV3.ContextMaiden.NoteEventSource !== undefined) {
		JV3.log("instantiating note event source");
		this.note_event_source = newify(JV3.ContextMaiden.NoteEventSource,this);
	    }
	    this.util = JV3.plumutil;
	},
	shutdown:function()  {
	    var this_ = this;
	    if (this.committer_alarm) {
		try {
		    this.clearInterval(this.committer_alarm);
		} catch(e) {
		    this_.log(e + " " + e.lineNumber + " " + e.fileName);
		}
	    }
	    this.sources.map(function(s) {
		if (s.shutdown !== undefined) {
		    try { s.shutdown(); } catch(e) { this_.log(e); }
		}
	    });
	},
	getSourceByName:function(name) {
	    var hits = this.sources.filter(function(x) { return x.id == name });
	    if (hits.length) {	return hits[0];    }
	},
	updateAll:function() {
	    var this_= this;
	    try {
		this.sources.map(function(x) {
		    try {
			if (x.update !== undefined) {
			    x.update(this_.event_store, this_.entity_store);
			}
		    } catch(e) { this_.log(e); } 
		});
	    } catch(e2) {this_.log(e2);    }	    
	},
	initOverlays:function(w,document) {
	    var this_ = this;
	    return this.sources.map(function(source) {
		if (source.initOverlay) {
		    try {
			return source.initOverlay(w,document,this_.JV3);
		    } catch(e) {
			this_.log("Error initializing overlay : " + e + " " + e.lineNumber + " " + e.fileName);
		    }
		}
	    });
	}
    };
})