
(function(JV3,ContextMaiden) {
    
    var newify = JV3.PROTOCROCK.newify;
    var plumutil = JV3.plumutil;
    var INSTANT_HOLD = 60000;
    
    ContextMaiden.TimeStruct = {
	makeInstant : function(d) {
	    plumutil.assert(d !== undefined, "makeInstant passed an undefined");
	    return newify(ContextMaiden.TimeStructType, ({instant:this._makeIntoDateLike(d)}));
	},
	makeDuration : function(start, end, padding) {
	    plumutil.assert(start !== undefined && end !== undefined, "makeInstant passed an undefined");
	    if (padding === undefined) { padding = 0; }
	    return newify(ContextMaiden.TimeStructType, ({start:start.valueOf()-padding,end:end.valueOf()+padding}));
	},    
	// statics
	_makeIntoDateLike: function(d) {
	    if (d === undefined || d === null) { return undefined; }
	    if (typeof(d) == 'object' && d.type =='schemas.DateTime') { return d.getDate(); } 
	    if (plumutil.isDateLike(d)) { return d; }
	    if (typeof(d) == 'string' && Date.parseDate && Date.parseDate(d) !== null) { return Date.parseDate(d); }
	    if (typeof(d) == 'number') { return new Date(d); }
	    plumutil.assert(false, "Cannot convert " + d + " into a date ");
	    return undefined;
	}
    };

    ContextMaiden.TimeStructType =  {
	type:"TimeStruct",
	initialize: function(def) {
	    for (var k in def) {
		this[k] = ContextMaiden.TimeStruct._makeIntoDateLike(def[k]);
	    }
	},
	isInstant : function() { return this.start === undefined && this.end === undefined && this.instant !== undefined; },
	isInterval : function() { return this.start !== undefined && this.end !== undefined; },
	asDuration : function(padded) {
	    // makes an instant look like an interval
	    var mkduration = ContextMaiden.TimeStruct.makeDuration;
	    if (this.isInstant()) {
		return mkduration(this.instant.valueOf() - (padded ? this.INSTANT_HOLD/2 : 0), this.instant.valueOf() + (padded ? this.INSTANT_HOLD/2 : 1));
	    }
	    return padded ? mkduration(this.start.valueOf() - this.INSTANT_HOLD/2, this.end.valueOf() + this.INSTANT_HOLD/2) : this;
	},
	intersects : function(st,padded) {
	    var d1 = this.asDuration(padded);
	    var d2 = st.asDuration(padded);
	    return (d1.end >= d2.start && d1.start <= d2.end); 
	},
	length : function (padded) {
	    var d = this.asDuration(padded);
	    return d.end.valueOf() - d.start.valueOf();
	},
	intersection: function(other) {
	    var d1 = this.asDuration(false), d2 = other.asDuration(false);
	    if (this.intersects(other)) {
		return ContextMaiden.TimeStruct.makeDuration(Math.max(d1.start, d2.start),Math.min(d1.end, d2.end));			      
	    }
	    return undefined;
	},
	toString : function() {
	    if (this.isInterval()) { return "TSInterval ("+this.start+"-"+this.end+")"; }
	    return "TSInstant " + this.instant;	    
	}
    };
});


