// Copyright (c) 2009,2008 Massachusetts Institute of Technology

// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:

// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.

// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

ContextMaiden.ContextBars = {
    cattypes:  { 'web' : ['schemas.Webpage'],
		 'music' : ['schemas.MusicTrack'],
		 'place' : ['schemas.Location'],
		 'events' : ['schemas.Event'],
		 'time-of-day' : ['schemas.Hour'],
		 'day-of-week' : ['schemas.DayOfWeek']
	       }, /* 		 'note': ['schemas.NoteContext'], */
    cathsb : {
	'web' : {h:200,s:80,b:90},
	'music' : {h:41,s:80,b:90},
	'place' : {h:120,s:80,b:90}, //{h:334/360.0*255.0,s:80,b:100},
	'events' : {h:300,s:0,b:60},
	'time-of-day' : {h:360,s:0,b:60},
	'day-of-week' : {h:360,s:0,b:60}	
    },
    
    rows : {},
    rowlabel : {},
    _makeRowName:function(cat) { return "row-"+this.id+"-"+cat; },
    _makeRowLabelName:function(cat) { return "row-label-"+this.id+"-"+cat; },
    getRowEl:function(cat) { return jQuery("#"+this._makeRowName(cat)); },
    getRowLabelEl:function(cat) { return jQuery("#"+this._makeRowLabelName(cat)); },
    getCanvasEl:function(cat) { return jQuery("#"+this.id);   },
    
    initialize:function(JV3) {
	this.id = JV3.plumutil.guid();
	this.JV3 = JV3;
	this.log = JV3.log;
	this.cmodel = JV3.CMS.pstats;
	this.util = JV3.plumutil;
    },
    buildDOM:function(parent) {
	this.parent = parent;
	var container = jQuery("<div id=\""+this.id+"\" class=\"plot\">").appendTo(parent);
	var this_ = this;
	var catnames = this.cmodel.getTypes(); 
	var height_pct = 0.5/catnames.length; // % of 1
	var percat = 2*(height_pct)*this.getCanvasEl().height(); // number of pixels
	catnames.map(function(c) {
	    var idx = catnames.indexOf(c);
	    var y = idx*percat;	    
	    this_.rows[c] = jQuery("<div id=\""+this_._makeRowName(c)+"\" class=\"row\"/>").
		addClass(c).
		appendTo(container);	    
	    this_.rowlabel[c] =
		jQuery("<div id=\""+this_._makeRowLabelName(c)+"\" class=\"rowlabel\">"+c+"</div>").
		appendTo(container);		
	});
    },
    _getAssociationsFor: function(note) {
	// what we want to know is 
	// JV3.CMS.pstats.getCorrelatingEventsFor({type:"note-edited", entity:{id:"note-580257"}})
	return this.cmodel.getCorrelatingEventsFor({type:"note-edited",entity:{id:"note-"+note.id,type:"note"}});	
    },
    _closeFocusPlots:function() {
	if (this.focus) { this.focus.close(); this.focus = undefined; } 
    },
    getEventLabel:function(event) {
	var valid = this.cmodel.getEntityId(event);
	var store = this.JV3.CMS.entity_store;
	if (this.label_cache === undefined) {
	    this.label_cache = {};
	}
	if (this.label_cache[valid]) {
	    return this.label_cache[valid];
	}
	var label = valid;
	try {
	    var ent = store.getEntityById(valid);		    
	    if (ent.artist && ent.name) { label = ent.artist + " : " + ent.name; }
	    else if (ent.title) { label = ent.title; }
	    else if (ent.name) { label = ent.name; }
	} catch(e) {
	    console.log("ERROR!!!");
	    console.log(e);
	}
	this.label_cache[valid] = label;
	return label;
    },
    plotAssociationsForNote: function(note) {
	this._closeFocusPlots();
	var this_ = this;
	var store = this.JV3.CMS.entity_store;
	var evstore = this.JV3.CMS.event_store;
	var nc = this._getAssociationsFor(note);
	var catnames = this.cmodel.getTypes();

	catnames.map(function(cat) {
	    this_.getRowEl(cat).empty();
	    this_.getRowEl(cat).append("<div class=\"empty\">No " + cat + " associations</div>");
	});	
	if (this.util.objKeys(nc).length == 0) {  this.log("No associations for note : " + nc); return;}
	
	// sort by strength and compute total 
	this.util.objKeys(nc).map(function(cat) {
	    var items = nc[cat].concat();
	    items.sort(function(x,y) { return x.conditionalP < y.conditionalP; });
	    nc[cat] = items;
	});
	
	var plotwidth = jQuery("#"+this.id).width();
	var max_assoc = 1.0;
	
	var color_helper = function(cat) {
	    var c = this_.util.clone(this_.cathsb[cat]);
	    	    /*
	    console.log(strength);
	    c.h = c.h - strength; // (strength*100.0);
	    if (c.h < 0) { c.h = 255-c.h; }
	    console.log("Setting HSB to : " + c.h + " " + c.s + " " + c.b);
	    c = plumutil.HSBToRGB(c);
*/	    var cstring =  "rgb("+[Math.round(c.r),Math.round(c.g),Math.round(c.b)].join(',')+")";
	    //	    console.log(cstring); var N = nc[catname].length;
	    return cstring;    	    
	};

	catnames.map(function(catname) {
	    var row = this_.rows[catname];
	    row.empty();
	    if (!nc[catname]) { return; }	    
    	    var N = nc[catname].length;
	    nc[catname].map(function(event) {
		// var strength = nc_assoc[itemid]/(cat_total[catname]*1.0)*(plotwidth-N*2-1);
		// compute presentation title 
		// this is ghetto. what we really want is a lens
		var label = this_.getEventLabel(event);
		console.log("strength of " + label + " " + event.conditionalP);
		var strength_width = event.conditionalP*(plotwidth-N*2-1);
		
		jQuery("<div class=\"barslice\">"+"</div>").
		    width(0).
		    attr("title",label).
		    addClass("tooltip").
		    css("background-color",color_helper(catname)).
		    click(function() {
			console.log("click " + this_.getCanvasEl().length);
			this_.getCanvasEl().fadeOut("fast",
			    function() {
				var focus = newify(ContextMaiden.ContextFocus,this_.JV3);
				focus.show(this_,this_.parent,this_.getCanvasEl(),nc[catname],1.0);
				focus.getCanvasEl().fadeIn("fast");
				this_.focus = focus;
			    }
			);			       
		    }).
		    mouseover(function() {
			this_.getRowLabelEl(catname).html(catname + ": " + label);
			jQuery(this).addClass("over");
			/*
			if (jQuery(this).width() < 80) {
			    jQuery(this).width(80); 
			}*/
		    }).
		    mouseout(function() {
			this_.getRowLabelEl(catname).html(catname + ": ");
			jQuery(this).removeClass("over");
			//jQuery(this).width(strength);
			//jQuery(this).animate({width:strength});
		    }).
		    appendTo(row).
		    animate({width:strength_width});
	    });	    
	});
    }
};

ContextMaiden.ContextFocus = {
    initialize:function(JV3) {
	console.log("ContextFocus initialize");
	this.id = JV3.plumutil.guid();
	this.cmodel = JV3.CMS.pstats;
	this.JV3 = JV3;
	this.log = JV3.log;
	this.util = JV3.plumutil;
    },
    show:function(visual_parent,parent,enclosing,values,total) {
	this.parent = parent;
	this.enclosing = enclosing;
	jQuery(parent).find(".focusplot").remove();
	var container = jQuery("<div id=\""+this.id+"\" class=\"focusplot\">").hide().appendTo(parent);
	var this_ = this;
	var rowheight = 50;
	var store = this.JV3.CMS.entity_store;
	values.map(function(event) {
	    var idx = values.indexOf(event);
	    var label = visual_parent.getEventLabel(event);
	    var y = idx*rowheight;
	    var row = jQuery("<div id=\""+this_._makeRowName(event)+"\" class=\"row\"/>").
		addClass(label).
		height(rowheight).
		css("top", ""+y+"px").
		css("width","100%").
		appendTo(container);
	    var strength = event.conditionalP * 100.0;	    
	    jQuery("<div class=\"barslicefocus\">"+label+"</div>").resizable( {
		minHeight:50,maxHeight:50,
		resize:function() {
		    var w = jQuery(this).width()*1.0/jQuery(this).parent().width();
		    var ws = Math.floor((1.0-w)*255);
		    //console.log("rgb("+ws+","+ws/2+","+ws/4+")");
		    row.css("background-color","rgba(255,0,0,"+(w)+")"); 
		},
		stop: function() { console.log("stopped resize" + jQuery(this).width()); }
	    }).css("width", Math.max(5,strength) + "%" ).appendTo(row);
	});
	this.button = jQuery("<button label='closeme'>close</button>").
	    click(function() { this_.close(); }).appendTo(parent);
    },
    close:function() {
	var this_ = this;
	this.getCanvasEl().fadeOut("fast",function() {
	    this_.getCanvasEl().remove();
	    this_.button.remove();
	    this_.enclosing.fadeIn("fast");
	});
    },				  
    _makeRowName:function(evt) { return "row-"+this.id+"-"+this.cmodel.getEntityId(evt); },
    getRowEl:function(val) { return jQuery("#"+this._makeRowName(val)); },
    getCanvasEl:function(cat) { return jQuery("#"+this.id);   }
};

