// Copyright (c) 2008,2009 Massachusetts Institute of Technology
// Copyright (c) 2009 Nokia

// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:

// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.

// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

/**
  Context Maiden WIFI
  Uses FF3.5+'s WiFI watching API to identify revisitations to locations

  Code based on https://developer.mozilla.org/En/Monitoring_WiFi_access_points
**/

__defn_module("context-maiden-wifi",function(JV3) {
		  var plumutil = JV3.plumutil;
		  var newify = JV3.newify;
		  var ContextMaiden = JV3.ContextMaiden;
		  ContextMaiden.WiFiLocation = {
		      id: 'ContextMaiden.WiFiLocation',
		      type:"location",
		      server: "localhost",
		      lastevent : undefined,
		      initialize:function() {
			  this.log = JV3.log;
			  this.hash = JV3.doEncryptActivities();
			  this.set_up_listener();
			  JV3.log("WiFiLocation initialized");	    			  
		      },
		      set_up_listener:function() {
			  var this_ = this;
	    		  this.wifi_service = Components.classes["@mozilla.org/wifi/monitor;1"].getService(Components.interfaces.nsIWifiMonitor);
			  var _convertAPs = function(aps) {
			      var this_ = this;
			      var prepare = function(x) {
				  return this_.hash ? { ssid: JV3.salt_hash(x.ssid), mac: JV3.salt_hash(x.mac), signal: x.signal} : { ssid:""+x.ssid, id:""+x.mac, mac:""+x.mac, signal:""+x.signal  };
			      };
			      return aps.map(prepare);
			  };
			  var wifilistener = function() {};
			  wifilistener.prototype = ({  
							onChange: function(accessPoints)  {
							    if (!JV3.isPrivateBrowsing() && JV3.is_wifi_logging_enabled()) {
								try {
								    let aPs = _convertAPs(accessPoints);
								    JV3.setTimeout(function() { this_._read(aPs); }, 100);			    			    
								} catch (x) {    JV3.log(x);	}
							    }
							},
							onError: function (value) {  this_.error(value); },
							QueryInterface: function(iid) {  
							    if (iid.equals(Components.interfaces.nsIWifiListener) ||  
								iid.equals(Components.interfaces.nsISupports))  
								return this;  
							    throw Components.results.NS_ERROR_NO_INTERFACE;
							}
						    });
			  var wifil_instance = new wifilistener();
			  this.wifi_service.startWatching(wifil_instance);
			  try {
			      let refresh_ival = JV3.prefs.getIntPref("context.wifi.force_refresh");
			      if (refresh_ival > 0) {
				  this._i = JV3.setInterval(function() {
								try {
								    this_.wifi_service.stopWatching(wifil_instance);
								    this_.wifi_service.startWatching(wifil_instance);				   
								} catch (x) {  JV3.log(x);       }			       
							    }, refresh_ival);
			      }
			  } catch (x) { JV3.log(x); }
		      },
		      getLastKnownLocation:function() {
			  return (this.last_event !== undefined && this.last_event.entity !== undefined) ? this.last_event.entity : undefined;
		      },
		      error:function(err) {  JV3.log(err);     },
		      overlap_threshold:0.4,
		      _find_best_match:function(accessPoints,cont) {	    
			  var this_ = this;	    
			  if (accessPoints.length == 0) {
			      // we should fetch unkown
			      JV3.CMS.entity_store.get_or_createAsync({type:"schemas.Location",id:"location_unknown_no_wifi", name:"Somewhere"},function(unknown_loc) { cont([unknown_loc]);});
			      return;
			  }
			  
			  var aps = accessPoints.concat();	    
			  var to_macs = function(xaps) {  return xaps.map(function(x) { return x.mac; }); };		
			  var ap_macs = to_macs(aps);
			  
			  // retrieve matches from context db using sneaky "like" feature
			  var _compute_overlap_with_us = function(t_location) {
			      var loc_macs = to_macs(t_location.aps);
			      var union = JV3.plumutil.uniq(loc_macs.concat(ap_macs));
			      return loc_macs.length + ap_macs.length - union.length;
			  };
			  
			  // FIRST: collect up all known previous locations with visible APs that share at least 1 AP with us
			  JV3.CMS.entity_store.getEntitiesAsync({type:"schemas.Location"},undefined,
								function(matches) {
								    try {							  
									// SECOND: out of these APs, rank by # of APs that overlap
									var overlaps_with_us = matches.map(function(t_location) {
													       if (t_location.aps === undefined) { return 0; }
													       return _compute_overlap_with_us(t_location);
													   });
									var _matches = matches.concat();
									var overlap = function(L) overlaps_with_us[_matches.indexOf(L)];							  
									// sort aps from MOST to LEAST overlapedcx
									matches.sort(function(x,y) overlap(y)-overlap(x));									
									// take the top aps that have (# SSIDs in Common) > Threshold
									var hits = matches.slice(0,matches.map(overlap).filter(function(x) x > this_.overlap_threshold*ap_macs.length).length);
									cont(hits);
								    } catch (x) {  JV3.log(x);      }
								});	
		      },
		      common_threshold: 2,
		      time_smidge:0, // 60*1000,
		      _read:function(accessPoints) {
			  try {
			      var this_ = this;
			      var new_start = new Date();
			      var new_end = new Date(new_start.valueOf() + this.time_smidge);
			      this._find_best_match(accessPoints,
						    function(best_match) {
							try {
							    // JV3.log("BEST MATCH? ", best_match.length > 0 ? best_match[0] : "new location");
							    var best_ap = best_match.length > 0 ? best_match[0] : this_._make_new_location(accessPoints, new_start);
							    // debug
							    //JV3.log("wifi >> using existing? " + (best_match.length > 0));
							    var event = this_.last_event;	    
							    if (event !== undefined && event.entity.id == best_ap.id) {
								//JV3.log("wifi >> extending last event");
								//JV3.log("__wifi__updating existing location : " + event.entity.id);		    
								event.end = new_end;
								// robustify current location's aps by aggregate current observed aps into model
								var new_aps = JV3.plumutil.union(event.entity.aps,accessPoints,function(x) { return x.mac; });
								if (new_aps.length > event.entity.aps) {
								    JV3.log("__wifi__updating existing location with new aps : " + new_aps.length);		    
								    event.entity.aps = new_aps;
								    JV3.CMS.entity_store.saveAsync(event.entity);
								}						  
							    } else {
								JV3.log("wifi >> Creating new event " + new_start);
								event = this_._make_new_event(new_start, new_end, best_ap);
							    }	   
							    this_._geo_update_evt(event);
							    this_.last_event = event;
							    JV3.CMS.event_store.saveAsync(event);
							} catch (x) { JV3.log(x); }
						    });
			  } catch(e) {  JV3.log(e);  }
		      },
		      _geo_update_evt:function() {
			  // abstract. override me with lovely context-maiden-wifi-geo.js
		      },
		      _make_new_event:function(start_date, end_date, location_entity, event_type) {
			  if (event_type === undefined) { event_type = "location"; }
			  return { type: event_type, start: start_date.valueOf(), end:end_date.valueOf(), entity: location_entity };
		      },
		      _make_new_location:function(aPs, creation_date) {
			  var timesig = (creation_date ? creation_date.valueOf() : new Date().valueOf());
			  var id = "Location-"+aPs.map(function(x) { return x.ssid; }).slice(0,3).join("/") + "_" + timesig;
			  return JV3.CMS.entity_store.get_or_create({
									type: "schemas.Location",
									firstspotted:timesig,
									id:id,
									name:id,
									aps:aPs
								    }
								   );
		      }
		  };

		  try {
		      JV3.log("Instantiating WiFiLocation --------------| ");

		      if (JV3.prefs.getBoolPref("context.wifi.geolocation")) {
			  // turn on geolocation
			  try {
			      JV3.log("ENABLING GEOLOCATION");
			      JV3.xuloadCB("chrome://listit/content/context/context-maiden-wifi-geo.js",function(init){ init(JV3,ContextMaiden); });
			  } catch (x) {
			      JV3.log(x);
			  }
		      }	
		      var instance = JV3.newify(ContextMaiden.WiFiLocation);
		      JV3.CMS.sources.push(instance);
		  } catch(E) {
		      JV3.log(E);
		  }		
	      });
