
(function(JV3, schemas) {
    schemas.PairwiseContextTable = {
	/*
           In this abstraction, a _context type_ defines a random variable
           and all the values that context variable can take defines its domain
        */
	superclass:schemas.Entity,	
	id: "schemas.PairwiseContextTable",
	type:"type",
	name:"Pairwise Context Keeper",
	fields: ["id","prior_counts","prior_times","conditional_counts","conditional_times"],
	update_mode: "time",    	
	initialize:function(props) {
	    PROTOCROCK.superclass(this,"initialize", {id:"pairwisept"});
	},
	getInstance:function(CMS) {
	    var this_ = this;
	    var table = CMS.entity_store.get_or_create({type:this.id, id:"pairwisept"});
	    if (table.prior_counts === undefined) {
		JV3.log("Could not locate existing pairwise context table of type " + this.type + " " + this.id );
		table.prior_lu_time={}; // last update times
		table.conditional_counts={};
		table.conditional_times={};
		table.prior_counts={};
		table.prior_times={};
		CMS.entity_store.save(table);
	    } else {
		JV3.log("Located existing pairwise table. using that." + table + table.prior_times);
	    }
	    CMS.event_store.addStoreUpdateListener(undefined, {
		    itemUpdated:function(event) {
			try {
			    //JV3.log("CMS event! " + event.type + " " + table.getEntityId(event) + " " + event.changed +  " " + event.start + " " + ((event.end.valueOf() - event.start.valueOf())/1000.0) + "s");
			    table.update_associations(event);
			} catch (e) {
			    JV3.log("CMS: ERROR in CMS.itemUpdated:"+ e + " " + e.lineNumber + " " + e.fileName);
			}
		    }
		});		
	    return table;
	},
	set: function(value,t,n) {
	    // value is a struture like { type: <context type>,  entity: { id: type } }
	    // prior COUNT, number of times we've seen value of this type
	    var type = this.getEventType(value);
	    var val = this.getEntityId(value);
	    if (this.prior_counts[type] === undefined) { this.prior_counts[type] = {}; }
	    if (this.prior_times[type] === undefined) { this.prior_times[type] = {}; }
	    if (t !== undefined) { this.prior_times[type][val] = t; }
	    if (n !== undefined) { this.prior_counts[type][val] = n; }
	    this.save();
	},
	getCount:function(value) {
	    // value is a struture like { type: <context type>,  entity: { id: type } }
	    var type = this.getEventType(value);
	    var val = this.getEntityId(value);	
	    if (this.prior_counts[type] === undefined || this.prior_counts[type][val] === undefined) {   return 0;	}	    
	    return this.prior_counts[type][val];
	},
	getTime:function(value) {
	    // value is a struture like { type: <context type>,  entity: { id: type } }
	    var type = this.getEventType(value);
	    var val = this.getEntityId(value);	
	    if (this.prior_times[type] === undefined || this.prior_times[type][val] === undefined) {   return 0;	}	    
	    return this.prior_times[type][val];
	},
	getTotalTime:function(type) {
	    return JV3.plumutil.sum( JV3.plumutil.objVals(this.prior_times[type]) );
	},
	getTotalCount:function(type) {
	    return JV3.plumutil.sum( JV3.plumutil.objVals(this.prior_counts[type]) );
	},
	getPTime:function(value) {
	    // value is a struture like { type: <context type>,  entity: { id: type } }
	    var type = this.getEventType(value);
	    return this.getTime(value)/this.getTotalTime(type);
	},
	getPCount:function(value) {
	    // value is a struture like { type: <context type>,  entity: { id: type } }
	    var type = this.getEventType(value);
	    return this.getCount(value)/this.getTotalCount(type);
	},
	getTypes:function() { return JV3.plumutil.objKeys(this.prior_counts);   },
	reorder:function() {
	    var vala_evttype = this.getEventType(arguments[0]);
	    var valb_evttype = this.getEventType(arguments[1]);
	    JV3.plumutil.assert(vala_evttype !== valb_evttype, "Dont do this to me.");
	    return (vala_evttype < valb_evttype) ? [arguments[0],arguments[1]]:[arguments[1],arguments[0]];
	},
	makeValKey:function(val1,val2) {
	    var id1 = this.getEntityId(val1), id2 = this.getEntityId(val2);
	    return JV3.plumutil.toJSON([id1,id2]);    
	},
	decodeValKey:function(valkey) {
	    return JV3.plumutil.fromJSON(valkey);
	},
	getCoCount:function(vala,valb) {
	    var values = this.reorder.apply(this,arguments); 
	    var val1 = values[0];
	    var val2 = values[1];
	    var val1type = this.getEventType(val1);
	    var val2type = this.getEventType(val2);
	    var valkey = this.makeValKey(val1,val2); 
	    try {
		return this.conditional_counts[val1type][val2type][valkey] ? this.conditional_counts[val1type][val2type][valkey] : 0;
	    } catch(e) {
		// we dereferenced an undefined. oops!
		// JV3.log(e); // debug, we dont need to worry about it
	    }
	    return 0;
	},
	getCoTime:function(vala,valb) {
	    var values = this.reorder.apply(this,arguments); 
	    var val1 = values[0];
	    var val2 = values[1];
	    var val1type = this.getEventType(val1);
	    var val2type = this.getEventType(val2);
	    var valkey = this.makeValKey(val1,val2); 
	    try {
		// warning: if you cahnge this you should also change the dump output function
		// in context-maiden-scheduler.js
		return this.conditional_times[val1type][val2type][valkey] ? this.conditional_times[val1type][val2type][valkey] : 0;
	    } catch(e) {
		// we dereferenced an undefined. oops!
		// JV3.log(e); // debug, we dont need to worry about it
	    }
	    return 0;
	},    
	setCo:function(vala,valb,t,n) {
	    var values = this.reorder.apply(this,arguments);
	    var val1 = values[0];
	    var val2 = values[1];
	    var val1type = this.getEventType(val1);
	    var val2type = this.getEventType(val2);	
	    var valkey = this.makeValKey(val1,val2);
	    if (this.conditional_counts[val1type] === undefined) { this.conditional_counts[val1type] = {}; }
	    if (this.conditional_counts[val1type][val2type] === undefined) { this.conditional_counts[val1type][val2type] = {}; }
	    if (this.conditional_times[val1type] === undefined) { this.conditional_times[val1type] = {}; }	
	    if (this.conditional_times[val1type][val2type] === undefined) { this.conditional_times[val1type][val2type] = {}; }
	    if (t !== undefined) { this.conditional_times[val1type][val2type][valkey] = t; }
	    if (n !== undefined) { this.conditional_counts[val1type][val2type][valkey] = n; }
	    this.save();
	},
	getConditionalPCount:function(val,givens,minval) {	
	    // value is a struture like { type: <context type>,  entity: { id: type } }
	    // givens is a struture like [ { type: <context type>,  entity: { id: type } } ] or { context-type: {id:type} }
	    var this_ = this;

	    // compute p(x | given) for all givens
	    var conditional = givens.map(function(given) {
		var overlap = this_.getCoCount(val,given)/2.0;
		var given_n = this_.getCount(given);
		if (given_n === 0) { this_.log("Given : " + given + " was null"); return 0; }
		if (overlap === 0 && minval) { overlap = minval; }
		return overlap*1.0/given_n;
	    }).reduce(function(x,y) { if (y === undefined) { return x; } return x*y; });
	    return conditional; // * this_.getPriorP(val);
	},
	getConditionalPTime:function(val,givens,minval) {	
	    // value is a struture like { type: <context type>,  entity: { id: type } }
	    // givens is a struture like [ { type: <context type>,  entity: { id: type } } ] or { context-type: {id:type} }
	    var this_ = this;

	    // compute p(x | given) for all givens
	    var conditional = givens.map(function(given) {
		var overlap = this_.getCoTime(val,given)/2.0;
		var given_T = this_.getTime(given);
		if (given_T === 0) { this_.log("Given : " + given + " was 0"); return 0; }
		if (overlap === 0 && minval) { overlap = minval; }
		return overlap*1.0/given_T;
	    }).reduce(function(x,y) { if (y === undefined) { return x; } return x*y; });
	    return conditional; // * this_.getPriorP(val);
	},
	save:function() {
	    var entity_store = JV3.CMS.entity_store;
	    entity_store.save(this);
	},
	getEntityId: function(x) {
	    if (typeof x == "object" && x.entity) { return this.getEntityId(x.entity.id); }
	    if (typeof x == 'object' && x.entityid) { return this.getEntityId(x.entityid); }
	    if (typeof x == "string") { return x; }
	    if (typeof x == 'number') { return ""+x; }
	    try {
	    JV3.plumutil.assert(false, "Cannot get id for object : " + x + " " + JV3.plumutil.toJSON(x));
	    } catch(e) {
		JV3.log(e+" "+e.fileName+":"+e.lineNumber+" "+e.stack);
	    }
	},
	getEntityType: function(x) {
	    if (typeof x == "object" && x.entity) { return x.entity.type; }
	    if (typeof x == 'object' && x.entitytype) { return x.entitytype; }
	    JV3.plumutil.assert(false, "Cannot get type for object : " + x + " " + JV3.plumutil.toJSON(x));
	},
	getEventType: function(x) {	return x.type;  },
	update_associations:function(source_event) {
	    // this deserves some explanation:
	    // either source_event is a "note-edit" event OR some OTHER event
	    // "note-edit" events do not come from the event store -- they come directly from noteeventsource
	    // all others comef rom the event store.
	    JV3.log("update_associations yo-- type : " + source_event.type + " changed? " + source_event.changed + " " +
		     source_event.start.valueOf() + " " + source_event.end.valueOf() + " " +  (source_event.entity && source_event.entity.id));
	    var entity_store = JV3.CMS.entity_store;
	    var event_store = JV3.CMS.event_store;
	    var pstats = this;
	    var TS = JV3.ContextMaiden.TimeStruct;
	    var this_ = this;
	    var time_padding = 2500;
 	    var events = undefined;
	    var update_mode = this.update_mode;
	    var util = JV3.plumutil;
	    var log = JV3.log;

	    // handle correlations for time	
	    // update p(event|Note)
	    JV3.plumutil.assert(source_event.changed !== undefined, "Source event changed is undefined : " + source_event.type);
	    JV3.plumutil.assert(source_event.changed == true || source_event.timeextension >= 0 , "Timeextension was " +
				source_event.timeextension + " for " + pstats.getEntityId(source_event));

	    var get_intersecting_events = function(source_interval) {
		var where_clause = "start <= " + source_interval.end.valueOf() + " AND end >= " + source_interval.start.valueOf();
		JV3.log(where_clause);
		events = event_store.getEvents( undefined, where_clause );
		return events;
	    };
	    if (source_event.changed) {  // new event!
		// we want to COUNT an update, and full intersection
		var source_interval = JV3.ContextMaiden.TimeStruct.makeDuration(source_event.start.valueOf(),source_event.end.valueOf(),time_padding);		
		// set prior == 
		pstats.set( source_event, pstats.getTime(source_event) + source_interval.length(), pstats.getCount(source_event) + 1 );
		JV3.log("__________________" + get_intersecting_events(source_interval).length + " " + source_interval.length());
		get_intersecting_events(source_interval).map(function(evt) {
								 try {
								     // - JV3.log("&&&&&& ");
								     if (evt.type == source_event.type) { return; /* skip self */ }
								     var intersection = TS.makeDuration(evt.start,evt.end).intersection(source_interval);
								     // - change to note_interval if you want to only intersect with the REAL note edit interval
								     JV3.plumutil.assert(intersection, "Intersection undefined! :( ");
								     JV3.log("updating " + (source_event.entityid || source_event.entity &&
											    source_event.entity.id) + "-" +
									     (evt.entityid || evt.entity && evt.entity.id) + " " +
									      pstats.getCoCount(source_event,evt) + " " + " .. " +
									      pstats.getCoTime(source_event,evt) + " " );
								     
								     pstats.setCo( source_event, evt,
										   pstats.getCoTime(source_event,evt) + intersection.length(),
										   pstats.getCoCount(source_event,evt) + 1);								     	     
								 } catch (x) {  JV3.log(x);   }
							     });

	    } else {
		// update: do NOT update counts,
		var source_interval =
		    JV3.ContextMaiden.TimeStruct.makeDuration(source_event.end.valueOf() - source_event.timeextension, source_event.end.valueOf());
		pstats.set( source_event, pstats.getTime(source_event) + source_interval.length(), pstats.getCount(source_event));
		get_intersecting_events(source_interval).map(function(evt) {
								 try {
								     if (evt.type == source_event.type) { return; /* skip self */ }
								     var intersection = TS.makeDuration(evt.start,evt.end).intersection(source_interval);
								     // change to note_interval if you want to only intersect with the REAL note edit interval
								     JV3.plumutil.assert(intersection, "Intersection undefined! :( ");
								     pstats.setCo( source_event, evt,
										   pstats.getCoTime(source_event,evt) + intersection.length(),
										   pstats.getCoCount(source_event,evt));
								     
								 } catch (x) {  JV3.log(x); }
							     });
	    }
	    pstats.save();
	    this._refresh_table_window();
	    return;
	},	
	open_table_window:function() {
	    var win = JV3.plumutil.open_new_window("chrome://listit/content/listit/debugwindow.html",700,600);
	    this.twindow = win;
	},
	dumpStats: function() {
	    var s = "";
	    s += "Conditional counts:\n";
 	    util.objKeys(pstats.conditional_counts).map(function(key1) {
							    var i1 = "  ";
							    s += i1+key1+"\n";
							    util.objKeys(pstats.conditional_counts[key1]).map(function(key2) {
														  s += i1+i1+key2+"\n";
														  util.objKeys(pstats.conditional_counts[key1][key2]).map(
														      function(row) {
															  s += i1+i1+i1+row+": "+
															      pstats.conditional_counts[key1][key2][row] + "\n";
														      });
													      });
							});
	    s += "\n";
	    s += "Conditional times: (updated " + new Date().valueOf() + ") \n";
	    util.objKeys(pstats.conditional_times).map(function(key1) {
							   var i1 = "  ";
							   s += i1+key1+"\n";
							   util.objKeys(pstats.conditional_times[key1]).map(function(key2) {
														s += i1+i1+key2+"\n";
														util.objKeys(pstats.conditional_times[key1][key2]).
														    map(function(row) {
															    s += i1+i1+i1+row+": "+pstats.conditional_times[key1][key2][row]/2.0 + "\n";
															});

													    });
						       });
	    return s;								  
	},	
	_refresh_table_window: function() {
	    var this_ = this;
	    var pstats = this;
	    var util = JV3.plumutil;
	    if (this.twindow) {
		JV3.log("Updating table window");
		var win = this.twindow;
		win.document.getElementById("dump").innerHTML = this.dumpStats() ;
	    }
	}
    };
})


        
/*

    _getCorrelatingEvents:function(val) {
	// returns p(val | C1=c11), p(val|C1=c12), .... for all contextual elements. //
	
	var this_ = this;
	var types = this.getTypes();
	var results = {};
	var valOverlapTotal = 0;
	types.map(function(type) {
	    try {
		results[type] = [];
		if (type == this_.getEventType(val)) { return; } // only want results that are correlations with values of different types
		var tordered = this_.reorder(val,{type:type});
		var ttype0 = this_.getEventType(tordered[0]);
		var ttype1 = this_.getEventType(tordered[1]);
		var valindex = tordered.indexOf(val);
		var wantindex = (valindex+1)%2;
		
		if (this_.conditional_counts &&
		    this_.conditional_counts[ttype0] &&
		    this_.conditional_counts[ttype0][ttype1]) {
		    this_.util.objKeys(this_.conditional_counts[ttype0][ttype1]).map(function(valkey){
			// we're iterating through all the comibinations of <type1>,<type2> in this conditional table
			// looking for ours
			var decoded = this_.decodeValKey(valkey);
			var valid = decoded[valindex];
			var wantid = decoded[wantindex];
			if (valid ==  this_.getEntityId(val)) { // then this is the row we've been looking for
			    var count = this_.conditional_counts[ttype0][ttype1][valkey];
			    // valOverlapTotal += count;
			    results[type].push({ type: type, entity: { id : decoded[wantindex] }, conditionalCount: count });
			}
		    });
		}
	    }catch(e) { this_.log("ERROR_getCorrelatingEvents " + e + " " + e.lineNumber + " " + e.fileName);  }
	});

	valOverlapTotal = this_.getPriorP(val);
	results = JV3.plumutil.objMap(function(x) {
	    return x.map(function(t) {
		t.conditionalP = t.conditionalCount * 1.0/valOverlapTotal;
		delete t.conditionalCount;
		return t;		
	    });
	},results);
	return results;
    },
*/

/*&&

    getPriorDistributionMap:function(event_type){
	// returns an object
	if (this.priors[event_type] === undefined) { return {}; }
	var total = JV3.plumutil.sum( JV3.plumutil.objVals(this.priors[event_type]) );
	return JV3.plumutil.objMap(function(x) {
	    return x*1.0/total;
	},this.priors[event_type]);
    },
    getPriorDistributionArray:function(event_type){
	// returns an object
	if (this.priors[event_type] === undefined) { return []; }
	var total = JV3.plumutil.sum( JV3.plumutil.objVals(this.priors[event_type]) );
	this.priors[event_type].objVals(function(x) {
	    return x * 1.0/total;
	});
    },
    getConditionalDistributionMap:function(event_type1,event_type2) {
	var event_types = this.reorder.apply(this,[{type:event_type1},{type:event_type2}]);
	event_type1 = event_types[0];
	event_type2 = event_types[1];
	throw new Error("Not implemented fully yet.");
    },   






    getConditionalP2:function(val,givens) {	
	// value is a struture like { type: <context type>,  entity: { id: type } }
	// givens is a struture like [ { type: <context type>,  entity: { id: type } } ] or { context-type: {id:type} }
	var this_ = this;
	givens = this.toGivenMap(givens);
	JV3.plumutil.assert(JV3.plumutil.objKeys(givens).length > 0, "Some givens must be provided");
	return JV3.plumutil.objKeys(givens).map(function(t) {
	    if (this_.getEventType(val) == t) { return 1.0; }
	    return this_._getCorrelatingP(val,givens[t])/this_.getPriorP(val);
	}).reduce(function(x,y) { if (y === undefined) { return x; } return x*y; });	
    },
    _getCorrelatingP:function(val,val2) {
	var this_ = this;
	var results = {};
	var events2  = this._getCorrelatingEvents(val);
	if (events2 && events2[this.getEventType(val2)]) {
	    //JV3.log("!!!!!!!!!!!!!!!!!! looking at events type success  " + this.getEventType(val2) + " now looking for id " + this.getEntityId(val2));
	    var colmatches = events2[this.getEventType(val2)].filter(function(x) { return this_.getEntityId(x) == this_.getEntityId(val2); });
	    if (colmatches && colmatches.length > 0) {
		//JV3.log("Correlating p for " + this_.getEntityId(val) + " " + this_.getEntityId(val2) + " " + colmatches[0].conditionalP);
		return colmatches[0].conditionalP;
	    }
	};
	//JV3.log("NO correlating p for " + this.getEntityId(val) + " " + this.getEntityId(val2));
	return 0;
    },






*/


/**
    getConditionalPOld:function(val,givens,fill_in_missings) {
	
	// value is a struture like { type: <context type>,  entity: { id: type } }
	// givens is a struture like [ { type: <context type>,  entity: { id: type } } ] or { context-type: {id:type} }
	givens = this.toGivenMap(givens);
	var types = fill_in_missings ? this.getTypes() : JV3.plumutil.objKeys(givens);
	JV3.plumutil.assert(types.length > 0, "Some givens must be provided");
	var this_ = this;
	return types.map(function(t) {
	    if (this_.getEventType(val) == t) { return 1.0; }
	    if (givens[t]) {
// 		this_.log(">>");
// 		this_.log("cocount between " + val.entity.id + " and " + givens[t].entity.id + " " + this_.getCoCount(val,givens[t]));
// 		this_.log("prior count of  givens[t]" + givens[t].entity.id + " :: " + this_.getCount(givens[t]));
// 		this_.log("prior probability of givens[t]" + givens[t].entity.id + " :: " + this_.getPriorP(givens[t]));
// 		this_.log("<<");
		//return this_.getCoCount(val,givens[t])/this_.getCount(givens[t]) * this_.getPriorP(givens[t]);
		// this computes p(given|val)
		return this_._getCorrelatingP(val,givens[t])/this_.getPriorP(val);
	    }
	    this_.util.assert(this_.priors[t] , "No domain for missng type : " + t);
	    this_.log("Missing context " + t + " // what do we do now?");
	    return Math.pow(Math.min.apply(this, this_.util.objVals(this_.priors[t])),2);  // ]) 1.0/D * this_.util.sum( this_.util.objVals( this_.priors[t] ) );	    
	}).
	    reduce(function(x,y) { if (y === undefined) { return x; } return x*y; });	
    },

    getCorrelatingEventsFor:function(val) {
	var this_ = this;
	var types = this.getTypes();
	var results = {};
	var valCount = this_.getCount(val);
	types.map(function(type) {
	    try {
		this_.log("TYPE : " + type);
		results[type] = [];
		var tordered = this_.reorder(val,{type:type});
		var ttype0 = this_.getEventType(tordered[0]);
		var ttype1 = this_.getEventType(tordered[1]);
		var valindex = tordered.indexOf(val);
		var wantindex = (valindex+1)%2;


//		this_.log("ttype0 : " + ttype0 + " " +ttype1 + " " + this_.conditionals + " " + (this_.conditionals[ttype0] ? this_.conditionals[ttype0]: " conditionalsUNDEFINED") );
//		this_.log("!!!! " + (this_.conditionals[ttype0] && this_.conditionals[ttype0][ttype1] ? this_.util.objKeys(this_.conditionals[ttype0][ttype1]) : " conditionalsUNDEFINED") );
//		if (this_.conditionals[ttype0] && this_.conditionals[ttype0][ttype1]) {
//		    this_.log("JV3.FOO SET");
//		    this_.JV3.FOO = this_.conditionals[ttype0][ttype1];
//		}

		if (this_.conditionals &&
		    this_.conditionals[ttype0] &&
		    this_.conditionals[ttype0][ttype1]) {
		    this_.util.objKeys(this_.conditionals[ttype0][ttype1]).map(function(valkey){
			var decoded = this_.decodeValKey(valkey);
			this_.log("uh huh : " + valkey + " " + decoded[0] + " " + decoded[1] + " " + valindex + " " + val.entity.id );			
			var count = this_.conditionals[ttype0][ttype1][valkey];
			if (decoded[valindex] == val.entity.id) {
			    this_.log("uh huh YEAH BABY " + decoded[valindex] + " "  + count * 1.0/ valCount);
			    var conditionalP = count * 1.0/ valCount; // conditional probability of this given val
			    results[type].push({ type: type, entity: { id : decoded[wantindex] }, conditionalP:conditionalP });			    
			}
		    });
		}
	    }
	    catch(e) {
		this_.log("ERROR_getCorrelatingEvents " + e + " " + e.lineNumber + " " + e.fileName);
	    }
	});
	return results;
    },


*/