
/**
 Firewatcher - Page grabbing and user interaction monitoring for Firefox 3.0+

 This module listens for page loads and dispatches notifications to whatever
 or whomever might be listening.  It also has an overridable handleEvent()
 method that you can use to directly handle all incoming events.  Does
 appropriate filtering for event targets to prevent duplicated load/focus
 events [caused by embedded elements such as iframes]
 
 (c) 2007-2009 by electronic Max

 **/ 

(function(JV3,ContextMaiden) {
    ContextMaiden.FWS = {
	host:"localhost",
	plumutil:JV3.plumutil,
	log:JV3.log,
	port:4444,
	gbrowser:undefined,
	outstream:undefined,
	contents : false,
	mediator:undefined,
	TAB_SWITCHED:"Tab Changed",
	WINDOW_FOCUSED:"Window Focused",
	WINDOW_UNFOCUSED:"Window Unfocused",
	PAGE_LOADED:"Page Loaded",
	PAGE_UNLOADED:"Page Unloaded",
	USER_IDLE:"User idle",
	USER_AWAKE:"User awake",        
	handleEvent : function(X) { }, // replaceme!
	
	initialize: function() {
	    //  global thingies
	    var this_ = this;
            this.wins = [];
            this.listeners = [];            
            this.instance = this;
	    this.wm = Components.classes["@mozilla.org/appshell/window-mediator;1"].getService(Components.interfaces.nsIWindowMediator);
	    try {
		this.windowListener = {
		    onOpenWindow : function(xulwin) {
			try {
			    JV3.log("~~~~~ open window cb => ");
			    var Ci = Components.interfaces;
			    var chromeWindow = xulwin.docShell.QueryInterface(Ci.nsIInterfaceRequestor).getInterface(Ci.nsIDOMWindow);
			    // browser only accessible once the window has loaded, so we register for that event
			    chromeWindow.addEventListener( "pageshow", function() {
							       JV3.log("_______________________fire watcher page show >> adding listeners ");
							       chromeWindow.removeEventListener("pageshow", arguments.callee, false);
							       this_.addListeners(chromeWindow);
							   }, false ); 

			} catch(e) {
			    this_.log(e + " " + e.lineNumber + " " + e.fileName);
			}			    
		    },
		    //setTimeout(plumutil.partial(_FWS.enumerate_all_windows,_FWS),2000);  },
		    onCloseWindow: function(xulwin) {
			try { 
			    this_.log("firewatcher.closeWindow: deregistering listeners >> ");
			    var Ci = Components.interfaces;
			    var chromeWindow = xulwin.docShell.QueryInterface(Ci.nsIInterfaceRequestor).getInterface(Ci.nsIDOMWindow);
			    this_.removeListeners(chromeWindow);
			} catch(e) {
			    this_.log(e + " " + e.lineNumber + " " + e.fileName);
			}
		    },
		    onWindowTitleChange: function(window) { }
		};
		this.wm.addListener(this.windowListener);
	    } catch(e2) {
		this_.log(e2 + " " + e2.fileName + " " + e2.lineNumber);
	    }

	    // very reluctant to do this, but necessary
	    this_.enumerate_all_windows();
	    this_.tval = JV3.setTimeout(function() { this_.enumerate_all_windows(); }, 2000);
	    this_.ival = JV3.setInterval(function() { this_.enumerate_all_windows(); }, 10000);	    
	},
	shutdown: function() {
	    this.wm.removeListener(this.windowListener);
	    this.removeAllListeners();
	},
	enumerate_all_windows: function() {
	    try {
		var enumerator = this.wm.getZOrderDOMWindowEnumerator("navigator:browser",true);
		var this_ = this;
		var count = 0;
		while (enumerator.hasMoreElements()) {
		    try {
			var w = enumerator.getNext();
			// JV3.log("enumerator found  : " + w);
			this_.addListeners(w);
			count++;
		    } catch(e) {
			this_.log(e + " " + e.lineNumber + " " + e.fileName);	
		    }
		}
		// JV3.log("enumerated " + count + " windows");
	    } catch (x) {
		JV3.log(x);
	    }
	},    
	cb_pageLoaded: function(FWS,win,e) {
	    //FWS.log("documenURI?: " + e.target.documentURI);
	    if (!e.target.documentURI) { return; }
	    //FWS.log("pageload target is " + e.target + " " + keys(e.target) + " " + e.target.documentURI); //" currentTarger " + e.currentTarget + " " + keys(e.currentTarget));
	    FWS.getPages(FWS,win,FWS.PAGE_LOADED,e);
	},
	cb_pageUnloaded: function(FWS,win,e) {
	    if (!e.target.documentURI) return;
	    //FWS.log("UNLOAD target is " + e.target + " " + keys(e.target) + " " + e.target.documentURI); //" currentTarger " + e.currentTarget + " " + keys(e.currentTarget));
	    FWS.getPages(FWS,win,FWS.PAGE_UNLOADED,e);
	},
	cb_tabChanged: function(FWS,win,e) {
	    //FWS.log("TABCHANGE target is " + e.target + " " + keys(e.target) + " " + e.target.documentURI); //" currentTarger " + e.currentTarget + " " + keys(e.currentTarget));
	    FWS.getPages(FWS,win,FWS.TAB_SWITCHED,e);
	},
	cb_windowFocused: function(FWS,win,e) {
	    FWS.getPages(FWS,win,FWS.WINDOW_FOCUSED,e);
	    JV3.informObservers("window-focused",JV3.getWindowId(win));
	},
	cb_windowUnfocused: function(FWS,win,e) {
	    FWS.getPages(FWS,win,FWS.WINDOW_UNFOCUSED,e);
	    JV3.informObservers("window-unfocused",JV3.getWindowId(win));
	},
	// for the user //
	cb_userIdle:function(FWS) {
	    JV3.log("user-idle");
	    FWS.getPages(FWS,undefined,FWS.USER_IDLE,FWS.IDLE_DURATION);
	    JV3.informObservers("user-idle");
	},
	cb_userAwakeFromIdle:function(FWS,win) {
	    FWS.getPages(FWS,win,FWS.USER_AWAKE,(new Date().valueOf())-FWS.lastUserAwake.valueOf());
	    JV3.informObservers("user-awake",JV3.getWindowId(win));
	},
	// window by window basis //
	cb_windowIdle:function(win) {
	    JV3.informObservers("window-idle",JV3.getWindowId(win));
	},
	cb_windowAwake:function(win){
	    JV3.informObservers("window-awake",JV3.getWindowId(win));
	},
	
	POLLER_INTERVAL:10000,
	IDLE_DURATION:60000, // 20 seconds = idle
	lastUserAwake:new Date(),
	last_awake_by_window: [],

	_cb_idle_Poller_Alarm:function() {
	    try {
		var FWS = this;
		var now = new Date().valueOf();
		
		// alarm fires. check to see if we should change the state of the user
		if (this.lastUserAwake === undefined || this.lastUserAwake === "idle") { /* already idle, do nothing. */ return;  }
		if (now - this.lastUserAwake > FWS.IDLE_DURATION) {
		    // user died.
		    this.cb_userIdle(FWS);
		    this.lastUserAwake = "idle";
		}		
		// how about any windows?
		this.wins.map(function(win,i) {
				  // if it's been threshold while since we've seen a window, we mark it as idle
				  if ( typeof(FWS.last_awake_by_window[i]) == "number" && (now - FWS.last_awake_by_window[i]) > FWS.IDLE_DURATION) {
				      // window went idle
				      FWS.last_awake_by_window[i] = "idle";
				      FWS.cb_windowIdle(win);
				  }
			      });	    		 
	    } catch (x) { JV3.log(x);  }	    
	},
	_cb_resetidle:function(FWS,win) {
	    try {
		// deal with the user
		var this_ = this;
		var now = new Date().valueOf();
		if (this.lastUserAwake === "idle") { this.cb_userAwakeFromIdle(this,win); }
		this.lastUserAwake = now;
		// now for each browse
		var winindex = this.wins.indexOf(win);		
		JV3.plumutil.assert(winindex >= 0, "Could not find window in wins");
		if (this.last_awake_by_window[winindex] == 'idle') { this.cb_windowAwake(win); }
		this.last_awake_by_window[winindex] = now;

		if (this._idlepoller === undefined) {
		    JV3.log("starting idlepoller");
		    this._idlepoller = JV3.setInterval(function() { this_._cb_idle_Poller_Alarm(); }, this.POLLER_INTERVAL);
		}		
	    } catch(e) {JV3.log(e);}
	},
	
	addListeners:function(w) {
	    if ((this.wins.indexOf(w) < 0) && w.getBrowser !== undefined) {
		JV3.log("eyebrowse detected a new window ~ registering ~ " + w);
		this.listeners.push(this._addListeners(w)); // wm.getMostRecentWindow("navigator:browser"));
		this.wins.push(w);
	    }
	},
	_addListeners : function(win) {
	    // debug
	    // FWS.log('registering listeners for window ' + window);
	    // 	var props = "";
	    // 	for (p in window) {   props += ", "+p;	}
	    // 	FWS.log('properties of window  ' + props);
	    // 	FWS.log(" getBrowser " + window.getBrowser());
	    // end debug
	    try {
		var FWS = this;
		var browser = win.getBrowser();
		var listeners = [
		    { event: "pageshow", to: browser, listener:  function(evt) {
			try { FWS.cb_pageLoaded(FWS,win,evt);  } catch(e) {  FWS.log(e + " " + e.lineNumber + " " + e.fileName);}}, value: false },
		    //		    { event: "pagehide", to: browser, listener: function(evt) {
		    //			try { FWS.cb_pageUnloaded(FWS,win,evt);  } catch(e) {  FWS.log(e + " " + e.lineNumber + " " + e.fileName);   }}, value: true },
		    { event: "TabSelect", to: browser.tabContainer, listener: function(evt) {
			try { FWS.cb_tabChanged(FWS,win,evt);  } catch(e) {  FWS.log(e + " " + e.lineNumber + " " + e.fileName);   } }, value: false },
		    { event: "focus", to: win, listener: function(evt) {
			  try { FWS.plumutil.partial(FWS.cb_windowFocused,FWS,win)(evt);} catch(e) { FWS.log(e + " " + e.lineNumber + " " + e.fileName);}}, value: true},
		    { event: "blur", to: win, listener: function(evt) {
			try { FWS.cb_windowUnfocused(FWS,win,evt); } catch(e) { FWS.log(e + " " + e.lineNumber + " " + e.fileName);}}, value: true},
		    { event: "mousemove", to: win, listener: function() { FWS._cb_resetidle(FWS,win); }, value: false},
		    { event: "mousedown", to: win, listener: function() { FWS._cb_resetidle(FWS,win); }, value: false},
		    { event: "keydown", to: win, listener: function() { FWS._cb_resetidle(FWS,win); }, value: false}
		];
		
		listeners.map(function(x) {
		    x.to.addEventListener( x.event, function() { x.listener.apply(x,arguments); return false; }, true ) ; //x.value ); 
		});

		return listeners;
	    } catch(e) {
		this.log("noncritical warning: " + e + " " + e.lineNumber + " " + e.fileName);
	    }
	},
	removeListeners: function(chromeWindow) {
	    try { 
		    var this_ = this;
		    var index = this.wins.indexOf(chromeWindow);
		    if (index >= 0) {
		        // this_.log("removing listeners : " + this_.listeners[index].length );
		        this.listeners[index].map(function(x) {
			                                  // this_.log("removing event : " + x.event + " " + x.listener + " " + x.value);		    
			                                  try {
			                                      x.to.removeEventListener( x.event, x.listener, x.value );
			                                  } catch(e) {
			                                      this_.log(e);
			                                  }
		                                  });		    
		        this_.plumutil.removeIndices(this_.wins,index,index);
		        this_.plumutil.removeIndices(this_.listeners,index,index);
		    } else {
		        this_.log("Could not find chromeWindow in stack : " + chromeWindow);
		    }
	    } catch(e2) {
		    this.log(e2 + " " + e2.lineNumber + " " + e2.fileName);
	    }
	},
	removeAllListeners:function() {
	    // removes all listeners listened to by this
	    var wins = this.wins.concat();
	    var this_= this;
	    wins.map(function(x) {
		this_.removeListeners(x);
	    });
	},    
	getPages: function(FWS,w,eventType,event) {
	    try {
		if (w && w.getBrowser !== undefined && w.getBrowser !== null)  {
		    try {
			// this is a browser window, so we can query its guts
			var tabBrowser = w.getBrowser(); //window.getBrowser();	// w is your chrome window
			var tab = tabBrowser.mTabBox.selectedTab;
			var browser = tabBrowser.getBrowserForTab(tab);
			var htmlWindow = browser.contentWindow;

			var location = undefined;
			var host = undefined;
			var title = undefined;


			// iniital filtering to get rid of things that dont have us as a target
			try {
			    if ([FWS.USER_AWAKE, FWS.USER_IDLE].indexOf(eventType) < 0 && event !== undefined && event.target !== htmlWindow.document) {
				//if (JV3.ETGT == undefined) {    JV3.ETGT = [];	} JV3.ETGT.push(event.target);
				return;
			    }
			    /*else { JV3.TGT = event;   } */
			}  catch(e) {
			    JV3.log(e);
			}

			try {
			    location = htmlWindow.document.location.href; 
 			    host =  htmlWindow.document.location.host ;
			    title =  htmlWindow.document.title ;
			} catch(e3) {
			    //FWS.log("YO"+ e3 + " " + e3.lineNumber + " " + e3.fileName);
			}

			// filter for iframe events
			if ([this.PAGE_LOADED, this.PAGE_UNLOADED].indexOf(eventType) >= 0 && event && event.target && location != event.target.documentURI) {
			    // we got a page load even for an embedded iframe, etc. yuck!
			    return;
			}
			
			this.handleEvent(eventType,location,host,title,
					 (eventType == this.PAGE_LOADED && this.contents) ? this.traverseTextNodes(htmlWindow.document.documentElement) : undefined,
					 event);			
			
		    } catch (e2) {
			// various errors may occur if things arent completely loaded yet,
			// don't freak out
			FWS.log(e2 + " " + e2.lineNumber + " " + e2.fileName);
		    }
		} else {
		    FWS.log("FWS: logging nonstandard event : " + eventType );
		    this.handleEvent(eventType,undefined,undefined,undefined,undefined,event);
		}
	    } catch(e) {
		this.log(eventType + " " + event + " " + e + " " + e.lineNumber + " " + e.fileName + " " + w);
	    }
	},
	traverseTextNodes : function(parent) {
	    if (!parent) return "";
	    var s = "";
	    var children = parent.childNodes;
	    for (var i = 0; i < children.length; i++) {
		var child = children[i];
		if (child.nodeType == child.ELEMENT_NODE && child.nodeName.toLowerCase() == 'script') {
		    // skip skripts
		    continue;
		} else if (child.nodeType == child.ELEMENT_NODE || child.nodeType == child.DOCUMENT_NODE) {
		    s = s + traverseTextNodes(child);
		} else if (child.nodeType == child.TEXT_NODE)  {
		    s = s + child.textContent;
		}
	    }
	    return s;
	}
    };
});
