// Copyright (c) 2008 Nokia
// Copyright (c) 2008 Massachusetts Institute of Technology

// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:

// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.

// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

(function(ContextMaiden) { 
    var WatchmeStore = {
	superclass: PlumStore,
	SCHEMA : "(id integer(100), type varchar, location varchar, host varchar, title varchar)",
	plumutil:plumutil,
	initialize: function(params) {
	    this.super.initialize(params);
	    this.table = params.table;
	    this.log('initializing with table ' + this.table + " and debug " + this.DEBUG );
	    try {
		this.createTable(this.table);
	    } catch(e) { /* error might occur if table already exists */ }
	},
	save : function (obj) {
	    var id = obj.id;
	    if (typeof(id.valueOf()) == 'string') { id = parseInt(id); }
	    var c = this._sqlfromVal;
	    var values = "("+c(id)+","+c(obj.type)+","+c(obj.location)+","+c(obj.host)+","+c(obj.title)+")";
	    var sqlstmt = 'INSERT INTO ' + this.table + ' VALUES ' + values + ';';
	    return this.execute(sqlstmt);
	},
	_makeEntry:function(id,type,loc,host,title) {
	    this.plumutil.assert((id !== undefined && type != undefined), "Id or type null " + id + " " + type );
	    var obj = {id:id,type:type};
	    if (loc != undefined) obj.location = loc;
	    if (host != undefined) obj.host = host;
	    if (title != undefined) obj.title = title;
	    return obj;
	},
	_sqlfromVal:function(val) {
	    if (val == undefined) return "NULL";
	    if (typeof(val.valueOf()) == 'number') return val.valueOf();
	    if (typeof(val.valueOf()) == 'string') return "\""+encodeURIComponent(val.valueOf())+"\"";
	    this.plumutil.assert(false, "cannot convert " + val);
	},
	_valfromSQL:function(statement,idx) { 
	    if (statement.getIsNull(idx) ||  statement.getTypeOfIndex(idx) == statement.VALUE_TYPE_NULL)
		return undefined;
	    if (statement.getTypeOfIndex(idx) == statement.VALUE_TYPE_INTEGER)
		return statement.getInt64(idx);
	    if (statement.getTypeOfIndex(idx) == statement.VALUE_TYPE_STRING || statement.getTypeOfIndex(idx) == statement.VALUE_TYPE_TEXT)
		return decodeURIComponent(statement.getString(idx));
	    this.log("Warning: could not handle parameter at index " + idx);
	    return undefined;
	},
	loadbyID : function(id) {
	    var sqlstmt = 'SELECT * from ' + this.table + ' WHERE id=' + this._sqlfromVal(id) + ';';
	    this.log(sqlstmt);
	    var db = this._db;
	    var statement = db.createStatement(sqlstmt);
	    var results = [];
	    try {            
		while(statement.executeStep()) {
		    results.push(this._makeEntry(statement.getInt64(0), // 
						 this._valfromSQL(statement,1), // 
						 this._valfromSQL(statement,2), // 
						 this._valfromSQL(statement,3), //
						 this._valfromSQL(statement,4))); 							    
		}
	    } catch (e) {
		alert(e);
	    } finally {
		statement.reset();
	    }
	    return results;					
	},
	loadbyIDs : function(ids) {
	    var results = [];
	    for (var i = 0; i < ids.length; i++) {
		results = results.concat(this.loadbyID(ids[i]));
	    }
	    return results;
	},
	deletebyID: function(ids) {
	    var c = this._sqlfromVal;
	    for (var i = 0; i < ids.length; i++) {
		var id = ids[i];
		var SQLQUERY = 'DELETE FROM ' + this.table + " WHERE id="+c(id)+";";
		this.execute(SQLQUERY);
	    }
	},
	clear: function() {
	    var SQLQUERY = 'DELETE FROM ' + this.table + ";";
	    this.execute(SQLQUERY);
	},
	find: function(criteria) {
	    // returns a list of IDs for stored objects that match a conjunction
	    // of properties expressed as an object, e.g.,
	    // f = new PlumTripleStore('test','bladerunner');
	    // f.find({name:"Decker",type:"Bounty Hunter"})
	    // results in:
	    //   SELECT a0.subj FROM bladerunner a0, bladerunner a1 WHERE a0.pred="name"
	    //     AND a0.obj="Decker^^#string" AND a1.pred="type" AND a1.obj="Bounty Hunter^^#string" 
	    //     AND a0.subj=a0.subj AND a0.subj=a1.subj;
	    
	    var AND = " AND " ;
	    // fields id integer(100), type varchar, location varchar, host varchar, title varchar
	    var where_clause = '';
	    
	    var fields = ['id','type','location','host','title'];
	    for (var ii = 0; ii < fields.length; ii++) {
		var f = fields[ii];
		if (criteria != undefined && criteria[f] != undefined) {
		    where_clause = ""+f+"="+this._sqlfromVal(criteria[f])+AND;
		}
	    }
	    var sqlstmt = "SELECT id from " + this.table + (where_clause.length > 0 ?
							    (" WHERE " + where_clause.substring(0,where_clause.length-AND.length))
							    : "") + ";";
	    var hits = [];
	    this.log(sqlstmt);
	    var statement = this._db.createStatement(sqlstmt);
	    try {            
		while(statement.executeStep()) {hits.push(statement.getInt64(0));    }
	    } catch (e) {
		alert(e);
	    } finally {
		statement.reset();
	    }
	    return this.plumutil.uniq(hits);					
	},
	findByIdRange: function(lower,upper) {
	    var sqlstmt = "SELECT id from " + this.table +" WHERE id >= " + this._sqlfromVal(lower) + " AND id <" + this._sqlfromVal(upper) +";";
	    var hits = [];
	    
	    // this.log(sqlstmt);
	    var statement = this._db.createStatement(sqlstmt);
	    try {            
		while(statement.executeStep()) {
		    hits.push(statement.getInt64(0));
		}
	    } catch (e) {
		alert(e);
	    } finally {
		statement.reset();
	    }
	    return this.plumutil.uniq(hits);					
	},				
	load : function(criteria) {
	    return this.loadbyIDs(this.find(criteria));
	},
	loadByIdRange : function(low,high) {
	    return this.loadbyIDs(this.findByIdRange(low,high));
	},
	createTable: function(tname){
	    this.super.createTable(tname,this.SCHEMA);
	}
    };

    ContextMaiden.WatchmeStore = WatchmeStore;
});