
/** introduced by emax to speed up activity log reads + writes **/

(function(JV3) {    
    // modernized actlog class based on context stores to make more scalable
    JV3.ActLogStore = ({
	superclass: ContextMaiden.CMStore,
	table:"activity_log",
	debug:false,
        index:["id","type", "noteid", "search"], 
	log:function(x) { JV3.log(x); },
	schema: "(id NUMERIC, type TEXT, noteid TEXT, noteText TEXT, search TEXT)",
	schemacols: ["id", "type", "noteid", "noteText", "search"], 
	column_grabbers: {id : ContextMaiden.mozGetType.double,
			  type : ContextMaiden.mozGetType.string,
			  noteid : ContextMaiden.mozGetType.string,
			  noteText : ContextMaiden.mozGetType.string,
			  search : ContextMaiden.mozGetType.string },    
	initialize: function(params) {
	    JV3.PROTOCROCK.superclass(this,"initialize", {log: this.log , debug:true, table: this.table});
	    this._createTable();
	    this._client_id = JV3.getClientId();
	},
	_deconstitute:function(hydrated) {
	    delete hydrated.client;  // remove the client (which is always us) before saving to disk
	    return hydrated;
	},
	_reconstitute:function(dehydrated) {
	    dehydrated.client = this._client_id;
	    return dehydrated;
	},
	save:function(action) {	    
	    try {
		/*
		if (action.type == 'note-add') {
		    JV3.log("note-add "+action.noteid+"|"+ action.id + "|" + JV3.notes[action.noteid].created + "|" +
			    Math.abs(JV3.notes[action.noteid].created - action.id));
		}
		 */
		return this.saveCommonAsync({id:action.id,type:action.type}, this._deconstitute(action), function() { });
	    } catch(E) { JV3.log("async failed"); }
	    
	    try {
		this.saveCommon({id:action.id,type:action.type}, this._deconstitute(action));
	    } catch(E) {
		JV3.log(E);
	    }
	},			   
	getMinID:function() {
	    var request = "select min(id) from " + this.table;
	    JV3.log(request);
	    return this.getIntResults(request);
	},
	getMaxID:function() { return this.getIntResults("select max(id) from "+this.table ); },
	loadbyID:function(id) {
	    // warning: synchronous
	    try {
		var this_ = this;
		var results = [];
		this._forResultsArray([{col:"id", op:"=", val:id}], "AND", function(r) { results.push(this_._reconstitute(r));   });
		return results;
	    } catch(E) { JV3.log(E);  }
	},
	loadIdsGreaterThan:function(id,cont) {
	    var this_ = this;
	    // ASYNC:
	    try {
		// try to do it using async first
		this._forResultsArrayAsync([{col:"id", op:">", val:id}], "AND", function(r) { return this_._reconstitute(r);}, cont);
		return;
	    } catch(E) { JV3.log(E);  }
	    try {
		var results = [];
		this._forResultsArray([{col:"id", op:">", val:id}], "AND", function(r) { results.push(this_._reconstitute(r));   });	    
		cont(results);
	    } catch(E) {
		JV3.log(E);
	    }
	},	
	open: function(tablename) {
	    try {
		return JV3.newify(this,{});
	    } catch(E) {
		JV3.log(E);
	    }
	},
	countType:function(actType) {
	    var request = "select count(*) from " + this.table + " where type=\""+actType+"\";";
	    JV3.log(request);
	    return this.getIntResults(request);
	},
	countAsync:function(cont) {
	    try {
		var request = "select count(*) from " + this.table + ";";
		this.getIntResultsAsync(request, function(x) {
					    cont(x[0]);
					});
	    } catch (x) {
		JV3.log(x);
	    }
	}			   
    });

    JV3.Old_ActLog={
	_makeActivityLogTableName:function(d) {
	    return "actlog_" + JV3.plumutil.midnightOf(d).valueOf();
	},
	openActivityLogforDay: function(date) {
	    var this_ = this;
	    var logName = this._makeActivityLogTableName(date);
	    return JV3.newify(JV3.PlumTripleStore,
			       {file:"jourknow", table:logName,
				debug:false,
				log:function(x) {
				    this_.errors.push("log store: " + x);
				    this_.log(x);
				}});
	},
	openActivityLog: function(tablename) {
	    var this_ = this;
	    return JV3.newify(JV3.PlumTripleStore,
			       {file:"jourknow", table:tablename,
				debug:false,
				log:function(x) {
				    this_.errors.push("log store: " + x);
				    this_.log(x);
				}});
	},
	getActivityLogTablesNewerThan: function(d) {
	    var this_ = this;
	    var d_midnight = JV3.plumutil.midnightOf(d);
	    var master = JV3.newify(JV3.PlumStore, {debug:false,
						    log:function(x) {
							this_.errors.push("log store: " + x);
							this_.log(x);
						    }});
	    var ts = master.getTables();
	    var results = [];
	    for (var i = 0; i < ts.length; i++) {
		try {
		    if (!ts[i].indexOf("actlog_") === 0) {
			continue;
		    }
		    var dateval = parseInt(ts[i].slice(ts[i].indexOf("_")+1));
		    if (dateval >= d_midnight.valueOf()) {
			results.push(ts[i]);
		    }
		} catch (e) {
		    JV3.log(e + " " + e.lineNumber + " " + e.fileName);
		}
	    }
	    return results;
	}
    };
})


