
/** introduced by emax to speed up activity log reads + writes **/

(function(JV3) {    
     // modernized actlog class based on context stores to make more scalable
     JV3.NoteStore = ({
			  superclass: JV3.ContextMaiden.CMStore,
			  table:"notebook",
			  debug:false,
			  index:["id", "modified", "deleted", "version"],
			  log:function(x) { JV3.log(x); },
			  schema: "(id NUMERIC, created NUMERIC, edited NUMERIC, version INTEGER, text TEXT, modified INTEGER, deleted INTEGER, analysis TEXT, data TEXT)",
			  schemacols: ["id", "created", "edited", "version", "text", "modified", "deleted", "analysis", "data"],
			  column_grabbers: {id : ContextMaiden.mozGetType.double,
					    created : ContextMaiden.mozGetType.double,			  
					    edited : ContextMaiden.mozGetType.double,
					    version : ContextMaiden.mozGetType.int,
					    text : ContextMaiden.mozGetType.string,
					    modified : ContextMaiden.mozGetType.int,
					    deleted : ContextMaiden.mozGetType.int,
					    analysis : ContextMaiden.mozGetType.string,
					    data : ContextMaiden.mozGetType.string},

			  
			  initialize: function(params) {
			      JV3.PROTOCROCK.superclass(this,"initialize", {log: this.log , debug:true, table: this.table});
			      this._upgrade_table();
			  },
			  _deconstitute:function(hydrated) {
			      hydrated.modified = hydrated.modified ? 1 : 0;
			      hydrated.deleted = hydrated.deleted ? 1 : 0;
			      return hydrated;
			  },
			  _reconstitute:function(dehydrated) {
			      dehydrated.modified = dehydrated.modified == 1 ? true : false;
			      dehydrated.deleted = dehydrated.deleted == 1 ? 1 : false;	    
			      return dehydrated;
			  },		     
			  save:function(note,async) {
			      if (async !== false) {
				  try {
				      return this.saveCommonAsync({id:note.id}, this._deconstitute(note),function() {});
				  } catch(E) { JV3.log(E);   }			      				  
			      }
			      try { this.saveCommon({id:note.id}, this._deconstitute(note));   } catch(E) { JV3.log(E); }
			  },			   
			  loadbyID:function(id) {
			      var vv = this._load_clauses([{col:"id", op:"=", val:id}]);
			      if (vv && vv.length > 0) {
				  return vv[0];
			      }
			      return undefined;
			  },
			  loadAll:function(id) {
			      JV3.log("LOADALL>> " + this._get_number_of_new_notes());
			      return this._load_clauses([]);
			  },			  
			  _load_clauses:function(clauses) {
			      try {
				  var this_ = this;
				  var results = [];
				  this._forResultsArray(clauses, "AND", function(r) { results.push(this_._reconstitute(r)); });
				  return results;
			      } catch(E) {  JV3.log(E);  }
			      return undefined;
			  },
			  loadAllAsync:function(cont) {
			      return this._load_clausesAsync([],cont);
			  },
			  findTextMatch: function(txt,callback) {
			      var select_keys = this.schemacols;
			      // var t = "%"+txt+"%";
			      // var _temp = this._makeWhere({text:t}, " LIKE ", " AND " );
			      // var where_selector = _temp[0], bindings = _temp[1];
			      // var query = "SELECT " + select_clause + " FROM " + this.table + (where_selector.length > 0 ? " WHERE " + where_selector : "") + ";";
			      // JV3.log("FIND CLAUSE ", query);
			      var where = txt.trim().split(" ").map(function(x) { return x.trim(); }).filter( function(x) { return x.length > 0; }).
				  map(function(x) { return "text LIKE \'%"+x.trim().replace(new RegExp("\'","g"),"\'\'")+"%\'"; }).join(" AND ");			      
			      var query = "SELECT * FROM " + this.table + " WHERE deleted=0 AND " + where + "; "; //text LIKE \"%"+txt+"%\";";
			      // JV3.log(query);
			      var statement = this._db.createStatement(query);			      
			      this._execute_async_statement(statement,
							    function(results) {
								callback(results.map(function(x) { return x.id; }));
							    });
			  },				  
			  _load_clausesAsync:function(clauses,cont) {
			      try {
				  var this_ = this;
				  var results = [];
				  this._forResultsArrayAsync(clauses, "AND", function(r) { results.push(this_._reconstitute(r)); },function() { cont(results); });
			      } catch(E) {  JV3.log(E);  }
			      return undefined;
			  },			  
			  open: function(tablename) {
			      try { return JV3.newify(this,{});  } catch(E) {  JV3.log(E);  }
			      return undefined;
			  },
			  _get_number_of_old_notes:function() {
			      JV3.log("Get Number of Old Notes");
			      try {  return this.getIntResults("select count(*) from notes where pred=\"id\";");  } catch (x) {  JV3.log("no old notes");     }
			      return 0;
			  },
			  _get_number_of_new_notes:function() {
			      JV3.log("Get Number of New Notes");
			      try {  return this.getIntResults("select count(*) from "+this.table+";");  } catch (x) {  JV3.log(x);     }
			      return 0;
			  },			  
			  _upgrade_table : function() {
			      // to determine whether we're done with the upgrade, look in the NEW database for magic note and check client id upgrade
			      // try {
			      // 	  var magic = this.loadbyID(-1);
			      //          var this_ = this;
			      // 	  if (magic !== undefined && magic !== null) {
			      // 	      var json = JV3.plumutil.fromJSON(magic.text);
			      // 	      if (JV3.plumutil.objKeys(json).length > 0) {
			      // 		  // successful upgrade,
			      // 		  JV3.log(" ----- upgrade was already completed.");
			      // 		  return;
			      // 	      }
			      // 	  }
			      // } catch (x) {
			      // 	  JV3.log("upgrading table: " + x);				  
			      // }
			      // JV3.log("no table, initalizing/upgrading..");

			      var this_ = this;
			      try {				  
				  this._createTable();
				  JV3.log("CHECKING necessary :: Old notes store: " + this._get_number_of_old_notes() + " >? " + this._get_number_of_new_notes());
				  if (this._get_number_of_new_notes() < this._get_number_of_old_notes()) {
				      JV3.log("MIGRATION necessary :: Old notes store: " + this._get_number_of_old_notes() + " > " + this._get_number_of_new_notes());
				      var ostore = this.newify(JV3.PlumTripleStore, {file:"jourknow", table:"notes",  debug:false, log:function(x) { this_.log(x);  }});
				      var onotes = this._load_notes_old(ostore);
				      if (onotes !== undefined) {
					  JV3.log("UPGRADING xfer.................");
					  JV3.plumutil.objVals(onotes).map(function(old) {
									       var newn = undefined;
									       try { newn = this_.loadbyID(old.id);  } catch (x) { JV3.log(x);  }
									       if (newn !== undefined) { return; }
									       this_.save(old,false);
									   });
					  JV3.log("UPGRADING commit.................");				  
					  // transferring magic note last
					  JV3.log("UPGRADING done!");
				      }
				  }
			      } catch (x) {
				  JV3.log("UPGRADE FAILED: ");
				  JV3.log(x);
			      }			      
			      return true;
			  },
			  _load_notes_old:function(store,skip_verify) {
			      // runs an integrity check over all notes (verifying all fields are .
			      // tries to salvage one sthat are not intact, by replacing fields iwht
			      // default values, and saving these back to disk
			      var notelist = store.load();
			      var notes = {};
			      for (var i = 0; i < notelist.length; i++) {
				  //if (!notelist[i].deleted) {} // include the deleted notes!
				  notes[notelist[i].id] = notelist[i];
			      }
			      //JV3.log("patching!");
			      for (var n_i in notes) {
				  //JV3.log(n_i);
				  var note = notes[n_i];
				  var valid = JV3._verify_and_construct_note(note);
				  if (valid.is_complete) {
				      notes[n_i] = note;
				  } else {
				      if (valid.is_patched_complete) {
					  //store.save(valid.note);
					  notes[n_i] = valid.note;
				      } else {
					  this.log("Note cannot be salvaged -- " + JV3.plumutil.objVals(valid.missing_fields) + " " + JV3.plumutil.objVals(note));
					  continue;
				      }
				  }
			      }
			      return notes;
			  }			  
		      });
 });


