// Copyright (c) 2008, 2009 Massachusetts Institute of Technology

// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:

// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.

// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

var JV3 = {};

FloatyViewController = {
    grace_period:3600,
    history:2*3600,
    N:10,
    initialize:function(JV3, log) {
	this.JV3 = JV3;
	this.util = JV3.plumutil;
	this.$j = JV3.$j;
	this.TS = JV3.CMS.TS;
	this.log = JV3.log;
	this.util = JV3.plumutil;
	this.visible = false;
	this.visibleNotes = {};
	this.noteShowTime = {};
	this.cprovider = JV3.CMS.pstats;
	this.current_context = this._preload_context();	
    },
    buildDOM:function(entries_element,context_element) {
	this.log('build.');
	this.element = this.$j(entries_element);
	this.current_context_element = this.$j(context_element);
	var $j = this.$j;

	this.cprovider.getTypes().map(function(type) {
	    this.$j("<div class=\"context-state\" id=\"context-state-"+type+"\">"+type+":<span class=\"value\"></span></div>").
		click(function(x) { $j(this).toggleClass("context-disabled"); }).
		appendTo(context_element);
	});
    },
    getContextValueElement:function(type) {
	return this.current_context_element.find("#context-state-"+type).find(".value");
    },
    start:function() {
	this.log('floatyview: start.');
	// AUTOMATICALLY starts looking for notes and repopulates according
	// to changes in context
	var this_ = this;
	var TS = this.TS;
	this.JV3.CMS.event_store.addStoreUpdateListener(undefined,{
	    itemUpdated: function(event) {
		var now = this_._getNowTS();
		if (TS.makeDuration(event.start,event.end).intersection(now)) {
		    this_.log("floaty: itemUpdated event : " + event);
		    this_.current_context[this_.cprovider.getEventType(event)] = event;
		}
		try {
		    this_._update_display();
		} catch(e) {
		    this_.log(e + " " + e.lineNumber + " " + e.fileName);
		}
	    }
 	});
    },    
    _preload_context:function() {
	var c = this.cprovider;
	var store = this.JV3.CMS.event_store;
	var types = this.cprovider.getTypes();
	var nowTS = this._getNowTS();
	var current_context = {};
	types.map(function(type) {
	    var evts = store.getEvents(type,nowTS);
	    var evt = evts.length > 0 ? evts[0] :undefined;
	    if (evts.length > 1) {
		evts.sort(function(x,y) { return x.end < y.end; });
		evt = evts[evt.length-1];
	    }
	    if (evt) {	current_context[c.getEventType(evt)] = evt;   }
	});
	return current_context;
    },
    _getNowTS:function() {
	var TS = this.TS;
	var nowdate = new Date().valueOf();
	return TS.makeDuration(nowdate-this.history,nowdate);
    },
    _update_display:function() {
	var this_ = this;	
	var store = this.JV3.CMS.entity_store;
	var currently_displaying_ids = this.getShowingNoteIds();
	var note_rank = {};
	var N = this.N;	
	var TS = this.TS;
	var nowTS = this._getNowTS();
	var cprovider = this.cprovider;
	
	// determine givens based on current context
	var givens = {};
	this.util.objKeys(this.current_context).map(function(contexttype) {
	    if (!this_.getContextValueElement(contexttype).parent().hasClass("context-disabled")) {
		givens[contexttype] = this_.current_context[contexttype];
	    } else {
		this_.log("skipping context element : " + contexttype);
	    }
	});
	this_.log("context elements included in givens : " + this.util.objKeys(givens));	
	this.JV3.givens = givens;
	cprovider.getTypes().map(function(type) {
	    if (givens[type]) {
		this_.getContextValueElement(type).html( givens[type].entity ? givens[type].entity.id : givens[type].entityid );
	    } else {
		this_.getContextValueElement(type).html("none");
	    }
	});
	
	if (givens) { delete givens["note-edited"]; }
	if (this.util.objKeys(givens).length == 0) {  this.empty();    return;	}

	this_.log("context elements included in givens : " + this.util.objKeys(givens));

	var displayRankedNotes = function(rankings) {
	    // let's see how many of the top N
	    this_.element.find(".note-display:gt("+(rankings.length-1)+")").remove();
	    var displays = this_.element.find(".note-display");
	    displays = this_.element.find(".note-display");
	    rankings.map(function(noterank) {
		var rank = rankings.indexOf(noterank);
		if (rank >= displays.length) {
		    this_.addNote(noterank.note);
		} else {
		    this_.$j(displays[rank]).hide("slow", function() {
			this_.showNoteInElement(noterank.note,displays[rank]);
			this_.$j(displays[rank]).show("slow");
		    });
		}
	    });
	};
	if (this._map) { this._map.stop();}

	// speedyrank notes
	var speedyRank = function(notes,givens,minval) {
	    var PI = function(l) {
		return l.reduce(function(x,y) { if (y === undefined) { return x; } return x*y; });
	    };
	    // notes -> note val
	    notes = notes.map(function(x) {
		return { type:"note-edited", entity:{id:"note-"+x.id, note:x} };
	    });
	    var denominator = PI(
		givens.map(function(given) {
		    var given_T = cprovider.getTime(given);
		    if (given_T === 0) { this_.log("Given : " + given + " was 0"); return 0; }
		    return 1.0/given_T;
		}));
	    return notes.map(function(val) {
		return {
		    note:val.entity.note,
		    p:(PI(givens.map(function(given) {
 			var overlap = cprovider.getCoTime(val,given);
			if (overlap === 0 && minval) { overlap = minval; }
			return overlap;
		    })))*denominator
		};
	    });
	};
	var rankings = speedyRank(this.util.objVals(this.JV3.notes).filter(function(x) { return !x.deleted; }),this.util.objVals(givens),1e-4);
	rankings.sort(function(x,y) { return x.p < y.p; });
	this_.log("floaty: probability ranges : " + rankings[0].p + " - " + rankings[0].p + " given " + this.util.objVals(givens).map(function(x) { return cprovider.getEntityId(x); }).join(" + "));
	rankings = rankings.slice(0,N);		    //rankings.filter(function(x) { return x.p > 1e-12; }).slice(0,N);		    
	displayRankedNotes(rankings);		    
	
    },
    
    showNoteInElement:function(note,element) {
	if (element.id == note.id) { return; }
	this.$j(element).find(".note-display-text").html(note.text);
	element.id = note.id;
    },    
    empty:function(note,element) {this.$j(element).find(".note-display-text").remove();    },
    addNote:function(n) {
	var nd = this._makeNoteDisplay(n);
	if (nd) { this.element.append(nd);}
	this.visibleNotes[n.id] = nd;
	this.noteShowTime[n.id] = new Date();
    },
    removeNote:function(n) {
	if (typeof n != 'object') {   n = { id: n };} // check if someone gave us the id instead	
	this.$j(this.visibleNotes[n.id]).remove();
	delete this.visibleNotes[n.id];
	delete this.noteShowTime[n.id];
    },
    getShowingNoteIds: function() { return this.util.objKeys(this.visibleNotes); },
    getShowingNotes:function() {
	var this_ = this;
	return this.getShowingNoteIds().map(function(x) {
	    return this_.JV3.getNoteById(x);
	});
    },
    _makeNoteDisplay: function(note) {
	var text = note.text;
	var this_ = this;
	if (!note.text) { return; }
	var noteDisplay = document.createElement('div');
	noteDisplay.setAttribute('class', "note-display");
	noteDisplay.id = note.id;
	
	var markerSpan = document.createElement('ul');
	markerSpan.setAttribute('class','note-display-list');
	
	var textSpan = document.createElement('li');
	textSpan.setAttribute('class', "note-display-text");
	textSpan.innerHTML = this.JV3.util.defang(text);
	
	var sinkSpan = document.createElement('div');
	sinkSpan.setAttribute('class', "sink-button");
	
	markerSpan.appendChild(textSpan);
	noteDisplay.appendChild(sinkSpan);
	noteDisplay.appendChild(markerSpan);
	
	sinkSpan.addEventListener('click', function(e)  {
            var entry = e.target.parentNode.parentNode;
	    this_.$j(entry).fadeOut("slow", function () {
		/* JV3.deleteNote(entry.id); */
	    });
	    e.cancelBubble = true; // stop propagation
	}, false);
	return noteDisplay;
    },
    setVisible:function(visible) {	
	if (visible) { this.clear(); }
	this.visible = visible;
	try {
	    window.opener.document.getElementById("floatybox").setAttribute("hidden", visible);
	    return;
	} catch(e) {
	    this.log(e);
	}
	this.log("window.opener is null, can't control whether we're showing. or not");
    },
    show: function() { this.setVisible(true); },
    hide:function() { this.setVisible(false); }
};

jQuery(document).ready(function() {
    Components.utils.import("resource://listit/share.jsm");
    self.JV3 = share.JV3;

    var size = self.JV3.prefs.getIntPref('general.fontSize');
    if (size === 0) {size = "100%";}
    var entries = document.getElementById("entries");
    jQuery(entries).css('font-size',size);
    try {
	self.view = self.JV3.newify(FloatyViewController,self.JV3);
    } catch(e) {
	self.JV3.log(e + " " + e.lineNumber + " " + e.fileName);
    }
    self.view.buildDOM(entries, jQuery("#context"));
    self.view.start();
    self.view._update_display();
});

	/*this._map = this.util.interval_map( ,
	    function(note) {
		try{
 		    if (note.deleted || note.text.length == 0) { return {note: note, p:0};    }
		    var ret = {note: note, p:cprovider.getConditionalPTime({ type:"note-edited", entity:{id:"note-"+note.id} },this_.util.objVals(givens), 1e-2)};
		    this_.log("evaluating " + note.id + " " + ret.p);
		} catch(e)  {  this_.log(e + " " + e.lineNumber + " " + e.fileName);	}
	    },
	    function(rankings) {
		try {		    
		    
		} catch(e) {
		    this_.log(e + " " + e.lineNumber + " " + e.fileName);
		}
	    },this_.log,100);
*/
