// Copyright (c) 2008, 2009, 2010 Massachusetts Institute of Technology
// Copyright (c) 2008, 2009 Nokia

// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:

// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.

// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

// Scroll.js - Main driver code for controller
// scroll-controller.js - viewController definition

function getJV3() {
    if (self.JV3 === undefined) {
	self.JV3 = Components.classes['@electronic.max/listit;1'].getService().wrappedJSObject;
    }
    return self.JV3;	
};

// init!
$j(document).ready(function() {
		       // scroll.js init --------------------- //
		       var s = new Date().valueOf();
		       var JV3 = getJV3();
		       self.JV3 = JV3;
		       
		       
		       var size = JV3.prefs.getIntPref('general.fontSize');
		       if (size == 0) { size = "100%";  }
		       $j("#entries").css('font-size',size);
		       $j("body").click(viewController.closeAndSaveAll);
		       $j(window).scroll(logger.check,2000);

		       /* ============== add integration for analysis view ================== */
		       if (JV3.analysis) {
			   JV3.analysis.initDisplay(self);
		       }
		       

		       viewController.initialize(JV3);
		       JV3.log(new Date().valueOf() - s);
		   });

var logger = {
    time : new Date().valueOf(),
    wasVisible : [],
    entryTimes: {},
    timer : {},

    check : function() {
        logger.time = new Date().valueOf();
        clearTimeout(logger.timer);
        logger.timer = setTimeout(function() {
            logger.read();
        }, 2000);
    },

    read : function() {
        var nowVisible = logger.getVisibleNotes($j(window).scrollTop());
        var exit = set.minus(logger.wasVisible, nowVisible);
        var enter = set.minus(nowVisible, logger.wasVisible);

	/*
	// DEBUG ZONE -- comment me out 
	try {
	    if (JV3.prefs.getBoolPref("debug")) {
		JV3.log("read >> ");
		JV3.log("wasVisible :" + logger.wasVisible.map(function(x) {  return logger.dfText(x) + ": " + x.id;  }).join("; "));
		JV3.log("nowVisible :" + nowVisible.map(function(x) {  return logger.dfText(x) + ": " + x.id;  }).join("; "));
		JV3.log("exit :" + exit.map(function(x) {  return logger.dfText(x) + ": " + x.id;  }).join("; "));
		JV3.log("enter :" + enter.map(function(x) {  return logger.dfText(x) + ": " + x.id;  }).join("; "));
		JV3.log("read << ");		
	    }
	} catch (x) {
	    JV3.log(x);
	}
	 */
	// -------------------------------
	
        logger.log(enter,exit);
        logger.wasVisible = nowVisible;
    },
    
    user_active : function() {
        logger.check();
    },

    user_inactive : function() {
        logger.log([],logger.wasVisible);
        logger.wasVisible = [];
    },

    dfText:function(nt) {
	try {  return JV3.notes[nt.id].text.substring(0,20).replace(/\n/g,'');	} catch (x) {   /*JV3.log(x);*/	}
	return "<undefined>";
    },

    log : function(enter,exit) {
        var sigscroll = exit.map(function(exitNote) {
            return { id : exitNote.id,
                     entryTime : logger.entryTimes[exitNote.id],
                     exitTime : logger.time
                   }
        });
        if (sigscroll && sigscroll.length > 0) {
/*
	    try {
		// DEBUG -- 
		JV3.log("sigscroll >> " + sigscroll.length);
		sigscroll.map(function(nvis) {
				  JV3.log(
				      logger.dfText(JV3.notes[nvis.id]) + " ["+nvis.id+"]: " + (new Date(nvis.entryTime)) + " " + (new Date(nvis.exitTime) + " start-now:" + ((new Date().valueOf()-nvis.entryTime)/1000.0) + "s // " +
				  " end-now:" + ((new Date().valueOf()-nvis.exitTime)/1000.0) + "s // "));
			      });
		JV3.log("<< sigscroll ");
		// end debug
	    } catch (x) { JV3.log(x); }
 */

	    // JV3.log("significant-scroll", sigscroll);
            JV3.logAction("significant-scroll", undefined, undefined, undefined, {window_id:JV3.getWindowId(window),note_visibilities:sigscroll});
        }
        enter.forEach(function(enterNote) {
            if (enterNote) { // apparently this is undefined sometimes. haven't figured out why yet.
                logger.entryTimes[enterNote.id] = logger.time;
            }
        });
    },

    getVisibleNotes : function(top) {
        var visible = [];
        var notes = document.getElementById("entries").childNodes;
        var bottom = top + $j(window).height();
        var min = -1;
        var max = notes.length;
        var mid;
        while (max - min > 1) {
            if($j(notes[mid = max + min >> 1]).offset().top < top) min = mid;
            else max = mid;
        }
        visible.push(notes[max]);
        for (var i = max + 1; i < notes.length; i++) {
            if ($j(notes[i]).offset().top + $j(notes[i]).height() < bottom) {
                visible.push(notes[i]);
            } else { 
                break;
            }
        }
        return visible;
    }
};

var set = {
    intersect : function(first,second) {
        return first.filter(function(item) {
            var i = second.indexOf(item);
            return i >= 0 ? item.id === second[i] : false;
        });
    },

    minus : function(first,second) {
        return first.filter(function(item) {
            var i = second.indexOf(item);
            return i >= 0 ? item.id != second[i] : true;
        });
    }
}

// DOM lifting methods --------
var display = {
    close : function(event) {
        var entry = event.target.parentNode.parentNode;
	$j(entry).hide("slow", function () {
	    JV3.deleteNote(entry.getAttribute("id"), _logActionHelper());
	});
	event.cancelBubble = true;
	return false;
    },
        
    create : function(note) {
	/* -- debug -- 
	JV3.log('create--------------------------------------------------');
	try {
	    throw new Error("foo.");
	} catch (x) {
	    JV3.log(x);
	}
	*/
        var noteDisplay = document.createElement('div');
        noteDisplay.setAttribute('class', "note-display");
        noteDisplay.innerHTML =  '<div class="handle-span"></div><div class="close-button"></div><div class="note-display-list"><div class="note-display-text">'+
	    JV3.util.defang(note.text)+'</div></div>';
	$j(noteDisplay).find(".close-button").click(display.close);
        augmentation("displayNote", self, note, {notedisplay: noteDisplay, textspan: noteDisplay.childNodes[2].firstChild});
        return noteDisplay;	
    },

    update_entry_text:function(entry,note) {
	$j(entry).find(".note-display-text").html(JV3.util.defang(note.text));
    }
}

_augmentfns = {};
function augmentation() {
    var args = JV3.plumutil.arguments2array(arguments);
    var method = args[0];
    args = args.slice(1);
    if (_augmentfns[method]) {
	_augmentfns[method].map( function(f) {
	    f.apply(self,args);
	});
    }
}
function addAugmentation(type,fn) {
    if ( self._augmentfns[type] ) {
	self._augmentfns[type].push(fn);
    }
    self._augmentfns[type] = [fn];
}

function _logActionHelper() {
    // this magic adds some metadata to our activity log entries
    // so that we can learn smart things (safely) about the ways
    // people take notes
    var misc = {};
    if (window) {
	var wid = JV3.getWindowId(window);
	// add window id
	if (wid !== undefined) { misc.wid = wid; }
	// add concurrent activities: web and loc
	var query = getQuery();
	if (query && query.length > 0) { misc.query = query; }
	var filters = JV3.plumutil.objKeys(viewController.getActiveFilters());
	if (filters.length > 0) {
	    misc.active_research = filters;
	}
	try { misc.ver = JV3.getClientVersion(); } catch (x) {	}
	if (JV3.getScienceMode() && !JV3.isPrivateBrowsing()) {
	    var doencrypt = JV3.doEncryptActivities();
	    var d = JV3.util.getMainWindow(window).content;
	    try {
		misc.viewing_host = doencrypt ? JV3.plumutil.hash(d.location.host) : d.location.host;
		misc.viewing_url = doencrypt ? JV3.plumutil.hash(d.location.href) : d.location.href;		
	    } catch (x) { JV3.log("warning - no documents logged");   }
	    
	    if (JV3 && JV3.CMS.getSourceByName('ContextMaiden.WiFiLocation') !== undefined) {
		try {
		    var wloc = JV3.CMS.getSourceByName('ContextMaiden.WiFiLocation').getLastKnownLocation();
		    if (wloc !== undefined){	    
			misc.loc = doencrypt ? JV3.plumutil.hash(wloc.id) : wloc.id;
		    }		    
		} catch (x) {JV3.log(x);}
	    }	    
	}
	return misc;
    }    
}

function closeAndSave(entry, animate) {
    // if animate, fade back to the other class
    // if !animate, just change class instantly
    if (entry.getAttribute("class").indexOf("editing") >= 0) {
        var editTable = entry.childNodes[1]; // note-display is first, editTable is second
        var ta = editTable.getElementsByTagName('textarea')[0];
        JV3.logAction("note-save", entry.getAttribute("id"), ta.value, undefined, _logActionHelper());
        var note = JV3.getNoteById(entry.getAttribute("id"));
	note.text = ta.value;
	note.edited = new Date().valueOf();
	note.modified = true;
        var editHeight = $j(entry).height();
	JV3.plumutil.assert(editTable !== undefined, "editTable is undefined problems!");
	try {
	    entry.removeChild(editTable);	    
	} catch(e) {
	    JV3.log(e + " " + e.lineNumber + " " + e.fileName);
	    return;
	} // NO SUCH NODE error

	try {	    	    
            var noteDisplay = entry.childNodes[0];
            // already done on note update ~ $j(noteDisplay).find('.note-display-text').html(JV3.util.defang(note.text));
	    //$j(noteDisplay).fadeIn(1000);
	    $j(noteDisplay).css("display","block");
	    var collapsed = viewController.isNoteDisplayCollapsed(noteDisplay) ;
	    if (animate) {

		var height = $j(noteDisplay).height();
		$j(entry).height(editHeight+4);
		$j(entry).css("border","none");
		$j(entry).css("-moz-border-radius","0");
		entry.setAttribute("class", "entry");
		JV3.saveNote(note.id,_logActionHelper());
		var targetHeight = collapsed ? viewController.getCollapsedHeight() : $j(noteDisplay).height();
 		$j(entry).animate({
 				      height: targetHeight,
 				      padding: 0
 				  }, 500, "linear", function() {
 				      try {					  
 					  $j(entry).css({
 							    height:"",
 							    padding: "",
 							    border: "",
 							    "-moz-border-radius":""
 							});
					  $j(noteDisplay).css("height", collapsed ? targetHeight : "");

 				      } catch (y) {  JV3.log(y);    }					  
 				  });
	    } else {
		entry.setAttribute("class", "entry");
		$j(entry).css({height: ""});
		$j(noteDisplay).css({ height: collapsed ? viewController.getCollapsedHeight() : "" });		
		JV3.saveNote(note.id, _logActionHelper());
	    }	    
	} catch (x) { JV3.log(x); }	    				   
    }
}

var edit = {
    
    click : function(event) {
	try {
            var entry = event.currentTarget;
	    if (event.target.className.indexOf("close-button") >= 0) { return true;  }
	    // filter out spurious clicks that are generated when user drags a note
	    if (this._dragclick(event)) {   event.cancelBubble = true;   return;	}
	    
	    if (entry.getAttribute("class").indexOf("editing") >= 0) {
		event.cancelBubble = true;    
	    }
            edit.open(entry);	    
	    
	} catch (x) {
	    window.JV3.log(x);
	}
    },

    _dragclick:function(event) {
	// checks internal sentinel to see if we recently dragged, and returns true if so
	// (HACK!)
	var tS = new Date().valueOf();
	JV3.log(JV3._dragged_);
	if (JV3._dragged_ !== undefined) { JV3.log(JV3._dragged_.time);}
	JV3.log(JV3._dragged_ !== undefined && tS > JV3._dragged_.time && (tS - JV3._dragged_.time) < 500 );
	return JV3._dragged_ !== undefined && tS > JV3._dragged_.time && (tS - JV3._dragged_.time) < 500 ;
    },

    open : function(entry) {
        viewController.closeAndSaveAll();

        var note = JV3.getNoteById(entry.getAttribute("id"));

        var editTable = edit.create(entry);
        if (entry.getAttribute("class").indexOf("editing") < 0) {

            try {
	    var height = $j(entry).height();
            $j(entry.firstChild).hide(); // hide for editing, come back on close

	    $j(editTable).height(height);
	    $j(editTable).find("textarea").height(height); // It will grow later
	    entry.appendChild(editTable);
	    editTable.text(note.text);
            } catch(e) {
                JV3.log(e);
            }
            height = editTable.getElementsByTagName("textarea")[0].scrollHeight;
	    $j(editTable).height(height); // readjust (sometimes necessary)
	    $j(editTable).find("textarea").height(height);
	    JV3.logAction("note-edit", note.id, note.text, undefined, _logActionHelper());

            $j(entry).animate({
	        height: height+30 // 30 == room for buttons
	    }, 250, "linear", function() {
	        entry.setAttribute("class", "editing");
            });
	    editTable.focus();
        }
    },

    text : function(editTable) {
        return function(text) { $j(editTable).find("textarea").val(text); };
    },

    focus : function(editTable) {
	return function() { editTable.getElementsByTagName('textarea')[0].focus(); };
    },

    save : function(entry) {
        return function(event) {
	    closeAndSave(entry, true);
	    event.cancelBubble = true;
        };
    },

    handleKeypress : function(entry) {
        return function(event) {
	    if (event.keyCode === 13) {
		var shift = event.shiftKey;
		var need = JV3.prefs.getBoolPref("general.shiftReturn");
		if ((shift && need) || (!shift && !need)) {
		    closeAndSave(entry, true);
		}
            }
	    event.cancelBubble = true;
	};
    },

    deleteNote : function(entry,editTable) {
        return function(event) {
	    $j(editTable.parentNode).hide("slow", function () {
	        closeAndSave(entry, true);
	        JV3.deleteNote(entry.getAttribute("id"), _logActionHelper());
	    });
	    event.cancelBubble = true;
        };
    },

    create : function(entry) {
        var editTable = document.createElement('div');
        editTable.setAttribute('class', 'note-editor');
        editTable.innerHTML = '<div class="note-edit-spacer"></div>'+
            '<div class="note-edit-x"></div>'+
            '<textarea class="note-edit-textarea" wrap="virtual" style="font-size: '+$j("#entries").css("font-size")+';"></textarea>'+
            '<div class="note-edit-buttons">'+
            '<button class="note-edit-save">Save</button>'+
            '<button class="note-edit-delete">Delete</button>'+
            '</div>';

        editTable.text = edit.text(editTable);
        editTable.focus = edit.focus(editTable);

        $j(editTable).find("textarea").keypress(edit.handleKeypress(entry));
        $j(editTable).find("textarea").keyup(edit.adjustHeight);
        $j(editTable).find(".note-edit-x").click(edit.save(entry));
        $j(editTable).find(".note-edit-save").click(edit.save(entry));
        $j(editTable).find(".note-edit-delete").click(edit.deleteNote(entry,editTable));
        editTable.onclick = function(event) {
	    event.cancelBubble = true;
	};
        return editTable;
    },

    adjustHeight : function(event) {
        var textarea = event.target;
        if (textarea.offsetHeight !== textarea.scrollHeight) {
            textarea.parentNode.parentNode.style.height = textarea.scrollHeight+30;
            textarea.style.height = textarea.scrollHeight;
        }
    }

}

function openLink(event) {
    event.preventDefault();
    event.cancelBubble = true;
    var Application = Components.classes["@mozilla.org/fuel/application;1"].getService(Components.interfaces.fuelIApplication);
    var newTab = Application.activeWindow.open(JV3.util.url(event.target.href));

    try {
	JV3.log("link-clicked",$j(event.target).parent().parent().parent().parent().attr("id"),event.target.href,undefined,_logActionHelper());
	JV3.logAction("link-clicked",$j(event.target).parent().parent().parent().parent().attr("id"),event.target.href,undefined,_logActionHelper());
    } catch (x) { JV3.log(x);   }    

    newTab.focus();
}

function findEntry(note) {
    var entries = document.getElementById("entries").childNodes;
    for (var i=0; i<entries.length; i++) {
        var entry = entries[i];
        if (note.id == entry.getAttribute("id")) {
            return entry;
        }
    }
    return false;
}
