// Copyright (c) 2009 Massachusetts Institute of Technology

// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:

// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.

// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.


// worker thread version of synchronize.js
// requires FF3.5 + 
// -- <3, emax

onmessage = function(event) {
    workerCall(event.data,
	       function(text,status) {
		   try {
		       var parsed = undefined;
		       try { parsed = JSON.parse(text); } catch (x) {	 }
		       postMessage(parsed||text);		       
		   } catch (x) {
		   }
	       },
	       function(req,status,text) {
		   if (status == undefined && text == undefined)
		       throw req;
		   throw "{\"status\":"+status+",\"responseText\":\""+text+"\"}"; // we signal an error that way, folks
	       });
};

workerCall = function(client, callback, error_callback) {
    try {
	var serialized = client.data;
	var type = client.type;
	var url = client.url;

	if (serialized !== undefined && typeof(serialized) == 'object') {
	    serialized = JSON.stringify(client.data);
	}    
	// Send
	var req = new XMLHttpRequest();
	req.open(type, url, false);

	// set defaults
	if (!client.contenttype) {    req.setRequestHeader('Content-Type', "text/json-comment-filtered"); }
	if (client.id) { req.setRequestHeader("X-ClientId", client.id); }
	if (client.username !== undefined && client.password !== undefined && client.client_auth_hash !== undefined) {
	    req.setRequestHeader('Authorization',client.client_auth_hash);
	}

	var kill_callbacks = function() { callback = undefined;  error_callback = undefined;};
	
	req.onreadystatechange = function(event) {
	    try {
    		if (req.readyState == 4) {
    		    if (req.status == 200) {
    			callback(req.responseText,req.status);
			kill_callbacks();
			return;			
    		    }
    		    /* error!! */
    		    req.abort();
    		    if (req.status == 401) {
    			if (error_callback) { return error_callback(req,401,"Username or password incorrect."); }
			kill_callbacks();			
    			return;
    		    }
    		    if (req.status == 404) {
    			if (error_callback) { return error_callback(req,404,"Page not found."); }
			kill_callbacks();			
    			return;
    		    }		
    		    if (req.status >= 500) {
    			if (error_callback) { return error_callback(req,500,req.responseText); }
			kill_callbacks();			
    			return;
    		    }
    		    if (error_callback) {
			error_callback(req,req.status,req.responseText);
			kill_callbacks();
			return;
		    }
    		    if (error_callback) { error_callback({status:0}, 0, "Unable to contact server - Check your connection.");  }
		    kill_callbacks();			
    		    return;
    		}		
	    } catch (X) {
		error_callback(X);		
		kill_callbacks();
		return;
	    }
	};
	req.send(serialized !== undefined ? serialized : null);	
    } catch (x) {
	error_callback(x);
    }
};
