// Copyright (c) 2008,2009 Massachusetts Institute of Technology
// Copyright (c) 2008,2009 Nokia

// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:

// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.

// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

// requires jQuery and protoCROCK. Don't even think about not having them. 
// -- <3, emax

(function(JV3) {
	var isnull = function(x) x === undefined || x === null;
	JV3.Synchronize = {    
	    methods : {"read": "GET",  "create": "POST",  "update": "POST", "delete": "DELETE"},
	    sync_status_to_message : {
		"json_parse" : "Error parsing server response",
		0: "Error connecting to server",
		404: "Wrong server path",
		200: "Success"
	    },
	    note_default_values : {edited: new Date(), created : new Date(), text: "", deleted: false, version: 0},
	    initialize : function(server_url,owner_id,path,store,activitylogstore,username,password) {
		if (path === undefined || path === null) { path = "/jv3/"; }
		this.server_url = server_url;
		this.url = server_url + path;
		this.owner_id = owner_id;
		this.store = store;
		this.activitystore = activitylogstore;
		this.username = username;
		this.password = password;
		this.define_ajaxjson();
		this.activitylog_bucket_size = prefs.getIntPref("sync.activityLogBucket");

		this._add_event_update_listener();
	    },
	    login:function(success_cb, failure_cb) {
		var this_ = this;
		this.ajaxJSON(this_.server_url + "/jv3/login", "GET", undefined,
			      function(data,status){
				  if (success_cb){
				      success_cb({status:200, couhes: JV3.plumutil.fromJSON(data)});				      
				  }
			      },
			      function(req,status,text) {
				  if (failure_cb) {
				      failure_cb.apply(this_,arguments);				      
				  }
			      });
	    },
	    set_consenting:function(value,success_cb, failure_cb) {
		var this_ = this;
		this.ajaxJSON(this_.server_url + "/jv3/set_consenting/", "POST",
			      {consenting:value},
			      function(data,status){
				  success_cb({status:200, couhes: JV3.plumutil.fromJSON(data)});
			      },
			      function(req,status,text) {
				failure_cb.apply(this_,arguments);
			      });
	    },	    
	    define_ajaxjson:  function(sync) {/*if (sync === undefined) { sync = this;}*/ },
	    ajaxJSON : function(url, type, data, callback, error_callback, doserialize) {
		try {
		    var username = this.username;
		    var password = this.password;
		    var this_ = this;
		    var serialized = data;
		    if (doserialize !== false) {
			serialized = (data === null || data === undefined) ? null : plumutil.toJSON(data);
		    }
		    // Send      
		    var req = Components.classes["@mozilla.org/xmlextras/xmlhttprequest;1"].createInstance(Components.interfaces.nsIXMLHttpRequest);
		    JV3.log("~type : " + type, " url: " + url);
		    req.open(type, url,  true); /* asynchronous! */  
		    req.setRequestHeader('Content-Type', "text/json-comment-filtered");
		    if (username !== undefined && password !== undefined) {
			req.setRequestHeader('Authorization',plumutil.make_base_auth(username,password));
		    }
		    // added to avoid caching
		    // req.setRequestHeader("If-Modified-Since", "Sat, 1 Jan 2005 00:00:00 GMT");
		    //
		    req.setRequestHeader("X-ClientId", JV3.getClientId());
		    req.onreadystatechange = function(event) {
			// 
			// JV3.log(event);
			// JV3.log("readystate ~ " + req.readyState + " status " + req.status );
			if (req.readyState == 4) {
			    if (req.status == 200) {
				// JV3.log("synchronize calling back ~ " + req.responseText);
				return callback(req.responseText,req.status);
			    }
			    /* error!! */
			    try {
				try {
				    JV3.log("___________error___________");
				    JV3.log("TRIED to access: " + url);
				    JV3.log(req.responseText);
				} catch (x) {   }
				req.abort();
				if (req.status == 401) {
				    if (error_callback) { return error_callback(req,401,"Username or password incorrect."); }
				    return;
				}
				if (req.status >= 500) {
				    if (error_callback) { return error_callback(req,500,"Server is down. Please try later."); }
				    return;
				}
				if (error_callback) { return error_callback(req,req.status,req.responseText);}
				return;					
			    } catch(E) {
				JV3.log(E);
				if (error_callback) { return error_callback(req,0,"Unable to contact server - Check your connection."); }
				return;
			    }
			    if (error_callback) { error_callback({status:0}, 0, "Unable to contact server - Check your connection.");  }
			    return;
			}
		    };
		    JV3.log("ZZZZZZ send", url);
		    req.send(serialized !== undefined ? serialized : null);
		} catch(e) {
		    JV3.log(e);
		    JV3.log("synchronize.ajaxJSON " + e + " " + e.lineNumber + " " + e.fileName);
		    error_callback("ajaxJSON XMLRPC - connection error.");
		}	    
	    },
	    updateAll: function(callback, error_cb) {
		// this.fixupDeleted(); // take me out when all notes have a deleted field
		var this_ = this;
		this._call_notes_ajax(
		    "notes", this.methods.read,
		    null, // no note for GET
		    function(remote_data) {
			var remote_notes = this_._translate_dj2js(remote_data);
			return this_._reconcileNotes(remote_notes, callback);
		    },
		    error_cb);		    
	    },
	    _serverHas:function(id, server_notes) {
		// checks to see if server contains note with the given id
		// in the server_notes list. these notes should be in the client schema
		for (var i = 0; i < server_notes.length; i++) {
		    if (id == server_notes[i].id) {
			return server_notes[i];
		    }
		}
		return undefined;
	    },
	    commitAll: function(success_cb, error_cb) {
		// performs a reconcile and then a commit, with the 
		// following semantics:
		// for each note that is modified on the client
		//   try to post it.
		// this.fixupDeleted(); // take me out when all notes have a deleted field
		var this_ = this;
		this._call_notes_ajax( "notes",
				       this_.methods.read,  null,
				       function(remote_data) {
					   log("commitall -- cont");
					   var remote_notes = this_._translate_dj2js(remote_data);
					   // now run local update, and then commit all of em.
					   this_._reconcileNotes(remote_notes,
								 function(updated_notes) {	
								     log("commitall -- reconcileNotes complete");

								     var server_ids = remote_notes.map(function(x) { return x.id; });
								     var pending = plumutil.objVals(JV3.notes).filter(function(n) {  return (n.modified || server_ids.indexOf(n.id) < 0);});
								     JV3.log("pending length " + pending.length/(1.0*plumutil.objVals(JV3.notes).length));
								     var translated = this_._translate_js2dj(pending);								     
								     log('notes.to_commit is ' + pending.length + " : " + pending.map(function(x) { return x.id; }).join(','));

								     var committed = [];
								     var errored = [];
								     var error_msgs = [];								     
								     var method = "POST";
								     
								     var checkpending = function() {
									 // determines whether we should call the errorcb or the success cb, or just one cb!
									 // checks pending queue to see if we've received all our callbacks
									 // either via error or whatever
									 if (pending.length > 0) {
									     return;
									 }
									 log("Sync: Nothing pending so continuing... ");
									 var reupdate_cont = function() {
									     log("Finished reupdating .. next!... ");
									     if (errored.length > 0 && error_cb !== undefined) {
										 log("Errors encountered, so calling error cont " + errored.length);
										 return error_cb(error_msgs.join(", "),
												 { commmitted_ids: committed.map(function(n) { return n.id; }), errored_ids: errored.map(function(n) { return n.id; })});
									     }
									     if (success_cb !== undefined) {
										 log("No errors encountered, so calling succes cont " + errored.length);
										 return success_cb(updated_notes,
												   { commmitted_ids: committed.map(function(n) { return n.id; }), errored_ids: errored.map(function(n) { return n.id; })});
									     }
									 };
									 // final update to get our version #s
									 this_.updateAll(reupdate_cont,reupdate_cont);
								     };			    								     
								     this_._call_notes_ajax("notespostmulti/",
											    method, translated,
											    function(response,code) {
												response.committed.map(
												    function(noteresp) {
													if ( [201,200].indexOf(noteresp.status) >= 0 )  {
													    var n = getNoteById(noteresp.jid);
													    n.modified=false;
													    this_.store.save(n);
													    committed.push(n);
													    pending = pending.filter(function(x) { return x.id !== n.id; });
													} else {
													    errored.push(n);
													}
												    });
												checkpending();												
											    },
											    function(status,code) {								  
												if (this_.sync_status_to_message[code] !== undefined) {
												    error_msgs.push(this_.sync_status_to_message[code]);
												}
												errored = pending;
												pending = [];
												checkpending();
											    });
								     checkpending(); // because we may have never executed a post if no modified
								 });
				       },  function(status,code) { /* error handler of GET phase */			
					   log("ERROR handler commit.GET phase  " + status + " code " + code);
					   if (this_.sync_status_to_message[code] !== undefined) {
					       if (error_cb !== undefined) {
						   return error_cb(this_.sync_status_to_message[code],code);
					       }
					       return;
					   }
					   if (error_cb !== undefined) {
					       return error_cb(status,code);
					   }
					   return undefined;
				       });

	    },

	    _reconcileNotes_helper:function(data,success_cb) {
		var this_ = this;
		var updated_notes = [];
		var store = this.store;
		JV3.plumutil.interval_map(data,function(remote_n) {
					      var id = remote_n.id;
					      try {
						  var local_n = JV3.getNoteById(id);
						  if (local_n && remote_n && local_n.version == remote_n.version)  { return; }		    
						  var verified = _verify_and_construct_note(remote_n,local_n || this_.note_default_values);						  
						  if (local_n === undefined) {
						      // NEW NOTE from REMOTE.
						      if (!verified.is_complete) {log("Warning: New note incomplete " + verified.note.id + " missing fields: " + plumutil.objVals(verified.missing_fields) );}
						      if (!verified.is_patched_complete) {
							  log("Error: Could not reconstruct note " + (verified.note.id) + "epic fail! ");
							  return;
						      }
						      updated_notes.push(verified.note);
						      JV3.remoteAddNote(verified.note);
						  } else if (local_n.version < verified.note.version) {
						      // UPDATE our local note
						      if (local_n.modified) {
							  // if our copy has been modified, sew the changes togetheer
							  local_n.text = this_.merge(local_n, verified.note);
						      } else {
							  // else whole sale replace the text
							  local_n.text = verified.note.text;
						      }
						      // either way update the version
						      local_n.version = verified.note.version;
						      local_n.deleted = verified.note.deleted; 
						      updated_notes.push(local_n);
						      if (local_n.deleted) {
							  remoteDelete(local_n); // does saving
						      } else {
							  remoteUpdate(local_n); // does saving
						      }
						  } else {
						      //log("note is apparently up to date : " + remote_n.id  + " local " + local_n.version + " remote " + remote_n.version);
						  }
					      } catch(e) {  JV3.log(e);	      }
					  },
					  function() {
					      JV3.log("successfully reconciled notes : ", updated_notes.length);
					      success_cb(updated_notes) ;
					  },
					  function(x) { JV3.log(x); },
					  10,
					  JV3);
	    },
	    
	    _reconcileNotes: function(data,success_cb) {
		// takes data received from the server and compares it with the
		// notes we have in our local kb and makes appropriate updates
		// with the following semantics:
		// for each note
		//   if the version from the server is newer (sn)
		//      if the corresponding local note (ln) is modified
		//             ln.body <- merge(sn.body,ln.body)
		//             ln.version <- sn.version
		//      else ln <- sn
		//   if the version from the server is the same, do nothing.
		//   if the version from the server is OLDER blow up the world.
		log("reconcileNotes()");
		var this_ = this;

		var special = data.filter(function(x) { return x.id < 0; });
		var regular = data.filter(function(x) { return x.id >= 0; });
		this._reconcileNotes_helper(special,function(updated_notes) {this_._reconcileNotes_helper(regular, function(updated_notes2) { success_cb(updated_notes.concat(updated_notes2));});   });
	    },
	    get_special_merge_handler:function(n) {
		return ({
		    "-1" : function(n1,n2) { return parseInt(n1.edited) > parseInt(n2.edited) ? n1.text : n2.text; }
		})[n.id];
	    },
	    merge:function(local, remote) {
		if (this.get_special_merge_handler(local) !== undefined) {
		    return this.get_special_merge_handler(local)(local,remote);
		}
		return local.text + "\n" + remote.text;
		// emax changed this on 9.1.2008 because diffString wasn't properly reversing -- causing infinite blowups
		// return diffString(local, remote);
	    },
	    // _unpack_response: function(data) {
	    // 	return plumutil.fromJSON(data).map(
	    // 	    function(datum) {
	    // 		if (datum.fields) { return datum.fields; }
	    // 		return datum;
	    // 	    }
	    // 	); 
	    // },
	    _commit_act_entries: function(url,to_commit,success_cb,error_cb) {
		// COMMIT
		var this_ = this;
		this.ajaxJSON(this_.server_url + url, "POST", to_commit, 
			      function(data,success) {
				  try {
				      var unpacked = typeof(data) == 'object' ? data : plumutil.fromJSON(data);
				      if (success_cb) { success_cb(unpacked); }
				  } catch (e) {
				      log(e + " " + e.lineNumber + " " + e.fileName);
				      error_cb(data, "json_parse");
				  }
			      },
			      function(xmlreq,code,response) {
				  try {
				      if ( this_.sync_status_to_message[code] !== undefined ) {
					  return error_cb(this_.sync_status_to_message[code], code);
				      }
				      return error_cb(response,code);
				  } catch (e) {
				      log(e + " " + e.lineNumber + " " + e.fileName);
				  }
			      });
	    },
	    _setLastCommittedActLogId:function(id) {
		this.actlog_last_entry_id = id;
	    },
	    _getLastCommittedActLogId:function() {
		return this.actlog_last_entry_id;
	    },	    
	    __find_old_uncomitted_activity_logs:function(newstore,last_committed_id)  {
		//  rather inefficient ---> loading each element one at a time
		var batch_size = this.activitylog_bucket_size; 
		//var batch_size = undefined;
		try {
		    //JV3.log("__find_old_uncomitted_activity_logs " + last_committed_id);
		    var min_new_id = newstore.getMinID();
		    //JV3.log("min_new_id " + min_new_id);
		    if (min_new_id === undefined || min_new_id.length == 0 || last_committed_id < min_new_id[0]) {
			// we are either just upgrading from an old client or
			JV3.log("LEGACY --------------- Retrieving old tables from old client version... ");
			var logtables = JV3.Old_ActLog.getActivityLogTablesNewerThan(last_committed_id);
			var to_commit = [];
			logtables.map(function(table) {
			    //JV3.log("considering table : " + table);
			    if (batch_size !== undefined && to_commit.length >= batch_size) { return; } // bucket full, die
			    var db = JV3.Old_ActLog.openActivityLog(table);
			    var maxid = db.getMaxID();			
			    if (maxid && last_committed_id > maxid) {
				// already committed, skip this one.
				//JV3.log("skipping this one");
				return;
			    }
			    //JV3.log("examining table : " + table);
			    var entries = db.getIDs().filter(function(x) { return x > last_committed_id; });			
			    entries.map(function(actlogid) {
				if (batch_size === undefined || to_commit.length < batch_size) {
				    var entry = db.loadbyID(actlogid);
				    if (entry !== undefined) {
					//JV3.log("adding : " + actlogid);
					to_commit.push(entry);
				    }			
				}
			    });
			});
			//JV3.log("Retrieved " + to_commit.length + " old records from old activity logs");
			return to_commit;
		    }
		    return [];
		} catch(E) {
		    JV3.log(E);
		}
		return [];
	    },

	    commitActivityLog:function(success_cb,error_cb) {
		var this_ = this;
		var astore = this.activitystore;
		var url = "/jv3/notelog/";


		var error_msgs = [];
		var error_codes = [];
		var to_commit = [];

		var all_committed = [];
		var all_errored = [];

		var check_cont = function() {
		    this_._setLastCommittedActLogId(Math.max.apply(this_,to_commit.map(function(x) { return x.id; })));
		    JV3.log("ALOG.check_cont ", to_commit.length, " ", all_committed.length, " ", all_errored.length);
		    if (to_commit.length == (all_committed.length + all_errored.length)) {
			JV3.log("ALOG: All ducks accounted for: commit complete at " + new Date() + " " + to_commit.length, all_errored);
			if (all_errored.length === 0) {			 
			    success_cb(all_committed);
			} else {
			    error_cb(error_msgs.join(","), error_codes[0], {success:all_committed,errored:all_errored});
			}
		    }
		};
		
		var part2 = function(last_entry_id, n_remote_logs) {
 		    try {
			log("activity log starting with last entry " + last_entry_id);
			if (last_entry_id === undefined) { last_entry_id = 0; }
			var part2_callee = arguments.callee;
			// retrieve last things
			// to_commit = []; // DISABLED OLD COMMIT this_.__find_old_uncomitted_activity_logs(astore, last_entry_id);
			//			if (to_commit !== undefined && to_commit.length == 0) {
			// new act 2.0 log code
			var __bench__start_load = new Date().valueOf();
			var chunk_size = 256;
			astore.loadIdsGreaterThan(last_entry_id, function(local_alogs) {
						      to_commit = local_alogs; // update parent 
						      JV3.log("Took " + (new Date().valueOf() - __bench__start_load) + " to load from " + last_entry_id);
						      astore.countAsync(function(n_local_logs) {
									    try {
										JV3.log("activity logs to commit : " + to_commit.length + " n_local_logs " + n_local_logs + "n_remote: ", n_remote_logs, " n_remote_logs+commit " + (n_remote_logs + to_commit.length));
										if ( n_remote_logs !== undefined && last_entry_id > 0 && n_remote_logs + to_commit.length < n_local_logs ) {
										    JV3.log("___________________ MISSING ALOGS, STARTING FROM SCRATCH ");
										    return part2_callee(0,n_remote_logs);										    
										}
										JV3.log("_actlog_ chunks ++ ", to_commit.length/chunk_size);
										JV3.plumutil.interval_map_chunked(to_commit,
														  function(batch) {														  
														      this_._commit_act_entries(url,batch,
																		function() {
																		    JV3.log("_commited batch ++ ", batch.length);
																		    all_committed = all_committed.concat(batch);
																		},
																		function(status,code) {
																		    // what do we do?
																		    try {
																			JV3.log("_actlog errored batch ++ ", batch.length, status, code);
																			error_msgs.push(status);
																			error_codes.push(code);
																			all_errored = all_errored.concat(batch);
																		    } catch (x) { JV3.log(x);  }
																		});
														  },
														  function() {  JV3.log("Checking cont -- "); check_cont(); },
														  function(x) { JV3.log(x); },
														  5000, /* seconds to wait .. turn this up */
														  chunk_size,  /* records per chunk */
														  JV3);
														  
														  
								
									    } catch (x) { JV3.log(x);   }
									    return undefined;
									});
						  });
		    } catch(ee) {
			JV3.log(ee);
			log(ee + " " + ee.lineNumber + " " + ee.fileName);
		    }
		};

		if (this_._getLastCommittedActLogId() !== undefined) {
		    JV3.log("skipping actlog call with last " + this_._getLastCommittedActLogId());
		    part2(this_._getLastCommittedActLogId());
		}
		
		// find out what the _last_ log entry we sent was by asking the server.
		this_.ajaxJSON(this_.server_url + url + "?type=get_max_log_id&client="+JV3.getClientId(), "GET", null,
			       function(data,success) {
				   try {
				       var obj = typeof(data) == 'object' ? data : plumutil.fromJSON(data);
				       var n_remote_logs = obj.num_logs;
				       JV3.log("setting n_remote_logs to ",obj.num_logs);
				       return part2(obj.value,obj.num_logs);							     
				   } catch (e) { log(e);  error_cb(data, "json_parse");  }
				   return undefined;
			       }, function(xmlrpc,code,status) {
				   try {
				       if (code == 404)  {
					   // then send the whole thing, please
					   return part2(0,0);
				       };
				   } catch(e) {  JV3.log(e);  }
				   return undefined;
			       });
	    },    
	    _call_notes_ajax: function(url,method, notes, cb, error_cb) {
		// only used for "/notes and /notepostmulti
		var this_ = this;

		var fj = function(data) { return typeof(data) == 'string' ? plumutil.fromJSON(data) : data; };
		
		//log("_call_notes_access start()------>");
		// notes === null ? null:
		// notes.map(function(note) {
		//     // have to manually add the owner here
		//     var owned = plumutil.clone(note);
		//     owned.owner = this_.owner_id;
		//     return owned; 
		// }
		//),
		
		try {
		    this.ajaxJSON(this.url+url,method, notes, 
				  function(data,success) {
				      try {
					  var unpacked = fj(data);
					  if (unpacked.map) {
					      unpacked = fj(data).map( function(datum) {
						  if (datum.fields) { return datum.fields; }
						  return datum;
					      });
					  } 
					  cb(unpacked);
				      } catch (e) {
					  log(e + " " + e.lineNumber + " " + e.fileName);
					  error_cb(data,"json_parse");
				      } 
				  },
				  function(xmlreq,code,response) {  error_cb(response,code);});
		} catch(ee) {
		    log(ee + " " + ee.lineNumber + " " + ee.fileName);
		}
	    },
	    _translate_dj2js: function(data) {
		return plumutil.map( function(n) {
		    return {
			id: n.jid,
			text: n.contents,
			edited : n.edited,
			created : n.created,
			version : n.version,
			modified: false,
			deleted : (n.deleted == 1)
		    };
		}, data);
	    },
	    _translate_js2dj: function(data) {
		return data.map( function(n) {
		    return {
			jid: n.id,
			contents: n.text,    
			edited : n.edited,  // edited time
			created : n.created, // created time
			version : n.version,
			deleted : n.deleted == 'true' || (!isNaN(parseFloat(n.deleted)) && parseFloat(n.deleted) > 0) || n.deleted === true ? true : false
		    };
		});
	    },

	    /** event **/ 
	    
	    /** new architecture for posting events **/	    
	    __entity_cache:{},
	    encrypted_entity_data_blacklist:["schemas.Location"],

	    get_max_events_url:"/jv3/eventlog/?type=get_max_log_id&client="+JV3.getClientId(),
	    post_events_url: "/jv3/post_events/",	    

	    get_max_event:function(success_cb, failure_cb) {
		var this_ = this;
		var url = "/jv3/eventlog/";
		this.ajaxJSON(this_.server_url + url + "?type=get_max_log_id&client="+JV3.getClientId(), "GET", undefined,
			      function(server_return_response){
				  if (typeof(server_return_response) == 'string') {
				      server_return_response = JV3.plumutil.fromJSON(server_return_response);
				  }
				  if (success_cb){ success_cb(server_return_response.value);   }
			      },
			      function(req,status,text) {
				  if (status == 404)  { JV3.log("NO event logs found -------->, returning success(0) "); return success_cb(0); }
				  JV3.log("GETMAXEVENTError -- " + req + " -- " + status + " -- " + text);
				  JV3.log(req);				  
				  if (failure_cb) { return failure_cb(req,status,text); }				  
			      });
	    },	    	     
	    get_entity_data : function(event) {
		if (isnull(event) || isnull(event.entityid)) {
		    JV3.log("warning - " + isnull(event) ? " event is undefined " : (" event entity id is undefined " + event.type + " " + event.start + " " + event.end));
		    return undefined;
		}
		if (JV3.doEncryptActivities() && this.encrypted_entity_data_blacklist.indexOf(event.entitytype) >= 0) { return "";  }
		if (!this.__entity_cache[event.entityid]) {
		    var entity = JV3.CMS.entity_store.getEntityByIdRaw(event.entityid);
		    if (entity !== undefined) {
			this.__entity_cache[event.entityid] = JV3.plumutil.toJSON(entity);
		    }			
		}
		return this.__entity_cache[event.entityid];
	    },	    
	    export_entity:function(raw_events) {
		var this_ = this;
		var client_id = JV3.getClientId();			
		raw_events.map(function(x) {
				 x.client = client_id;
				 try { x.entitydata = this_.get_entity_data(x); } catch(e) { JV3.log(e);  }
			     });
	    },
	    sync_events:function(success_cb,failure_cb) {
		var this_ = this;
		var url = "/jv3/post_events/";
		var event_store = JV3.CMS.event_store;
		// this code shamefully taken from listit
		if (this._get_last_committed_event() === undefined) {
		    return this.get_max_event(function(data) {
						  JV3.log("------------------- || sync_events() : setting last_committed_event_time to " + data);
						  this_._set_last_committed_event(data);
						  this_.sync_events(success_cb,failure_cb);
					      },
					      failure_cb);
		}
		try {
		    var new_logs = event_store.getEventsRawTime(undefined, [this._get_last_committed_event() + 1, new Date().valueOf()]);
		    this.log("event logs to commit: " + new_logs.length);
		    if (new_logs.length == 0) { return success_cb([]);   }
		    this.export_entity(new_logs);
		    this_.ajaxJSON(this_.server_url + url, "POST", new_logs,
				   function(server_return_response) {
				       try {
					   JV3.log("server response: " + server_return_response);
					   // unpack.
					   if (typeof(server_return_response) == 'string') {  server_return_response = JV3.plumutil.fromJSON(server_return_response);  }
					   // find committed
					   var committed = server_return_response.committed;
					   if (committed && committed.length > 0) {
					       this_._set_last_committed_event(Math.max.apply(this_,committed.map(function(x) { return x; })));
					       JV3.log("setting last entry: " + this_._get_last_committed_event());
					   }
					   if (success_cb) { success_cb(committed); }						   
				       } catch (x) {
					   JV3.log(x);
				       }
				   },failure_cb);
		    
		} catch (x) {
		    JV3.log(x);
		}
	    },

	    
	    // times,
	    _set_last_committed_event:function(last) { JV3.Synchronize._last_committed_event = last;   },
	    _get_last_committed_event:function(last) { return JV3.Synchronize._last_committed_event; },
	    _add_event_update_listener:function() {
		var this_ = this;
		if (JV3.CMS && JV3.CMS.event_store) {
		    JV3.CMS.event_store.addStoreUpdateListener(undefined,function(event) {
								   try {
								       if (this_._get_last_committed_event() !== undefined && event.start <= this_._get_last_committed_event()) {
									   JV3.log("sync ~ event update");
									   this_._set_last_committed_event( event.start - 1 );    
								       }								       								       
								   } catch (x) { JV3.log(x);   }
							       });
		}		    
	    },	    
	    
	    run_diagnostics : function(success_fn,fail_fn) {
		// sends diagnostic information even without login for purposes of troubleshooting
		// no login needed		
		if (this.ajaxJSON === undefined) {
		    // we're running on Sync
		    this.define_ajaxjson();
		}
		var this_ = this;
		var server_url = JV3.prefs.getCharPref("sync.server");
		var info = JV3._gen_diagnostics();
		var url = "/jv3/post_diagnostics/";
		this_.ajaxJSON(server_url + url, "POST", [info], function() {
				   JV3.log("success posting diagnostic.");
				   if(success_fn) { success_fn(); }
			       }, function(xmlrpc,code,status) {
				   JV3.log("error posting diagnostic: " + code + " " + status );
				   if(fail_fn) { fail_fn.apply(this_, arguments); }
			       });
	    }
	};
});


	    // /// old architecture for posting events **/
	    // commitEventLog:function(success_cb,error_cb) {
	    // 	// POYOZO/PRUNE code
	    // 	// "events" are user activity observations, un-related to listit
	    // 	if (JV3.CMS === undefined || JV3.CMS.event_store === undefined) {
	    // 	    log("Warning: Context Maiden is not enabled " + JV3.CMS);
	    // 	    return;
	    // 	}		
	    // 	var this_ = this;
	    // 	var event_store = JV3.CMS.event_store;
	    // 	var url = "/jv3/eventlog/";

	    // 	var error_msgs = [];
	    // 	var error_codes = [];
	    // 	var to_commit = [];

	    // 	var all_committed = [];
	    // 	var all_errored = [];

	    // 	var get_last_committed_event_log_start = function() { return this_.__last_committed_event_log_start; };
	    // 	var set_last_committed_event_log_start = function(time) { return this_.__last_committed_event_log_start = time; };

	    // 	var check_cont = function() {
	    // 	    if (to_commit.length === all_committed.length + all_errored.length) {
	    // 		if (all_errored.length === 0) {
	    // 		    success_cb(all_committed);
	    // 		} else {
	    // 		    error_cb(error_msgs.join(","), error_codes[0], {success:all_committed,errored:all_errored});
	    // 		}
	    // 	    }
	    // 	};


	    // 	/* add entity data to the event */
	    // 	var __entity_cache = {};

	    // 	var encrypted_entity_data_blacklist = ["schemas.Location"];
		
	    // 	var get_entity_data = function(event) {
	    // 	    if (event === undefined || event.entityid === undefined) { return; }
	    // 	    if (JV3.doEncryptActivities() && encrypted_entity_data_blacklist.indexOf(event.entitytype) >= 0) {
	    // 		 return "";
	    // 	    }
	    // 	    if (__entity_cache[event.entityid] === undefined) {
	    // 		var entity = JV3.CMS.entity_store.getEntityByIdRaw(event.entityid);
	    // 		//JV3.log(" ------> sending --------> " + JV3.plumutil.toJSON(entity));
	    // 		if (entity !== undefined) {
	    // 		    __entity_cache[event.entityid] = JV3.plumutil.toJSON(entity);
	    // 		}			
	    // 	    }
	    // 	    return __entity_cache[event.entityid];
	    // 	};
		
	    // 	var part2 = function(last_committed_event_time) {
 	    // 	    try {
	    // 		if (last_committed_event_time === undefined) { last_committed_event_time = 0; }
	    // 		//log("EVENT last_id : " + last_committed_event_time + " " + new Date(last_committed_event_time));			
	    // 		// new act 2.0 log code
	    // 		var client_id = JV3.getClientId();
			
	    // 		var new_logs = event_store.getEventsRawTime(undefined, [last_committed_event_time, new Date().valueOf()]);
	    // 		new_logs.map(function(x) {
	    // 		    x.client = client_id;
	    // 		    try {
	    // 			x.entitydata = get_entity_data(x);
	    // 			//JV3.log(x.entitydata);
	    // 		    } catch(e) {
	    // 			JV3.log(e);
	    // 		    }
	    // 		});
			
	    // 		JV3.log("event logs to commit : " + new_logs.length);
	    // 		to_commit = to_commit.concat(new_logs);			
	    // 		//JV3.log("calling commit event on " + to_commit.length);
	    // 		this_._commit_act_entries(url,to_commit,
	    // 					  function(batch) {
	    // 					      return function() {
	    // 						  // log("SUCCESS -> _commit_event_entries");
	    // 						  all_committed = all_committed.concat(batch);
	    // 						  if (batch && batch.length > 0) {
	    // 						      set_last_committed_event_log_start(Math.max.apply(this_,batch.map(function(x) { return x.start; })));
	    // 						  }
	    // 						  check_cont();
	    // 					      };
	    // 					  }(to_commit),						      
	    // 					  function(batch) {
	    // 					      return function(status,code) {
	    // 						  // what do we do?
	    // 						  log(status);
	    // 						  error_msgs.push(status);
	    // 						  error_codes.push(code);
	    // 						  all_errored = all_errored.concat(batch);					      
	    // 						  if (batch && batch.length > 0 && parseInt(code) >= 500) {
	    // 						      set_last_committed_event_log_start(Math.max.apply(this_,batch.map(function(x) { return x.id; })));
	    // 						  }					      
	    // 						  check_cont();
	    // 					      };
	    // 					  }(to_commit));
	    // 	    } catch(ee) {
	    // 		JV3.log(ee);
	    // 		log(ee + " " + ee.lineNumber + " " + ee.fileName);
	    // 	    }
	    // 	};
		
	    // 	if (false && get_last_committed_event_log_start() !== undefined) {
	    // 	    // express train 
	    // 	    part2(get_last_committed_event_log_start());
	    // 	} else {
	    // 	    // find out what the _last_ log entry we sent was by asking the server.
	    // 	    this_.ajaxJSON(this_.server_url + url + "?type=get_max_log_id&client="+JV3.getClientId(), "GET", null,
	    // 			   function(data,success) {
	    // 			       try {
	    // 				   var obj = plumutil.fromJSON(data);
	    // 				   log("Server sez last EVENT log received was : " + obj.value);
	    // 				   set_last_committed_event_log_start(obj.value);
	    // 				   part2(obj.value);
	    // 			       } catch (e) {
	    // 				   log(e);
	    // 				   error_cb(data, "json_parse");
	    // 			       }
	    // 			   },
	    // 			   function(xmlrpc,code,status) {
	    // 			       try {
	    // 				   if (code == 404)  {
	    // 				       // the server HAS no activity logs, so we have to send it some
	    // 				       return part2();
	    // 				   }			       
	    // 				   log("Activitylog Errorcb " + status + " code : " + code);
	    // 				   if ( this_.sync_status_to_message[code] !== undefined ) {
	    // 				       return error_cb(this_.sync_status_to_message[code], code);
	    // 				   }
	    // 				   return error_cb(status,code);					   
	    // 			       } catch (e) {
	    // 				   log(e + " " + e.lineNumber + " " + e.fileName);
	    // 			       }
	    // 			   });
	    // 	}
	    // },	    
