(function(JV3) {
     
     JV3.PlumTripleStore = {
	 superclass:PlumStore,
	 TRIPLESTORE_SCHEMA : "(subj varchar, pred varchar, obj varchar)",
	 _formatSubj : function(x) { return ""+x.valueOf(); },
	 _formatPred : function(x) { return ""+x.valueOf(); },
	 _formatObj  : function(x) {
	     var sobj = undefined;
	     if (x === undefined || x === null) { return; }
	     if (this.plumutil.isArrayLike(x)) {
		 // invoke array saving routine
		 sobj = this._saveArray(x);
	     } else if (typeof(x) == 'object' && x.id) {
		 // invoke object saving routine
		 sobj = this.save(x);
	     } else {
		 // primitive saving routine 
		 // must be a primitive type (string, int, ..)
		 sobj = this.serializer.serialize(x);
		 if (sobj === undefined) {
		     sobj = this.serializer.serialize(x,this.PlumStoreTypeNames.STRING);
		 }
	     }
	     return sobj;
	 },
	 _deformatSubj : function(x) { return x; },
	 _deformatPred : function(x) { return x; },
	 _deformatObj  : function(x) {
	     if (x === undefined || x === null) return;
	     //this.log('trying ' + x + "is Serialized Seq? " + this.serializer.isSerializedType(x,PlumStoreTypeNames.SEQ));
	     if (this.serializer.isSerializedType(x,this.PlumStoreTypeNames.SEQ)) {
		 var sid = this.serializer.deserialize(x,this.PlumStoreTypeNames.SEQ);
		 var arr = this.loaders[this.PlumStoreTypeNames.SEQ](this.loadbyID(sid));
		 return arr;
	     } else if (this.serializer.isSerializedType(x,this.PlumStoreTypeNames.RESOURCE)) {
		 var sid = this.serializer.deserialize(x,this.PlumStoreTypeNames.RESOURCE);
		 var obj = this.loadbyID(sid);
		 return obj; 
	     } else {
		 var sid = this.serializer.deserialize(x);
		 if (sid !== undefined) {
		     return sid;
		 } 
	     }
	     this.plumutil.assert(false, "Warning, could not deserialize " + x);
	     return undefined;
	 },    
	 initialize: function(params) {
	     JV3.PROTOCROCK.superclass(this, "initialize", params); // this.super.initialize.apply(this.super,arguments);
	     this.serializer = JV3.newify(this.PlumStoreTypeHandler);
	     this.table = params.table;
	     //this.log('initializing with table ' + this.table);
	     try {
		 this.createTable(this.table);
	     } catch(e) { /* error might occur if table already exists */ }
	     this.loaders[this.PlumStoreTypeNames.SEQ] = function(obj) {
		 // unpack an array
		 var toret = [];
		 var i_ = 0;
		 while (true) {
		     var field = this.PlumStoreTypeNames.SEQ+i_;
		     if (obj[field]) {
			 toret.push(obj[field]);
		     } else {
			 break;
		     }
		     i_ = i_ + 1;
		 }
		 return toret;
	     };
	 },
	 
	 // object saving routine
	 save : function (obj) {
	     //this.log('save object: ' + obj + " with keys: " + this.plumutil.keys(obj));
	     var sqlstmt = undefined;
	     try {
		 this.plumutil.assert(obj.id !== undefined, "obj to serialize  must have an id" + obj);
		 var serializer = this.serializer;					 
		 var table = this.table;
		 var subj = this._formatSubj(obj.id);
		 
		 this._db.beginTransaction();
		 // delete all with subj
		 this.execute('DELETE FROM ' + table + ' WHERE subj=\"'+subj+'\";');
		 for (var k_i = 0; k_i < this.plumutil.keys(obj).length; k_i++) {
		     var field = this.plumutil.keys(obj)[k_i];
		     var pred = this._formatPred(field);
		     this.plumutil.assert(obj[field] !== undefined && obj[field] !== null, "Key " + pred + " was null/undefined");
		     var sobj = this._formatObj(obj[field]);					     
		     var values = "(\""+subj+"\",\""+pred+"\",\""+sobj+"\")";
		     sqlstmt = 'INSERT INTO ' + table + ' VALUES ' + values + ';';
		     // reference is already created, now we just have to serialize IT
		     this.execute(sqlstmt);					     
		 }
		 this._db.commitTransaction();
		 return this.serializer.byType[this.PlumStoreTypeNames.RESOURCE].serialize(obj.id);
	     } catch (e) {
		 this.log("PLUMSTOREERROR: " + e + " " + e.lineNumber + " " + e.fileName);
		 this.log(sqlstmt);
		 //alert(sqlstmt);
		 this._db.rollbackTransaction();
	     }
	     return undefined;
	 },
	 // array saving routine
	 _saveArray: function(obj) {
	     // obj is an array
	     var subj  = this.plumutil.guid(10); //this.serializer.byType[PlumStoretypeNames.SEQ].serialize(obj);
	     for (var k_i = 0; k_i < obj.length; k_i++) {
		 var pred = this.PlumStoreTypeNames.SEQ+k_i;
		 var sobj=  this._formatObj(obj[k_i]);
		 var values = "(\""+subj+"\",\""+pred+"\",\""+sobj+"\")";
		 var sqlstmt = 'INSERT INTO ' + this.table + ' VALUES ' + values + ';';
		 this.execute(sqlstmt);					     
	     }					 
	     return this.serializer.byType[this.PlumStoreTypeNames.SEQ].serialize(subj);
	 },
	 loaders : {},
	 loadbyID : function(id) {
	     var SQLQUERY = "SELECT pred,obj FROM " + this.table + " WHERE subj=\""+id+"\";";
	     this.log(SQLQUERY);
	     var toret = {};
	     var statement = this._db.createStatement(SQLQUERY);
	     try {            
		 while(statement.executeStep()) {
		     var pred = this._deformatPred(statement.getString(0));
		     var obj = this._deformatObj(statement.getString(1));
		     toret[pred] = obj;
		 }
	     } catch (e) {
		 this.log("error executing statement " + statement + " " + e); //alert(e);
	     } finally {
		 statement.reset();
	     }
	     // if we have a type-specific loader module, invoke that to de-serialize the craziness.
	     if (toret[this.PlumStoreTypeNames.TYPE] && this.loaders[toret[this.PlumStoreTypeNames.TYPE]]) {
		 return this.loaders[toret[this.PlumStoreTypeNames.TYPE]](toret);
	     }
	     return (this.plumutil.keys(toret).length > 0) ? toret : undefined;
	 },
	 deletebyID: function(ids) {
	     if (!this.plumutil.isArrayLike(ids)) {
		 ids = [ids];
	     }
	     this.plumutil.assert(ids.length > 0, "Must provide at least 1 id. Otherwise use clear()");
	     var where_clause = "";
	     var OR = " OR ";
	     for (var i = 0; i < ids.length; i++) {
		 where_clause = where_clause + "subj=\""+this._formatSubj(ids[i])+"\""+OR;
	     }
	     var SQLQUERY = 'DELETE FROM ' + this.table + ' WHERE ' + where_clause.substring(0,where_clause.length-OR.length)+";";
	     this.execute(SQLQUERY);
	     return ids.length > 0;					 
	 },
	 clear: function() {
	     var SQLQUERY = 'DELETE FROM ' + this.table + ";";
	     this.execute(SQLQUERY);
	 },
	 _field_conditions:function(val) {
	     if (val === undefined) { return []; } 
	     if (typeof(val) !== "object") {
		 return [ val ] ; // scalar
	     }
	     if (!this.plumutil.isArrayLike(val)) {
		 // single criterion
		 return [ val ];
	     }
	     return val;	
	 },
	 _fieldmatch_comparator:function(val) {
	     if (typeof(val) == 'object' && val.operator !== undefined) {
		 return val.operator;
	     }
	     return '=';
	 },
	 _fieldmatch_target:function(val) {
	     if (typeof(val) == 'object' && val.target !== undefined) {
		 return val.target;
	     }
	     return val;
	 },
	 fieldmatch:function(operator,value) {
	     // this is specific to a _field_ e.g., objects should be like { 'id' : fieldmatch('>',123) }
	     return ({operator:operator, target:value });
	 },
	 find: function(criteria) {
	     // returns a list of IDs for stored objects that match a conjunction
	     // of properties expressed as an object, e.g.,
	     // f = new PlumTripleStore('test','bladerunner');
	     // f.find({name:"Decker",type:"Bounty Hunter"})
	     // results in:
	     //   SELECT a0.subj FROM bladerunner a0, bladerunner a1 WHERE a0.pred="name"
	     //     AND a0.obj="Decker^^#string" AND a1.pred="type" AND a1.obj="Bounty Hunter^^#string" 
	     //     AND a0.subj=a0.subj AND a0.subj=a1.subj;


	     var variables = [];
	     var criteria_str = "";
	     var table_str = "";
	     var where_id_eqs_str ="";
	     var where_obj_criteria_str = "";
	     var AND = " AND " ;

	     var SQLQUERY = undefined;

	     if (criteria !== undefined) {
		 for (var f_i in criteria) {
		     if (criteria.prorotype == undefined || criteria.prototype[f_i] === undefined) {
			 var pred  = this._formatPred(f_i);
			 // if val is a primitive then this gets serialized
			 // otherwise serialize returns us a reference anyway
			 // allocate a new variable
			 var sqlvar  = "a"+variables.length;
			 table_str = table_str + this.table + " " + sqlvar + ", ";		
			 where_obj_criteria_str = where_obj_criteria_str + sqlvar+".pred=\""+pred+"\""+AND;
			 where_id_eqs_str = where_id_eqs_str + "a0.subj="+sqlvar+".subj"+AND;
			 variables.push(sqlvar);
			 
			 // deal with the object clauses
			 var pred_constraints = this._field_conditions(criteria[f_i]);
			 self.pc = pred_constraints;
			 for (var c_i = 0; c_i < pred_constraints.length; c_i++) {
			     var comparator = this._fieldmatch_comparator(pred_constraints[c_i]);
			     var sobj  = this.serializer.serialize(this._fieldmatch_target(pred_constraints[c_i]));
			     where_obj_criteria_str += sqlvar+".obj "+comparator+" \""+sobj+"\"" + AND;
			 }
		     }
		 }
		 var where_str = where_obj_criteria_str + where_id_eqs_str.substring(0,where_id_eqs_str.length-AND.length);
		 SQLQUERY = "SELECT a0.subj FROM " + table_str.substring(0,table_str.length-2) + " WHERE " + where_str + ";";
		 this.log(SQLQUERY);
	     } else {
		 this.log("unconstrained " + SQLQUERY);
		 SQLQUERY = "SELECT subj FROM " + this.table + ";";
	     }
	     
	     var hits = [];
	     this.log("SQL>>> " + SQLQUERY);
	     var statement = this._db.createStatement(SQLQUERY);
	     try {            
		 while(statement.executeStep()) {
		     hits.push(statement.getString(0));
		 }
	     } catch (e) {
		 this.log("error executing statement " + statement + " " + e); //alert(e);
	     } finally {
		 statement.reset();
	     }
	     return this.plumutil.uniq(hits);
	 },
	 load : function(criteria) {
	     // find()s objects according to a certain criteria (as documented by find)
	     // and loads all of them, returning an array of instantiated objects
	     // warning: can NOT handle loopy graphs, as JSON cannot either! will blow up the world

	     // some people call us when they mean to call loadbyID.
	     if (criteria !== undefined && typeof(criteria) != 'object') {
		 return this.loadbyID(criteria);
	     }
	     // otherwise pass it off to find
	     var hits = this.find(criteria);
	     var results = [];
	     for (var hit_i = 0; hit_i < hits.length; hit_i++) {
		 results.push(this.loadbyID(hits[hit_i]));
	     }
	     return results;
	 },
	 createTable: function(tname) {
	     JV3.PROTOCROCK.superclass(this, "createTable", tname, this.TRIPLESTORE_SCHEMA);	// this.super.createTable(tname,this.TRIPLESTORE_SCHEMA);
	 },
	 getIDs:function(where_clause) {
	     // returns all ids contained in this triple store
	     var query = "select distinct subj from " + this.table + (where_clause ? "WHERE " + where_clause : "") + ";";
	     var toret = [];	
	     var statement = this._db.createStatement(query);
	     try {            
		 while(statement.executeStep()) {
		     toret.push(this._deformatSubj(statement.getString(0)));
		 }
	     } catch (e) {
		 this.log("error executing statement " + statement + " " + e); //alert(e);
	     } finally {
		 statement.reset();
	     }
	     return toret;			
	 },
	 getMaxID:function(where_clause) {
	     // returns all ids contained in this triple store
	     var query = "select max(subj) from " + this.table + (where_clause ? "WHERE " + where_clause : "") + ";";
	     var toret = [];	
	     var statement = this._db.createStatement(query);
	     try {            
		 while(statement.executeStep()) {
		     toret.push(this._deformatSubj(statement.getString(0)));
		 }
	     } catch (e) {
		 this.log("error executing statement " + statement + " " + e); //alert(e);
	     } finally {
		 statement.reset();
	     }
	     return toret;			
	 }
     };
 })
