
STACK = [];

var primitives = {
    "Date" : function(d) {
	var wrapped = object(d);
	wrapped._PLUMTYPE = 'RDF_Date';
	wrapped.isBefore = function(x) {
	    assert(x.getTime, "Cannot compare a time with a " + typeof(x));
	    return this.getTime() < x.getTime();
	}
	wrapped.isAfter = function(x) {
	    assert(x.getTime, "Cannot compare a time with a " + typeof(x));
	    return this.getTime() > x.getTime();
	}
	bindMethods(wrapped);
	// yadda dadda..
	return wrapped;
    },
    "string" : function(s) {
	var wrapped = object(s);
	wrapped._PLUMTYPE = 'RDF_String';
	wrapped.is = function(sarg) {
	    sarg = primitives.unwrap(sarg);
	    if (typeof(sarg) == 'string') sarg = new String(sarg);
	    assert(typeof(sarg) == 'object', "Cannot compare a string to a " + typeof(sarg));
	    assert(sarg.constructor.name == 'String', "Cannot compare a string to a " + sarg.constructor.name);
	    if (s.toLowerCase().indexOf(sarg.toLowerCase()) >= 0) return true;
	    if (sarg.toLowerCase().indexOf(s.toLowerCase()) >= 0) return true;
	    return false;
	}
	wrapped.toLowerCase = function() { return primitives.wrap(s.toLowerCase()); };
	wrapped.strlen = function() { return primitives.wrap(s.valueOf().length); };
	wrapped.indexOf = function(arg) { return primitives.wrap(s.valueOf().indexOf(primitives.unwrap(arg))); };
	wrapped.valueOf = function() { return s.valueOf(); }
	return wrapped;
    },
    "number" : function(n) {
	// numeric ops here. very useful
	var mkmap = MochiKit.Base.map;	
	var N = new Number(n);
	var wrapped = object(N);
	wrapped._PLUMTYPE = 'RDF_Number';
	wrapped._value = n;
	wrapped.valueOf = function() { return n;} 
	forEach(['eq','ne','gt','ge','lt','le','ceq','add','sub','div','mod','mul'],
		function(op) {
		    // must watch out for binding op
		    wrapped[op] = function(op_) {
			return function() {
			    return primitives.wrap(operator[op_].apply(null,[wrapped._value].concat(mkmap(primitives.unwrap,arguments))));;
			};
		    }(op);
		});		    
	return wrapped;
    },
    "boolean": function(b) {
	var B = new Boolean(b);
	var wrapped = object(B);
	wrapped._PLUMTYPE = 'RDF_Boolean';
	wrapped.valueOf = function() { return b; }
	return wrapped;
    },
    "unwrap": function(wrapped) {
	if (wrapped === undefined) return undefined;
	if (wrapped === null) return null;
	if (typeof(wrapped) != "object") return wrapped;
	if (!primitives.isWrapped(wrapped)) return wrapped;
	if (wrapped.valueOf && typeof(wrapped.valueOf) == 'function')
	    return wrapped.valueOf();
	return wrapped;
    },
    "wrap" : function(prim) {
	if (prim === undefined /*|| isNaN(prim) */) { logWarning('cannot transform ' + prim); return undefined; }

	// _never_ try to wrap an fset!
	if (isfset(prim)) return prim;
	
	// if already wrapped, wrapped .
	if (primitives.isWrapped(prim))  return prim;

	// 
	var primtype = typeof(prim);
	if (typeof(prim) != 'object' && primitives[primtype]) {
	    return primitives[primtype](prim); 
	}
	if (typeof(prim) == 'object' && primitives[prim.constructor.name]) {
	    return primitives[prim.constructor.name](prim);		
	}
	// can't convert.. just as it is!
	return prim;
    },
    "isWrapped": function(wrapped) {
	return !isfset(wrapped) && wrapped._PLUMTYPE && typeof(wrapped._PLUMTYPE) === 'string' && wrapped._PLUMTYPE.indexOf('RDF_') == 0;
    },



//     These are dangerous ... at least if you try to use them as replacements
//     for main map/filter
//     "filter" : function(/*arguments*/) {
// 	var args = MochiKit.Base.extend(null,arguments);
// 	var fn = args[0];
// 	var args = args.slice(1,arguments.length);
// 	var mkmap = MochiKit.Base.map;
// 	var mkfilter = MochiKit.Base.filter;
// 	var result = mkfilter.apply(null, [function() {
// 	    var _mfargs = extend(null,arguments);
// 	    return fn.apply( null, mkmap(primitives.unwrap,_mfargs));}].concat(args));
// 	return result;
//     },
//     "map" : function(/*arguments*/) {
// 	/* wait can this support multiarg map ? */
// 	var args = MochiKit.Base.extend(null,arguments);
// 	var fn = args[0];
// 	var args = args.slice(1,arguments.length);
// 	var mkmap = MochiKit.Base.map;
// 	var result =  mkmap.apply(null, [
// 	    function() {
// 		var _mfargs = extend(null,arguments);
// 		return primitives.wrap(fn.apply(null,mkmap(primitives.unwrap,_mfargs)));
// 	    }
// 	].concat(args));
// 	return result;
//     },

    "safefilter" : function(/*arguments*/) {
        var args = MochiKit.Base.extend(null,arguments);
	var fn = args[0];
	var remainders = args.slice(1,arguments.length);
        return MochiKit.Base.filter.apply(this, [function() {
             var fargs = MochiKit.Base.extend(null,arguments);
             var result = fn.apply(null, fargs);
             return primitives.unwrap(result);
        }].concat(remainders));
    }

    /*      "Array" : function() {    // ??     };    */
}

// bind these to the global env
//self.filter = primitives.filter;
//self.map = primitives.map;
self.filter = primitives.safefilter;