var PlumStoreTypeHandler = {
    plumutil:plumutil,
    factory : [],
    byType : {},
    newify:PROTOCROCK.newify,
    PlumStoreTypeNames:PlumStoreTypeNames,
    push: function(c) {
	this.byType[c.type] = c;
	this.factory.push(c);
    },
    initialize: function() {
	var _plumutil = this.plumutil;
	this.push(
	    {
		plumutil:_plumutil,
		// serializes REFERENCES to resources, e.g., URIs,
		type : this.PlumStoreTypeNames.DATETIME,
		testserial : function(obj) {
		    return obj !== undefined && typeof(obj) == 'object' && obj.getTimezone && obj.toJSONString;
		},
		serialize: function(obj) {
		    return ""+obj.valueOf()+this.type;
		},
		testdeserial:function(sobj) {
		    if (sobj === undefined) return false;
		    sobj = sobj.valueOf();
		    var index = sobj.indexOf(this.type);
		    return typeof(sobj) == "string" && index >= 0 && (index + this.type.length == sobj.length);
		},
		deserialize: function(sobj) {
		    return new Date(parseInt(sobj.substring(0,sobj.indexOf(this.type))));
		}
	    });
	this.push(
	    {
		plumutil:_plumutil,
		type : this.PlumStoreTypeNames.BOOLEAN,
		testserial :function(obj) {
		    return obj !== undefined && typeof(obj.valueOf()) == "boolean" && (obj === true || obj === false);
		},
		serialize: function(obj) {
		    var serialized = ""+ obj.valueOf() + this.type;
		    return ""+ obj.valueOf() + this.type;
		},
		testdeserial:function(sobj) {
		    if (sobj === undefined) return false;
		    sobj = sobj.valueOf();
		    var index = sobj.indexOf(this.type);
		    return typeof(sobj) == "string" && index >= 0 && (index + this.type.length == sobj.length);
		},
		deserialize: function(sobj) {
		    this.plumutil.assert(sobj.indexOf(this.type) >= 0, "Error deserializing boolean " + sobj);
		    var s =sobj.substring(0,sobj.indexOf(this.type));
		    if (s == 'true') return true;
		    if (s == 'false') return false;
		    this.plumutil.assert(false,"Error deserializing boolean " + s + " " + sobj);
		}
	    });	
	this.push(
	    {
		plumutil:_plumutil,
		type : this.PlumStoreTypeNames.INTEGER,
		testserial :function(obj) {
		    return obj !== undefined && obj.valueOf && typeof(obj.valueOf()) == "number" &&
			parseFloat(obj.valueOf()) - Math.round(parseFloat(obj.valueOf())) === 0;
		},
		serialize: function(obj) {
		    return ""+ obj.valueOf() + this.type;
		},
		testdeserial:function(sobj) {
		    if (sobj === undefined) return false;
		    sobj = sobj.valueOf();
		    var index = sobj.indexOf(this.type);
		    return typeof(sobj) == "string" && index >= 0 && (index + this.type.length == sobj.length);
		},
		deserialize: function(sobj) {
		    return parseInt(sobj.substring(0,sobj.indexOf(this.type)));
		}
	    });
	this.push(
	    {
		plumutil:_plumutil,
		type : this.PlumStoreTypeNames.FLOAT,
		testserial :function(obj) {
		    return obj !== undefined && obj.valueOf && typeof(obj.valueOf()) == "number" && parseFloat(obj.valueOf())%1>0;
		},
		serialize: function(obj) {
		    return "" + obj.valueOf() + this.type;
		},
		testdeserial:function(sobj) {
		    if (sobj === undefined) return false;
		    sobj = sobj.valueOf();
		    var index = sobj.indexOf(this.type);
		    return typeof(sobj) == "string" && index >= 0 && (index + this.type.length == sobj.length);
		},
		deserialize: function(sobj) {
		    return parseFloat(sobj.substring(0,sobj.indexOf(this.type)));
		}
	    });
	this.push(
	    {
		plumutil:_plumutil,
		// serializes REFERENCES to resources, e.g., URIs,
		type : this.PlumStoreTypeNames.RESOURCE,
		testserial : function(obj) {
		    return obj !== undefined && typeof(obj) == 'object' && obj.id;
		},
		serialize: function(obj) {
		    if (obj.id) {
			return ""+obj.id+this.type;
		    }
		    return ""+obj+this.type;
		},
		testdeserial:function(sobj) {
		    if (sobj === undefined) return false;
		    sobj = sobj.valueOf();
		    var index = sobj.indexOf(this.type);
		    return typeof(sobj) == "string" && index >= 0 && (index + this.type.length == sobj.length);
		},
		deserialize: function(sobj) {
		    return obj !== undefined && parseFloat(sobj.substring(0,sobj.indexOf(this.type)));    
		}
	    });	
	this.push( {
	    plumutil:_plumutil,
	    type: this.PlumStoreTypeNames.SEQ,
	    testserial: function(obj) {
		return obj !== undefined && this.plumutil.isArrayLike(obj);
	    },
	    serialize: function(obj) {
		return ""+obj+this.type;
	    },
	    testdeserial: function(sobj) {
		var index = sobj.indexOf(this.type);
		return typeof(sobj) == "string" && index >= 0 && (index + this.type.length == sobj.length);
	    },
	    deserialize: function(sobj) {
		if (sobj === undefined) return false;
		return sobj.substring(0,sobj.indexOf(this.type));
	    }
	});
	// string
	this.push(
	    {
		plumutil:_plumutil,
		type : this.PlumStoreTypeNames.STRING,
		testserial :function(obj) {
		    return obj !== undefined && obj.valueOf && typeof(obj.valueOf()) == "string";
		},
		serialize: function(obj) {
		    return "" + encodeURIComponent(obj.valueOf()) + this.type;
		},
		testdeserial:function(sobj) {
		    if (sobj === undefined) return false;
		    sobj = sobj.valueOf();
		    var index = sobj.indexOf(this.type);
		    return typeof(sobj) == "string" && index >= 0 && (index + this.type.length == sobj.length);
		},
		deserialize: function(sobj) {
		    return decodeURIComponent(sobj.substring(0,sobj.indexOf(this.type)));		    
		}
	    });
    },
    serialize: function(obj,type) {
	if (type) {
	    var sfn = this.byType[type];
	    this.plumutil.assert(sfn, "No serializer found for type : " + type);
	    return sfn.serialize(obj); 
	}
	// no serializer, find the "best fit" 
	var factory = this.factory;
	for (var ci = 0; ci < factory.length; ci++) {
	    if(factory[ci].testserial(obj)) {
		return factory[ci].serialize(obj);
	    }
	}
	return undefined;
    },
    deserialize: function(obj) {
	var factory = this.factory;
	for (var ci = 0; ci < factory.length; ci++) {
	    if(factory[ci].testdeserial(obj)) {
		return factory[ci].deserialize(obj);
	    }
	}
	return undefined;
    },
    indexOfSerializer: function(type) {
	for (var i=0; i < this.factory.length; i++) {
	    if (this.factory[i].type == type || this.factory[i].type == this.PlumStoreTypeNames[type]) return i;
	}
	return -1;
    },
    isSerializedType: function(obj,type) {
	var serializer = this.byType[type];
	if (serializer === undefined && this.PlumStoreTypeNames[type.toUpperCase()])  {
	    serializer = this.byType(this.PlumStoreTypeNames[type.toUpperCase()]);
	}
	return serializer && serializer.testdeserial(obj);
    },
    isType: function(obj,type) {
	var serializer = this.byType[type];
	if (serializer === undefined && this.PlumStoreTypeNames[type.toUpperCase()]) {
	    serializer = this.byType(this.PlumStoreTypeNames[type.toUpperCase()]);
	}
	return serializer && serializer.testserial(obj);
    },
    test: function() {
	var psth = this.newify(PlumStoreTypeHandler);
	b = psth.serialize(1.2);
	this.plumutil.assert(b=="1.2^^http://www.w3.org/2001/XMLSchema#float","Failed test 1");
	this.plumutil.assert(psth.deserialize(b) == 1.2, "Failed test 2");
	this.plumutil.assert(typeof(psth.deserialize(b))=='number',"Failed test 3");
	b = psth.serialize(1);
	log("b is " + b);
	this.plumutil.assert(b=="1^^http://www.w3.org/2001/XMLSchema#int","Failed test 4");
	this.plumutil.assert(psth.deserialize(b) == 1, "Failed test 5");
	var d = new Date();
	b = psth.serialize(new Date());
	this.plumutil.assert(psth.deserialize(b).valueOf() == d.valueOf(), "Failed test 6");	
    }
};

